/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/** \example int-list.c int list treedb example
In this example the tree "payload" is a 32 bit signed integer.

int-list.c and int-list.h provide a backend for a treedb_t so that it can manage
int32_t list payloads.

The added complication this example addresses is the alignment requirement of
the payload, in addition to the alignment requirements of the list node.
*/

#include "int-list.h"
#include <stdio.h>
#include <unistd.h>

static PVOID STDCALL L2IntList_AllocNode
( treedb_list_backend_ptr_t This
, treedb_list_runtime_ptr_t runtime
, HTREEDB_LISTDATA pv
) {
	(void)This;
	uintptr_t size = runtime->m_pImpl->m_pDescription->m_aligned_node_size;
	char * ret = (char *)treedb_allocator_t_PointerAlloc(
				  runtime->m_pImpl->m_pAllocator
				, size + sizeof(int32_t));
	if(!ret)
		return 0;
	size_t misalign = treedb_list_t_PointerPack(runtime, ret + size)
		% sizeof(int32_t);
	META_TREEDB_printf2("L2IntList_AllocNode : Alloc'd %zu, storing %d to %zu "
		", misalign %zu.\n"
		, treedb_list_t_PointerToOffset(runtime, ret)
		, *(int32_t *)pv
		, treedb_list_t_PointerToOffset(runtime, ret + size)
		, misalign
	);
	if(misalign) {
		treedb_allocator_t_FreePtr(runtime->m_pImpl->m_pAllocator, ret);
		return 0;
	}
	*(int32_t*)(ret + size) = *(int32_t *)pv;
	return ret;
}
static void STDCALL L2IntList_FreeNode
( treedb_list_backend_ptr_t This
, treedb_list_runtime_ptr_t runtime
, PVOID pv
) {
	(void)This;
	treedb_allocator_t_FreePtr(runtime->m_pImpl->m_pAllocator, pv);
}

#if 0
static void test(treedb_list_backend_vtbl_t * v)
{
	v->AllocNode = & L2IntList_AllocNode;
	v->FreeNode = & L2IntList_FreeNode;
}
#endif

treedb_list_backend_vtbl_t IntList_backend_vtbl = {
	treedb_list_backend_vtbl_t_entries(L2IntList_)
};

void STDCALL IntList_Dump
( treedb_list_ptr_t This
) {
	PVOID node = treedb_list_t_PointerHead(This);
	int n;
	for(n = 0; node; node = treedb_list_t_PointerNodeNextPtr(This, node), ++n)
		printf("[%d] %d\n", n, *(int32_t *)
			(((char *)node) + This->m_pImpl->m_pDescription->m_aligned_node_size)
		);
}
