#!/bin/sh

# Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)
# License: LGPLv3.  See LICENSE.txt for the full license.

{ run dump test.sb && {
	echo "Dump non-existing file test didn't fail as expected."
	exit 1
} } || true

################################################################################
# string set
################################################################################
run string set init test.sb $bits
if test ! -f test.sb; then
	echo "Failed to create test.sb."
	exit 1
fi
for a in apple orange lemon pear plum banana; do
	run string set insert test.sb $a
done
expected_fn()
{
	cat <<EOF
apple
banana
lemon
orange
pear
plum
EOF
}
expected=$(expected_fn)
got=$($app dump test.sb)

if test "$got" != "$expected"; then
	echo "After inserts: wanted : $expected\nGot:\n$got\ninstead."
	exit 1
fi
for a in apple orange pear banana; do
	run string set remove test.sb $a
done
expected_fn2()
{
	cat <<EOF
lemon
plum
EOF
}
expected=$(expected_fn2)
got=$($app dump test.sb)
if test "$got" != "$expected"; then
	echo "After deletes: wanted : $expected\nGot:\n$got\ninstead."
	exit 1
fi
{ run string set remove test.sb banana && {
	echo "Remove a non-existing item test didn't fail as expected."
	exit 1
} } || true
{ run string set insert test.sb lemon && {
	echo "Adding an existing item test didn't fail as expected."
	exit 1
} } || true
{ run string set contains test.sb banana && {
	echo "Contains non-existing item test didn't fail as expected."
	exit 1
} } || true
run string set contains test.sb plum || {
	echo "Contains existing item test failed."
	exit 1
}
run check test.sb || {
	echo "Heap check failed."
	exit 1
}
testme()
{
	local arg="$1"
	if test -z "$arg" ; then arg="$bits"; fi
	run string set init test.sb $arg
	local n=$LARGE_VAL
	if test "$arg" = 16; then
		n=$LARGE_VAL_16
	fi
	v3c range 0 $n | run stream test.sb || {
		echo "Big string set insert test failed."
		exit 1
	}
	expected="test.sb is a $arg bit string set treedb [f6ef86e3-d8c6-407c-938b-d36d5e8a2835] with $n entries."
	got=$($app type test.sb)
	echo "### $got"
	got="$(echo "$got" | head -n 1)"
	if test "$got" != "$expected"; then
		echo -e "Stream type failed : wanted \n\t$expected"
		exit 1
	fi
}
testme
testme 16
testme 32
if test "$bits" = "64" ; then
testme 64
fi

################################################################################
# int set
################################################################################
run int set init test.sb
if test ! -f test.sb; then
	echo "Failed to create test.sb."
	exit 1
fi
for a in 1 9 8 4 6 5 7 3 2; do
	run int set insert test.sb $a
done
expected_fn()
{
	cat <<EOF
1
2
3
4
5
6
7
8
9
EOF
}
expected=$(expected_fn)
got=$($app dump test.sb)

if test "$got" != "$expected"; then
	echo "After inserts: wanted : $expected\nGot:\n$got\ninstead."
	exit 1
fi
for a in 1 3 5 7 9; do
	run int set remove test.sb $a
done
expected_fn2()
{
	cat <<EOF
2
4
6
8
EOF
}
expected=$(expected_fn2)
got=$($app dump test.sb)
if test "$got" != "$expected"; then
	echo "After deletes: wanted : $expected\nGot:\n$got\ninstead."
	exit 1
fi
{ run int set remove test.sb 3 && {
	echo "Remove a non-existing item test didn't fail as expected."
	exit 1
} } || true
{ run string set insert test.sb "a" && {
	echo "Adding the wrong type test didn't fail [XPASS]."
	exit 1
} } || { ret=$?
	if test "$ret" != "$eUnknownSchema" ; then
		echo -n "Adding the wrong type test failed with $(result_name $ret)"
		echo ", not $(result_name $eUnknownSchema)."
		exit 1
	fi
}
{ run int set insert test.sb 2 && {
	echo "Adding an existing item test didn't fail as expected."
	exit 1
} } || true
{ run int set contains test.sb 7 && {
	echo "Contains non-existing item test didn't fail as expected."
	exit 1
} } || true
run int set contains test.sb 8 || {
	echo "Contains existing item test failed."
	exit 1
}
run check test.sb || {
	echo "Heap check failed."
	exit 1
}
testme()
{
	local arg="$1"
	if test -z "$arg" ; then arg="$bits"; fi
	run int set init test.sb $arg
	local n=$LARGE_VAL
	if test "$arg" = 16; then
		n=$LARGE_VAL_16
	fi
	v3c range 0 $n | run stream test.sb || {
		echo "Big int set insert test failed."
		exit 1
	}
	expected="test.sb is a $arg bit int set treedb [882d186f-a880-4565-b9ab-bcb93614f96d] with $n entries."
	got=$($app type test.sb)
	echo "### $got"
	got="$(echo "$got" | head -n 1)"
	if test "$got" != "$expected"; then
		echo -e "Stream type failed : wanted \n\t$expected"
		exit 1
	fi
}
testme
testme 16
testme 32
if test "$bits" = "64" ; then
testme 64
fi
################################################################################
# string list
################################################################################
run string list init test.sb
if test ! -f test.sb; then
	echo "Failed to create test.sb."
	exit 1
fi
for a in apple orange lemon pear plum banana; do
	run string list push_back test.sb $a
done
expected_fn()
{
	cat <<EOF
apple
orange
lemon
pear
plum
banana
EOF
}
expected=$(expected_fn)
got=$($app dump test.sb)

if test "$got" != "$expected"; then
	echo "After pushes: wanted : $expected\nGot:\n$got\ninstead."
	exit 1
fi
for a in 1 2 3 4; do
	run string list pop_back test.sb
done
expected_fn2()
{
	cat <<EOF
apple
orange
EOF
}
expected=$(expected_fn2)
got=$($app dump test.sb)
if test "$got" != "$expected"; then
	echo "After pops: wanted : $expected\nGot:\n$got\ninstead."
	exit 1
fi
run check test.sb || {
	echo "Heap check failed."
	exit 1
}
testme()
{
	local arg="$1"
	if test -z "$arg" ; then arg="$bits"; fi
	run string list init test.sb $arg
	local n=$LARGE_VAL
	if test "$arg" = 16; then
		n=$LARGE_VAL_16
	fi
	v3c range 0 $n | run stream test.sb || {
		echo "Big string list insert test failed."
		exit 1
	}
	expected="test.sb is a $arg bit string list treedb [a1f7dba4-baa5-4d51-8c2a-8031db6961ad] with $n entries."
	got=$($app type test.sb)
	echo "### $got"
	got="$(echo "$got" | head -n 1)"
	if test "$got" != "$expected"; then
		echo -e "Stream type failed : wanted \n\t$expected"
		exit 1
	fi
}
testme
testme 16
testme 32
if test "$bits" = "64" ; then
testme 64
fi

################################################################################
# int list
################################################################################
run int list init test.sb
if test ! -f test.sb; then
	echo "Failed to create test.sb."
	exit 1
fi
for a in 1 9 8 4 6 5 7 3 2; do
	run int list push_back test.sb $a
done
expected_fn()
{
	cat <<EOF
1
9
8
4
6
5
7
3
2
EOF
}
expected=$(expected_fn)
got=$($app dump test.sb)

if test "$got" != "$expected"; then
	echo "After pushes: wanted : $expected\nGot:\n$got\ninstead."
	exit 1
fi
for a in 1 2 3 4; do
	run int list pop_back test.sb
done
expected_fn2()
{
	cat <<EOF
1
9
8
4
6
EOF
}
expected=$(expected_fn2)
got=$($app dump test.sb)
if test "$got" != "$expected"; then
	echo "After pops: wanted : $expected\nGot:\n$got\ninstead."
	exit 1
fi
run check test.sb || {
	echo "Heap check failed."
	exit 1
}
testme()
{
	local arg="$1"
	if test -z "$arg" ; then arg="$bits"; fi
	run int list init test.sb $arg
	local n=$LARGE_VAL
	if test "$arg" = 16; then
		n=$LARGE_VAL_16
	fi
	v3c range 0 $n | run stream test.sb || {
		echo "Big int list insert test failed."
		exit 1
	}
	expected="test.sb is a $arg bit int list treedb [0099bb6d-4262-4b98-b333-590f4f0bf33f] with $n entries."
	got=$($app type test.sb)
	echo "### $got"
	got="$(echo "$got" | head -n 1)"
	if test "$got" != "$expected"; then
		echo -e "Stream type failed : wanted \n\t$expected"
		exit 1
	fi
}
testme
testme 16
testme 32
if test "$bits" = "64" ; then
testme 64
fi

rm test.sb
