/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include "config.h"

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <v3c/bits.h>
#include "meta-comet.h"
#include "meta-avl.h"

#if 0
static int test()
{
	treedb_storage_on_file_t m_storage
		= TREEDB_STORAGE_ON_FILE_INIT;
	treedb_allocator_on_storage_impl_t m_allocator_impl
		= META_TREEDB_ALLOCATOR_ON_STORAGE_IMPL_INIT;
	treedb_allocator_on_storage_t m_allocator
		= META_TREEDB_ALLOCATOR_ON_STORAGE_INIT;
	treedb_tree_backend_t m_free_backend
		= META_TREEDB_TREE_BACKEND_INIT;
}
#endif

CometCommandStatus meta_comet_read_init_unknown
( meta_comet_ptr_t runtime
, const char * filename
, BOOL readonly
) {
	TreedbHeapStatus res = treedb_storage_t_OpenUnknown
		( & runtime->m_storage
		, filename
		, readonly ? eHeapOpenRead : (eHeapOpenRead | eHeapOpenWrite)
		, (readonly ? 0 : eHeapAccessWrite) | eHeapAccessRead
			| eHeapAccessShared
		, 0
		);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "meta_comet_read_init_unknown : Failed to open \"%s\" :"
			" %s.\n" , filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}
	/* We only know of one allocator implementation here - avl. */
	if(!treedb_init_allocator_on_storage_from_template
			( & runtime->m_allocator
			, & runtime->m_allocator_impl
			, runtime->m_storage.m_pTreedb->m_uBytesPerOffset
			, runtime->m_storage.m_pTreedb->m_uFeatures * 4
			, sizeof(TreedbHeader)
			, (treedb_storage_ptr_t)(& runtime->m_storage)
			)) {
		treedb_storage_t_Unlock(& runtime->m_storage);
		treedb_storage_t_Close(& runtime->m_storage);
		return eCometCommandUnknownSchema;
	}
	if(uuid_compare(treedb_allocator_t_Guid(& runtime->m_allocator)
			, runtime->m_storage.m_pTreedb->m_uAllocatorAlgorithm)) {
		treedb_storage_t_Unlock(& runtime->m_storage);
		treedb_storage_t_Close(& runtime->m_storage);
		return eCometCommandUnknownAllocatorAlgorithm;
	}
	return eCometCommandSuccess;
}
CometCommandStatus meta_comet_write_init
( meta_comet_ptr_t runtime
, const char * filename
, const uuid_t schema
, uint16_t bpo
, uint16_t uPayloadOffsCount
) {
	if(!treedb_init_allocator_on_storage_from_template
			( & runtime->m_allocator
			, & runtime->m_allocator_impl
			, bpo
			, eTreedbContainerHasCount | eTreedbContainerHasDepth
			, sizeof(TreedbHeader)
			, (treedb_storage_ptr_t)(& runtime->m_storage))
	)
		return eCometCommandBadCommand;

	TreedbHeapStatus res = treedb_storage_t_Create
		( & runtime->m_storage
		, (treedb_allocator_ptr_t)(& runtime->m_allocator)
		, schema
		, filename
		, eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0, TREEDB_PAGE_SIZE
		, uPayloadOffsCount
			* runtime->m_allocator.m_pImpl->m_pDescription->m_pointer.m_uSize
		);
	if(res != eTreedbHeapSuccess) {
		if(TreedbHeapMismatched(res))
			return eCometCommandUnknownSchema;
		fprintf(stderr, "meta_comet_write_init : Failed to open \"%s\" : %s.\n"
			, filename, TreedbHeapStatusDesc(res));
		if(res == eTreedbHeapFileOpenFailure)
			return eCometCommandFileOpenFailure;
		return eCometCommandHeapError;
	}

	return eCometCommandSuccess;
}

CometCommandStatus meta_comet_create
( meta_comet_ptr_t runtime
, const char * filename
, CometDetailsPtr pDetails
, uint16_t bpo
) {
	CometCommandStatus res = meta_comet_write_init(runtime, filename
		, pDetails->schema
		, bpo, pDetails->uPayloadOffsCount);
	if(res != eCometCommandSuccess)
		return res;
	res = pDetails->fn(runtime);
	treedb_storage_t_Unlock(& runtime->m_storage);
	return res;
}
CometCommandStatus meta_comet_read
( meta_comet_ptr_t runtime
, const char * filename
, PFN_META_COMET_FUNCTION fn
) {
	CometCommandStatus res = meta_comet_read_init_unknown(runtime, filename, 1);
	if(res != eCometCommandSuccess)
		return res;
	res = fn(runtime);
	treedb_storage_t_Unlock(& runtime->m_storage);
	return res;
}
CometCommandStatus meta_comet_read2
( meta_comet_ptr_t runtime
, const char * filename
, PFN_META_COMET_FUNCTION2 fn
) {
	CometCommandStatus res = meta_comet_read_init_unknown(runtime, filename, 1);
	if(res != eCometCommandSuccess)
		return res;
	res = fn(runtime, filename);
	treedb_storage_t_Unlock(& runtime->m_storage);
	return res;
}

extern CometDetails int_tree_details;
extern CometDetails string_tree_details;
extern CometDetails int_list_details;
extern CometDetails string_list_details;

typedef enum CometType
{ eCometTypeString = 1 /* bit not set == int */
, eCometTypeTree = 2   /* bit not set == list */
, eCometTypeUnknown = 4
} CometType;

CometDetailsPtr details_for_type(unsigned int type)
{
	switch(type) {
	case 0:
		return & int_list_details;
	case eCometTypeString:
		return & string_list_details;
	case eCometTypeTree:
		return & int_tree_details;
	case eCometTypeString | eCometTypeTree:
		return & string_tree_details;
	default:
		return 0;
	}
}

CometCommandStatus meta_comet_update
( meta_comet_ptr_t runtime
, const char * filename
, unsigned int type
, PFN_META_COMET_UPDATE_FUNCTION fn
, int argc
, char * argv[]
) {
	CometCommandStatus res = meta_comet_read_init_unknown(runtime, filename, 0);
	if(res != eCometCommandSuccess)
		return res;
	CometDetailsPtr pd = details_for_type(type);
	if(!uuid_compare(pd->schema, runtime->m_storage.m_pTreedb->m_uSchema))
		res = fn(runtime, argc, argv);
	else
		res = eCometCommandUnknownSchema;
	treedb_storage_t_Unlock(& runtime->m_storage);
	return res;
}

CometCommandStatus meta_comet_update2
( meta_comet_ptr_t runtime
, const char * filename
, PFN_META_COMET_FUNCTION fn
) {
	CometCommandStatus res = meta_comet_read_init_unknown(runtime, filename, 0);
	if(res != eCometCommandSuccess)
		return res;
	res = fn(runtime);
	treedb_storage_t_Unlock(& runtime->m_storage);
	return res;
}

extern CometCommandStatus int_tree_dump(meta_comet_ptr_t runtime);
extern CometCommandStatus string_tree_dump(meta_comet_ptr_t runtime);
extern CometCommandStatus int_list_dump(meta_comet_ptr_t runtime);
extern CometCommandStatus string_list_dump(meta_comet_ptr_t runtime);

CometCommandStatus meta_comet_dump(meta_comet_ptr_t runtime)
{
	if(!uuid_compare(UUID_treedb_int_tree_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return int_tree_dump(runtime);
	}else if(!uuid_compare(UUID_treedb_string_tree_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return string_tree_dump(runtime);
	}else if(!uuid_compare(UUID_treedb_int_list_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return int_list_dump(runtime);
	}else if(!uuid_compare(UUID_treedb_string_list_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return string_list_dump(runtime);
	}
	return eCometCommandUnknownSchema;
}

extern CometCommandStatus int_tree_count(meta_comet_ptr_t runtime);
extern CometCommandStatus string_tree_count(meta_comet_ptr_t runtime);
extern CometCommandStatus int_list_count(meta_comet_ptr_t runtime);
extern CometCommandStatus string_list_count(meta_comet_ptr_t runtime);

CometCommandStatus meta_comet_count(meta_comet_ptr_t runtime)
{
	if(!uuid_compare(UUID_treedb_int_tree_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return int_tree_count(runtime);
	}else if(!uuid_compare(UUID_treedb_string_tree_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return string_tree_count(runtime);
	}else if(!uuid_compare(UUID_treedb_int_list_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return int_list_count(runtime);
	}else if(!uuid_compare(UUID_treedb_string_list_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return string_list_count(runtime);
	}
	return eCometCommandUnknownSchema;
}

extern CometCommandStatus int_tree_stream(meta_comet_ptr_t runtime);
extern CometCommandStatus string_tree_stream(meta_comet_ptr_t runtime);
extern CometCommandStatus int_list_stream(meta_comet_ptr_t runtime);
extern CometCommandStatus string_list_stream(meta_comet_ptr_t runtime);

CometCommandStatus meta_comet_stream(meta_comet_ptr_t runtime)
{
	if(!uuid_compare(UUID_treedb_int_tree_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return int_tree_stream(runtime);
	}else if(!uuid_compare(UUID_treedb_string_tree_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return string_tree_stream(runtime);
	}else if(!uuid_compare(UUID_treedb_int_list_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return int_list_stream(runtime);
	}else if(!uuid_compare(UUID_treedb_string_list_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return string_list_stream(runtime);
	}
	return eCometCommandUnknownSchema;
}

extern CometCommandStatus int_tree_type(meta_comet_ptr_t runtime
, const char * filename);
extern CometCommandStatus string_tree_type(meta_comet_ptr_t runtime
, const char * filename);
extern CometCommandStatus int_list_type(meta_comet_ptr_t runtime
, const char * filename);
extern CometCommandStatus string_list_type(meta_comet_ptr_t runtime
, const char * filename);

CometCommandStatus meta_comet_type(meta_comet_ptr_t runtime
, const char * filename)
{
	if(!uuid_compare(UUID_treedb_int_tree_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return int_tree_type(runtime, filename);
	}else if(!uuid_compare(UUID_treedb_string_tree_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return string_tree_type(runtime, filename);
	}else if(!uuid_compare(UUID_treedb_int_list_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return int_list_type(runtime, filename);
	}else if(!uuid_compare(UUID_treedb_string_list_schema
			, runtime->m_storage.m_pTreedb->m_uSchema)) {
		return string_list_type(runtime, filename);
	}
	return eCometCommandUnknownSchema;
}
CometCommandStatus meta_comet_check(meta_comet_ptr_t runtime)
{
	if(treedb_allocator_t_Check(& runtime->m_allocator))
		return eCometCommandHeapError;
	return eCometCommandSuccess;
}
const char * program = 0;
const char * program_name(const char *p)
{
	const char * name = strrchr(p,'/');
	if(!name)
		return p;
	if(name != (p + strlen(p) - 1))
		name++;
	return name;
}
void usage()
{
	printf("Usage:"
		"\t%s <string|int> <list|set> <op> <filename> [arg]\n"
		"\t%s <string|int> <list|set> <op> <filename>\n"
		"\t%s <op> <filename> [arg]\n\n"
		"Common operations are type, init|create, count stream, check and"
		" dump.\n"
		"For sets, the extra operations are insert, remove and contains.\n"
		"For lists, the valid operations are reverse, push_back, push_front,"
		" pop_back, and pop_front.\n"
		"To create a default sandbox, use\n"
		"\t%s <string|int> <list|set> <create|init> <filename> [bits],\n"
		"where bits is one of 16,32,64. If omitted it uses the default for\n"
		"your platform - 32 or 64 bits.\n"
		"Stream is the slip-stream way of populating a treedb from a pipe,\n"
		"\tas in find . | %s stream <filename>\n"
		"You cannot create or read a 64 bit treedb on a 32 bit platform.\n"
		, program, program, program, program, program);
}
unsigned int get_comet_type(const char * a, const char * b)
{
	unsigned int res = 0;

	if(!strcmp(a, "int")) {
		;
	}else if(!strcmp(a, "string")) {
		res |= eCometTypeString;
	}else
		return eCometTypeUnknown;

	if(!strcmp(b, "list")) {
		;
	}else if(!strcmp(b, "set")) {
		res |= eCometTypeTree;
	}else
		return eCometTypeUnknown;

	return res;
}

extern CometCommandStatus int_list_process(meta_comet_ptr_t runtime, int argc
, char * argv[]);
extern CometCommandStatus string_list_process(meta_comet_ptr_t runtime, int argc
, char * argv[]);
extern CometCommandStatus int_tree_process(meta_comet_ptr_t runtime, int argc
, char * argv[]);
extern CometCommandStatus string_tree_process(meta_comet_ptr_t runtime, int argc
, char * argv[]);

CometCommandStatus meta_comet_process
( meta_comet_ptr_t runtime
, const char * filename
, unsigned int type
, int argc
, char * argv[]
) {
	switch(type) {
	case 0:
		return meta_comet_update(runtime, filename, type, int_list_process
			, argc, argv);
	case eCometTypeString:
		return meta_comet_update(runtime, filename, type, string_list_process
			, argc, argv);
	case eCometTypeTree:
		return meta_comet_update(runtime, filename, type, int_tree_process
			, argc, argv);
	case eCometTypeString | eCometTypeTree:
		return meta_comet_update(runtime, filename, type, string_tree_process
			, argc, argv);
	default:
		return eCometCommandUnrecognisedCommand;
	}
}
/** Entry point for meta-comet
*/
int main(int argc, char *argv[])
{
	meta_comet_t mc = META_COMET_INIT;
	unsigned int type = eCometTypeUnknown;
	program = program_name(argv[0]);
	switch(argc) {
	case 2:
		if(!strcmp(argv[1], "-?")) {
			usage();
			return 0;
		}else if(!strcmp(argv[1], "-h")) {
			usage();
			return 0;
		}else if(!strcmp(argv[1], "-v")) {
			printf(PACKAGE_VERSION "\n");
			return 0;
		}else if(!strcmp(argv[1], "--help")) {
			usage();
			return 0;
		}else if(!strcmp(argv[1], "--version")) {
			printf(PACKAGE_VERSION "\n");
			return 0;
		}else if(!strcmp(argv[1], "man")) {
			printf(
			".\\\"                                      Hey, EMACS: -*- nroff -*-\n"
			".\\\" First parameter, NAME, should be all caps\n"
			".\\\" Second parameter, SECTION, should be 1-8, maybe w/ subsection\n"
			".\\\" other parameters are allowed: see man(7), man(1)\n"
			".TH meta-treedb 1 \"" PACKAGE_SHORT_BUILD_TIME "\"\n"
			".\\\" Please adjust this date whenever revising the manpage.\n"
			".\\\"\n"
			".\\\" Some roff macros, for reference:\n"
			".\\\" .nh        disable hyphenation\n"
			".\\\" .hy        enable hyphenation\n"
			".\\\" .ad l      left justify\n"
			".\\\" .ad b      justify to both left and right margins\n"
			".\\\" .nf        disable filling\n"
			".\\\" .fi        enable filling\n"
			".\\\" .br        insert line break\n"
			".\\\" .sp <n>    insert n+1 empty lines\n"
			".\\\" for manpage-specific macros, see man(7)\n"
			".SH NAME\n"
			"v3c\\-meta\\-comet \\- a meta\\-treedb demo program\n"
			".SH SYNOPSIS\n"
			".B v3c\\-meta\\-comet\n"
			".RI [ command ] \" options\" ...\n"
			".SH DESCRIPTION\n"
			"This manual page documents briefly the\n"
			".B v3c\\-meta\\-comet\n"
			"command.\n"
			".PP\n"
			".\\\" TeX users may be more comfortable with the \\fB<whatever>\\fP and\n"
			".\\\" \\fI<whatever>\\fP escape sequences to invode bold face and italics,\n"
			".\\\" respectively.\n"
			"\\fBv3c\\-meta\\-comet\\fP is a meta\\-treedb demo program with the following commands\n"
			".br\n");
			char * rp = realpath(argv[0], NULL);
			if(!rp)
				return 2;
			char cmd[4096];
			sprintf(cmd, "%s --help | sed -e 's|-|\\\\-|g' "
				"-e 's|$|\\n\\.br|g' -re 's|(.+) :|\\\\fB\\1\\\\fP :|g'", rp);
			free(rp);
			printf(".\\\" %s\n", cmd);
			fflush(stdout);
			return system(cmd);
		}
		usage();
		return 1;
	case 3:
		/* prg dump <filename> */
		if(!strcmp(argv[1], "dump"))
			return meta_comet_read(& mc, argv[2], meta_comet_dump);
		if(!strcmp(argv[1], "check"))
			return meta_comet_read(& mc, argv[2], meta_comet_check);
		if(!strcmp(argv[1], "count"))
			return meta_comet_read(& mc, argv[2], meta_comet_count);
		if(!strcmp(argv[1], "stream"))
			return meta_comet_update2(& mc, argv[2], meta_comet_stream);
		if(!strcmp(argv[1], "type"))
			return meta_comet_read2(& mc, argv[2], meta_comet_type);
		break;
	case 5:
		/* prg string set <op> <filename> */
		type = get_comet_type(argv[1], argv[2]);
		if(type != eCometTypeUnknown) {
			if(!strcmp(argv[3], "dump"))
				return meta_comet_read(& mc, argv[4], meta_comet_dump);
			if(!strcmp(argv[3], "count"))
				return meta_comet_read(& mc, argv[4], meta_comet_count);
			if(!strcmp(argv[3], "create") || !strcmp(argv[3], "init")) {
				CometDetailsPtr pdetails = details_for_type(type);
				if(!pdetails)
					break;
				return meta_comet_create(& mc, argv[4], pdetails, POINTER_SIZE);
			}
		}
		return meta_comet_process(& mc, argv[4], type, argc, argv);
	case 6:
		/* prg int list <op> <filename> <arg> */
		type = get_comet_type(argv[1], argv[2]);
		if((type != eCometTypeUnknown)
				&& (!strcmp(argv[3], "create") || !strcmp(argv[3], "init"))) {
			uint16_t bpo;
			{
				char * tailptr;
				unsigned long item = strtoul(argv[5], & tailptr, 10);
				if(!tailptr || *tailptr) {
					fprintf(stderr, "Error \"%s\" isn't a number.\n", argv[5]);
					return eCometCommandBadCommand;
				}
				if(item >= (1 << (sizeof(uint16_t) * 8))) {
					fprintf(stderr, "Error \"%s\" is too large.\n", argv[5]);
					return eCometCommandBadCommand;
				}
				if(v3c_nbits_u64(item) != 1) {
					fprintf(stderr, "Error \"%s\" isn't a power of 2.\n"
						, argv[5]);
					return eCometCommandBadCommand;
				}
				bpo = (uint16_t)item;
			}
			CometDetailsPtr pdetails = details_for_type(type);
			if(!pdetails)
				break;
			return meta_comet_create(& mc, argv[4], pdetails, bpo/8);
		}
		return meta_comet_process(& mc, argv[4], type, argc, argv);
	default:
		usage();
		return eCometCommandUnrecognisedCommand;
	}
	usage();
	return eCometCommandUnrecognisedCommand;
}
