/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/** \example string-list.c string list treedb example
In this example the list "payload" is a variable length char array - a string.

string-list.c and string-list.h provide a backend for a treedb_t so that it can
manage string list payloads.
*/

#include "string-list.h"
#include <stdio.h>
#include <unistd.h>
#include <string.h>

static PVOID STDCALL StringList_AllocNode
( treedb_list_backend_ptr_t This
, treedb_list_runtime_ptr_t runtime
, HTREEDB_LISTDATA pv
) {
	(void)This;
	char * ret = (char *)treedb_allocator_t_PointerAlloc
		( runtime->m_pImpl->m_pAllocator
		, runtime->m_pImpl->m_pDescription->m_aligned_node_size
			+ strlen((const char *)pv) + 1);
	if(!ret)
		return 0;
	char * p = ret + runtime->m_pImpl->m_pDescription->m_aligned_node_size;
#if META_TREEDB_DEBUG
	printf("StringList_AllocNode : Alloc'd %zu, storing \"%s\" in %zu.\n"
		, treedb_list_t_PointerToOffset(runtime, ret), (const char *)pv
		, treedb_list_t_PointerToOffset(runtime, p)
	);
#endif
	if(0) treedb_allocator_t_Check(runtime->m_pImpl->m_pAllocator);
	fflush(stdout);
	strcpy(p, (const char *)pv);
	if(0) treedb_allocator_t_Check(runtime->m_pImpl->m_pAllocator);
	return ret;
}
static void STDCALL StringList_FreeNode
( treedb_list_backend_ptr_t This
, treedb_list_runtime_ptr_t runtime
, PVOID pv
) {
	(void)This;
	treedb_allocator_t_FreePtr(runtime->m_pImpl->m_pAllocator, pv);
}
treedb_list_backend_vtbl_t StringList_backend_vtbl = {
	treedb_list_backend_vtbl_t_entries(StringList_)
};

void STDCALL StringList_Dump
( treedb_list_ptr_t This
) {
	PVOID node = treedb_list_t_PointerHead(This);
	int n;
	for(n = 0; node; node = treedb_list_t_PointerNodeNextPtr(This, node), ++n)
		printf("[%d] %s\n", n,
			(((char *)node) + This->m_pImpl->m_pDescription->m_aligned_node_size)
		);
}
