/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/** \example string-tree.c string tree treedb example
In this example the tree "payload" is a variable length char array - a string.

string-tree.c and string-tree.h provide a backend for a treedb_t so that it can
manage string tree payloads.
*/

#include "string-tree.h"
#include <v3c/meta-avl.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

static PCVOID STDCALL StringTree_UnpackKey
( treedb_tree_backend_ptr_t This
, treedb_tree_runtime_ptr_t runtime
, PCVOID pv
) {
	(void)This;
	const char * p
		= (const char *)pv
			+ runtime->m_pImpl->m_pDescription->m_aligned_node_size;
	return p;
}
static int STDCALL StringTree_Compare
( treedb_tree_backend_ptr_t This
, treedb_tree_runtime_ptr_t runtime
, PCVOID av
, PCVOID bv
) {
	(void)This;
	(void)runtime;
	const char * a = (const char *)av;
	const char * b = (const char *)bv;
	int ret = strcmp(a, b);
#if META_TREEDB_DEBUG
	printf("StringTree_Compare : strcmp(%s,%s) = %d.\n", a, b, ret);
#endif
	return ret;
}
static PVOID STDCALL StringTree_AllocNode
( treedb_tree_backend_ptr_t This
, treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREEKEY pv
) {
	(void)This;
	char * ret = (char *)treedb_allocator_t_PointerAlloc(
		runtime->m_pImpl->m_pAllocator
		, runtime->m_pImpl->m_pDescription->m_aligned_node_size
			+ strlen((const char *)pv) + 1);
	if(!ret)
		return 0;
	char * p = ret + runtime->m_pImpl->m_pDescription->m_aligned_node_size;
#if META_TREEDB_DEBUG
	printf("StringTree_AllocNode : Alloc'd %zu, storing \"%s\" in %zu.\n"
		, treedb_tree_t_PointerToOffset(runtime, ret), (const char *)pv
		, treedb_tree_t_PointerToOffset(runtime, p)
	);
#endif
	if(0) treedb_allocator_t_Check(runtime->m_pImpl->m_pAllocator);
	fflush(stdout);
	strcpy(p, (const char *)pv);
	if(0) treedb_allocator_t_Check(runtime->m_pImpl->m_pAllocator);
	return ret;
}
static void STDCALL StringTree_FreeNode
( treedb_tree_backend_ptr_t This
, treedb_tree_runtime_ptr_t runtime
, PVOID pv
) {
	(void)This;
	treedb_allocator_t_FreePtr(runtime->m_pImpl->m_pAllocator, pv);
}
treedb_tree_backend_vtbl_t StringTree_backend_vtbl = {
	treedb_tree_backend_vtbl_t_entries(StringTree_)
};

typedef struct StringTree_tree_data_renderer_t StringTree_tree_data_renderer_t
, * StringTree_tree_data_renderer_ptr_t;
typedef struct StringTree_tree_data_renderer_vtbl_t
	StringTree_tree_data_renderer_vtbl_t;
typedef const StringTree_tree_data_renderer_vtbl_t
	* StringTree_tree_data_renderer_const_vtbl_ptr_t;

struct StringTree_tree_data_renderer_vtbl_t
{
	treedb_tree_data_renderer_t_methods(,,StringTree_tree_data_renderer_t);
};
struct StringTree_tree_data_renderer_t
{
	treedb_tree_data_renderer_t_members(
		StringTree_tree_data_renderer_const_vtbl_ptr_t);
	uint16_t m_aligned_node_size;
};
static PSTR STDCALL StringTree_tree_data_Dump
( StringTree_tree_data_renderer_ptr_t This
, treedb_tree_runtime_ptr_t runtime
, PCVOID pv
) {
	(void)runtime;
	const char * p = ((const char *)pv) + This->m_aligned_node_size;
#if META_TREEDB_DEBUG
	printf("StringTree_data_Dump [%zu] : Reading \"%s\" from %zu.\n"
		, treedb_tree_impl_t_PointerToOffset(runtime->m_pImpl, pv)
		, (const char *)p
		, treedb_tree_impl_t_PointerToOffset(runtime->m_pImpl, p)
	);
#endif
	char * ret = 0;
	if(0 > asprintf(& ret, "%s", p))
		return 0;
	return ret;
}
static StringTree_tree_data_renderer_vtbl_t StringTree_tree_data_renderer_vtbl =
{ treedb_tree_data_renderer_vtbl_t_entries(StringTree_tree_data_) };

void STDCALL StringTree_Dump
( treedb_tree_ptr_t This
) {
	fflush(stdout);
	StringTree_tree_data_renderer_t StringTree_tree_data_renderer =
	{ & StringTree_tree_data_renderer_vtbl
	, This->m_pImpl->m_pDescription->m_aligned_node_size
	};
	treedb_fd_tree_renderer_t renderer
		= INIT_TREEDB_FD_TREE_RENDERER(STDOUT_FILENO);
	treedb_allocator_t_Dump(This->m_pImpl->m_pAllocator
		, (treedb_tree_renderer_ptr_t)& renderer
		, (treedb_tree_data_renderer_ptr_t)& StringTree_tree_data_renderer);
}
