/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdio.h>
#include <iostream>
#include "meta-treedb-handle-array.hpp"

#if HAVE_64_BITS
uint16_t bpo_end = 16;
#else
uint16_t bpo_end = 8;
#endif

using namespace std;
using namespace v3c;
using namespace treedb;
using namespace meta;

char buf[4096 + 8];
char * g_buf() { return V3C_ROUND_UP_PTR(buf,8); }

const char * flags_desc(uint16_t flags)
{
	switch(flags) {
	case eTreedbContainerHasCount:
		return "count";
	case eTreedbContainerHasDepth:
		return "depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth:
		return "count | depth";
	case eTreedbContainerHasCount | eTreedbContainerIsPointer:
		return "pointer | count";
	case eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | count | depth";
	default:
		return "<unknown flags>";
	}
}
const char * xflags_desc(unsigned int flags)
{
	switch(flags) {
	case 0:
		return "";
	case eTreedbContainerIsPointer:
		return "[pointer]";
	default:
		return "[unknown flags]";
	}
}
struct IntArrayBackendTypes {
	typedef int32_t DataType;
	typedef int32_t InputType;
};
struct IntArrayBackend
: public HandleArrayBackendT< HandleArrayRuntime, IntArrayBackendTypes
	, sizeof(int32_t), sizeof(int32_t) >
{
	inline static int32_t unpack_data(PCVOID pv) { return *(const int32_t*)pv; }
	inline static int32_t & unpack_data(PVOID pv) { return *(int32_t*)pv; }
	inline static HTREEDB_ARRAYDATA pack_input(const int32_t & i)
	{ return (HTREEDB_ARRAYDATA)(& i); }
};
/*! A treedb_handle_array_t delegate.
*/
typedef HandleArrayImplT< IntArrayBackend > IntArrayImpl;
typedef HandleArrayWithImplT< IntArrayBackend > IntArrayWithImpl;

int test(uint16_t abytes, uint16_t aflags, uint16_t xbytes, uint16_t xflags)
{
	cout << "############################################################"
		"####################\n";
	if(aflags & eTreedbContainerIsPointer)
		cout << "# "
#ifdef __cplusplus
		"C++ "
#endif
		"Int handle array : allocator [" << flags_desc(aflags) << "], ";
	else
		cout << "# "
#ifdef __cplusplus
		"C++ "
#endif
		"Int handle array : allocator [" << abytes*8 << " : "
		<< flags_desc(aflags) << "], ";
	if(xflags & eTreedbContainerIsPointer)
		cout << "array" << xflags_desc(xflags) << ".\n";
	else
		cout << "array[" << xbytes*8 << " : " << xflags_desc(xflags) << "].\n";
	cout << "############################################################"
		"####################" << endl;

	/* The storage. */
	StorageOnMemory storage(g_buf(), 4096);

	/* The allocator. */
	AllocatorOnStorageImpl allocator_impl;
	AllocatorOnStorage allocator;
	allocator.m_inst.m_pImpl = & allocator_impl.m_inst;

	/* Even though only the size member is needed from the header, I'm including
	it to compare the numbers with meta-comet. */
	if(!allocator.init_runtime(& allocator_impl.c_type(), abytes, aflags
			, sizeof(TreedbHeader), & storage.base_instance())) {
		cerr << "No matching allocator template found." << endl;
		return 1;
	}
	treedb_const_handle_array_ptr_t parray_template
		= treedb_get_handle_array_template_for(xbytes, xflags);
	if(!parray_template) {
		cerr << "No matching array template found." << endl;
		return 2;
	}
	PackedPayload< IntArrayImpl > payload;
	uint64_t payload_64;
	// The array payload is an allocator handle storing the handle of the memory
	// allocated for the array.
	if(!allocator.Init
			( parray_template->m_pImpl->m_pBackend->m_pDescription->m_root.m_uSize
			, 4096
			, & payload_64))
		return 3;
	if(payload_64 > (uint64_t)((size_t)-1))
		return 4;
	payload.m_value = (size_t)payload_64;

	cout << "Header: " << sizeof(TreedbHeader)
		<< ", Allocator: " << payload.m_value - sizeof(TreedbHeader)
		<< ", Node: " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_node.m_uSize
		<< ", offset: " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_pointer.m_uSize
		<< ", Item " << allocator.description().m_aligned_item_size
		<< ", Node total " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_node.m_uSize
				+ allocator.description().m_aligned_item_size
		<< "\n"
		"Array root: size "
		<< parray_template->m_pImpl->m_pBackend->m_pDescription->m_root.m_uSize
		<< ", align "
		<< parray_template->m_pImpl->m_pBackend->m_pDescription->m_root.m_uAlign
		<< "\n";

	treedb_allocator_t_check_result
		( & allocator.base_c_type()
		, allocator.Check()
	);
	if(allocator.error()) {
		cerr << "Allocator check failed(0) : " << allocator.error() << "."
			<< endl;
		return 5;
	}
	size_t free_offset, free_size;
	if(!allocator.LargestFree(&free_offset, &free_size)) {
		cerr << "treedb_allocator_LargestFree failed." << endl;
		return 6;
	}
	if(xflags & eTreedbContainerIsPointer)
		payload.m_value = (uintptr_t)allocator.OffsetToPointer(payload.m_value);
	IntArrayWithImpl runtime(allocator, parray_template, payload);
	if(!runtime.Alloc(5, 5)) {
		cerr << "Array creation failed." << endl;
		return 7;
	}
	unsigned int n;
	//unsigned int n, m;

	treedb_allocator_t_check_result
		( & allocator.base_c_type()
		, allocator.Check()
	);
	if(allocator.error()) {
		cerr << "Allocator check failed(1) : " << allocator.error() << "."
			<< endl;
		return 8;
	}
	for(n = 0; n < 10; ++n) {
		/*cout << "[" << n << "] * handle == "
			<< allocator.Read(runtime.handle())
			<< "." << endl;*/
		bool res = runtime.push_back(n);
		if(!res) {
			cerr << "[" << n << "] insertion failed." << endl;
			return 9;
		}
#if META_TREEDB_DEBUG
		if(!allocator.error())
			allocator.Check();
#endif
		if(allocator.error()) {
			cerr << "Allocator check failed(2) : " << allocator.error() << "."
				<< endl;
			return 10;
		}
		/*cout << "   * handle == "
			<< allocator.Read(runtime.handle())
			<< "." << endl;
		int32_t * data = runtime.entries();
		for(m = 0; m < runtime.Count(); ++m)
			cout << "   [" << m << "] " << data[m] << "\n";*/
	}
	fflush(stdout);
	uintptr_t count = runtime.Count();
	cout << "Inserted " << count << " items.\n";

	int32_t * data = runtime.entries();

	for(n = 0; n < runtime.Count(); ++n) {
		//cout << "   [" << n << "] " << data[n] << "\n";
		if(data[n] != (int)n) {
			cerr << data[n] << " != " << n << "." << endl;
			return 11;
		}
	}

	runtime.Free();
	treedb_allocator_t_check_result
		( & allocator.base_c_type()
		, allocator.Check()
	);
	if(allocator.error()) {
		cerr << "Allocator check failed(3) : " << allocator.error() << ".\n";
		return 12;
	}

	size_t new_free_offset, new_free_size;
	if(!allocator.LargestFree(&new_free_offset, &new_free_size)) {
		cerr << "Error : treedb_allocator_LargestFree(2) failed.\n";
		return 13;
	}
	if(free_size != new_free_size) {
		cerr << "Error : free size " << new_free_size
			<< " != original free size " << free_size << ".\n";
		return 14;
	}
	if(free_offset != new_free_offset) {
		cerr << "Error : free offset " << new_free_offset
			<< " != original free offset " << free_offset << ".\n";
		return 15;
	}
	if(count != 10) {
		cerr << "Error : Expected a count of 10, got " << count << " instead.\n";
		return 16;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	unsigned int aflags[] =
	{ eTreedbContainerHasCount
	, eTreedbContainerHasDepth
	, eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int apflags[] =
	{ eTreedbContainerIsPointer | eTreedbContainerHasCount
	, eTreedbContainerIsPointer | eTreedbContainerHasDepth
	, eTreedbContainerIsPointer | eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int xflags[] =
	{ 0
	};
	unsigned int xpflags[] =
	{ eTreedbContainerIsPointer
	};
	unsigned int abytes, aflags_index;
	unsigned int xbytes, xflags_index;
	/* Work through the allocator byte values */
	for(abytes = 2; abytes < bpo_end; abytes <<= 1) {
		/* Work through the allocator flags */
		for(aflags_index = 0
				; aflags_index < sizeof(aflags)/sizeof(aflags[0])
				; ++aflags_index) {
			/* Work through the list byte values */
			for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
				/* Work through the list flags */
				for(xflags_index = 0
						; xflags_index < sizeof(xflags)/sizeof(xflags[0])
						; ++xflags_index) {
					int ret = test(abytes, aflags[aflags_index], xbytes
						, xflags[xflags_index]);
					if(ret)
						return ret;
				}
			}
			/* Work through the pointer list flags */
			for(xflags_index = 0;
					xflags_index < sizeof(xpflags)/sizeof(xpflags[0])
					; ++xflags_index) {
				int ret = test(abytes, aflags[aflags_index], 0
						, xpflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
	}
	/* Work through the pointer allocator flags */
	for(aflags_index = 0;
			aflags_index < sizeof(apflags)/sizeof(apflags[0])
			; ++aflags_index) {
		/* Work through the list byte values */
		for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
			/* Work through the list flags */
			for(xflags_index = 0
					; xflags_index < sizeof(xflags)/sizeof(xflags[0])
					; ++xflags_index) {
				int ret = test(0, apflags[aflags_index], xbytes
							, xflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
		/* Work through the pointer list flags */
		for(xflags_index = 0;
				xflags_index < sizeof(xpflags)/sizeof(xpflags[0])
				; ++xflags_index) {
			int ret = test(abytes, apflags[aflags_index], xbytes
					, xpflags[xflags_index]);
			if(ret)
				return ret;
		}
	}
	cout << "Success!" << endl;
	return 0;
}
