/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <iostream>
#include <v3c/meta-treedb-list.h> /* We need an implementation from somewhere. */
#include <v3c/meta-treedb.hpp>
#include <v3c/string_utils.hpp>

#if HAVE_64_BITS
uint16_t bpo_end = 16;
#else
uint16_t bpo_end = 8;
#endif

using namespace std;
using namespace v3c;
using namespace treedb;
using namespace meta;

char buf[4096 + 8];
char * g_buf() { return V3C_ROUND_UP_PTR(buf,8); }

string flags_desc(uint16_t flags)
{
	switch(flags) {
	case eTreedbContainerHasCount:
		return "count";
	case eTreedbContainerHasDepth:
		return "depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth:
		return "count | depth";
	case eTreedbContainerHasCount | eTreedbContainerIsPointer:
		return "pointer | count";
	case eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | count | depth";
	default:
		return "<unknown flags>";
	}
}
string xflags_desc(unsigned int flags)
{
	switch(flags) {
	case 0:
		return "";
	case eTreedbContainerHasCount:
		return "count";
	case eTreedbContainerIsPointer:
		return "pointer";
	case eTreedbContainerHasCount | eTreedbContainerIsPointer:
		return "pointer | count";
	default:
		return string_inline_printf("<unknown flags>[%u]", flags);
	}
}
struct IntListBackendTypes {
	typedef int32_t InputType;
	typedef int32_t DataType;
};
struct IntListBackend
: public FixedListBackendT< IntListBackend, ListRuntime
	, IntListBackendTypes, sizeof(int32_t) >
{
	inline static int32_t unpack_data(PCVOID pv) { return *(const int32_t*)pv; }
	inline static int32_t & unpack_data(PVOID pv) { return *(int32_t*)pv; }
	inline static int32_t unpack_input(PCVOID pv) { return *(int32_t*)pv; }
	inline static HTREEDB_LISTDATA pack_input(const int32_t & i)
	{ return (HTREEDB_LISTDATA)(&i); }
};
/*! A treedb_tree_t delegate.
*/
typedef ListImplT< IntListBackend > IntListImpl;
typedef ListWithImplT< IntListBackend > IntListWithImpl;

int test(uint16_t abytes, uint16_t aflags, uint16_t xbytes, uint16_t xflags)
{
	cout << "############################################################"
		"####################\n";
	if(aflags & eTreedbContainerIsPointer)
		cout << "# "
#ifdef __cplusplus
		"C++ "
#endif
		"Int list : allocator [" << flags_desc(aflags) << "], ";
	else
		cout << "# "
#ifdef __cplusplus
		"C++ "
#endif
		"Int list : allocator [" << abytes*8 << " : " << flags_desc(aflags)
		<< "], ";
	if(xflags & eTreedbContainerIsPointer)
		cout << "list[" << xflags_desc(xflags) << "].\n";
	else
		cout << "list[" << xbytes*8 << " : " << xflags_desc(xflags) << "].\n";
	cout << "############################################################"
		"####################" << endl;

	/* The storage. */
	StorageOnMemory storage(g_buf(), 4096);

	/* The allocator. */
	AllocatorOnStorageImpl allocator_impl;
	AllocatorOnStorage allocator;
	allocator.m_inst.m_pImpl = & allocator_impl.m_inst;

	/* Even though only the size member is needed from the header, I'm including
	it to compare the numbers with meta-comet. */
	if(!allocator.init_runtime(& allocator_impl.c_type(), abytes, aflags
			, sizeof(TreedbHeader), & storage.base_instance())) {
		cerr << "No matching allocator template found." << endl;
		return 1;
	}
	treedb_const_list_ptr_t plist_template
		= treedb_get_list_template_for(xbytes, xflags);
	if(!plist_template) {
		cerr << "No matching list template found." << endl;
		return 2;
	}
	PackedPayload< IntListImpl > payload;
	uint64_t payload_64;

	if(!allocator.Init
			( plist_template->m_pImpl->m_pDescription->m_root.m_uSize
			, 4096
			, & payload_64))
		return 3;
	if(payload_64 > (uint64_t)((size_t)-1))
		return 4;
	payload.m_value = (size_t)payload_64;

	cout << "Header: " << sizeof(TreedbHeader)
		<< ", Allocator: " << payload.m_value - sizeof(TreedbHeader)
		<< ", Node: " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_node.m_uSize
		<< ", offset: " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_pointer.m_uSize
		<< ", Item " << allocator.description().m_aligned_item_size
		<< ", Node total " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_node.m_uSize
				+ allocator.description().m_aligned_item_size
		<< "\n";

	size_t free_offset, free_size;
	if(!allocator.LargestFree(&free_offset, &free_size)) {
		cerr << "treedb_allocator_LargestFree failed." << endl;
		return 4;
	}

	if(xflags & eTreedbContainerIsPointer)
		payload.m_value = (uintptr_t)allocator.OffsetToPointer(payload.m_value);
	IntListWithImpl runtime(allocator, plist_template, payload);
	runtime.Init();
	unsigned int n;

	treedb_allocator_t_check_result
		( & allocator.base_c_type()
		, allocator.Check()
	);
	if(allocator.error()) {
		cerr << "Allocator check failed : " << allocator.error() << "." << endl;
		return 5;
	}
	for(n = 0; n < 10; ++n) {
		cout << "[" << n << "] inserting " << n << "." << endl;
		bool res = runtime.push_back(n);
		if(!res) {
			cerr << "[" << n << "] insertion failed." << endl;
			return 6;
		}
#if META_TREEDB_DEBUG
		if(!allocator.error())
			allocator.Check();
#endif
		if(allocator.error()) {
			cerr << "Allocator check failed : " << allocator.error() << "."
				<< endl;
			return 7;
		}
	}
	uintptr_t count = runtime.Count();
	cout << "Inserted " << count << " items.\n";

	PVOID node = runtime.PointerHead();

	for(n = 0; node; ++n, node = runtime.PointerNodeNextPtr(node)) {
		int d = runtime.data(node);
		cout << "[" << n << "] " << d << "\n";
		if(d != (int)n) {
			cerr << "[" << d << "] != " << n << "." << endl;
			return 8;
		}
	}

	runtime.Clear();
	treedb_allocator_t_check_result
		( & allocator.base_c_type()
		, allocator.Check()
	);
	if(allocator.error()) {
		cerr << "Allocator check failed : " << allocator.error() << ".\n";
		return 9;
	}

	size_t new_free_offset, new_free_size;
	if(!allocator.LargestFree(&new_free_offset, &new_free_size)) {
		cerr << "Error : treedb_allocator_LargestFree(2) failed.\n";
		return 10;
	}
	if(free_size != new_free_size) {
		cerr << "Error : free size " << new_free_size
			<< " != original free size " << free_size << ".\n";
		return 11;
	}
	if(free_offset != new_free_offset) {
		cerr << "Error : free offset " << new_free_offset
			<< " != original free offset " << free_offset << ".\n";
		return 12;
	}
	if(count != 10) {
		cerr << "Error : Expected a count of 10, got " << count << " instead.\n";
		return 13;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	unsigned int aflags[] =
	{ eTreedbContainerHasCount
	, eTreedbContainerHasDepth
	, eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int apflags[] =
	{ eTreedbContainerIsPointer | eTreedbContainerHasCount
	, eTreedbContainerIsPointer | eTreedbContainerHasDepth
	, eTreedbContainerIsPointer | eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int xflags[] =
	{ 0
	, eTreedbContainerHasCount
	};
	unsigned int xpflags[] =
	{ eTreedbContainerIsPointer
	, eTreedbContainerIsPointer | eTreedbContainerHasCount
	};
	unsigned int abytes, aflags_index;
	unsigned int xbytes, xflags_index;
	/* Work through the allocator byte values */
	for(abytes = 2; abytes < bpo_end; abytes <<= 1) {
		/* Work through the allocator flags */
		for(aflags_index = 0
				; aflags_index < sizeof(aflags)/sizeof(aflags[0])
				; ++aflags_index) {
			/* Work through the list byte values */
			for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
				/* Work through the list flags */
				for(xflags_index = 0
						; xflags_index < sizeof(xflags)/sizeof(xflags[0])
						; ++xflags_index) {
					int ret = test(abytes, aflags[aflags_index], xbytes
						, xflags[xflags_index]);
					if(ret)
						return ret;
				}
			}
			/* Work through the pointer list flags */
			for(xflags_index = 0;
					xflags_index < sizeof(xpflags)/sizeof(xpflags[0])
					; ++xflags_index) {
				int ret = test(abytes, aflags[aflags_index], 0
						, xpflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
	}
	/* Work through the pointer allocator flags */
	for(aflags_index = 0;
			aflags_index < sizeof(apflags)/sizeof(apflags[0])
			; ++aflags_index) {
		/* Work through the list byte values */
		for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
			/* Work through the list flags */
			for(xflags_index = 0
					; xflags_index < sizeof(xflags)/sizeof(xflags[0])
					; ++xflags_index) {
				int ret = test(0, apflags[aflags_index], xbytes
							, xflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
		/* Work through the pointer list flags */
		for(xflags_index = 0;
				xflags_index < sizeof(xpflags)/sizeof(xpflags[0])
				; ++xflags_index) {
			int ret = test(abytes, apflags[aflags_index], xbytes
					, xpflags[xflags_index]);
			if(ret)
				return ret;
		}
	}
	cout << "Success!" << endl;
	return 0;
}
