/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <iostream>
#include <v3c/meta-treedb-list.h> /* We need an implementation from somewhere. */
#include <v3c/meta-treedb.hpp>
#include <cstdlib> /* exit() */

#if HAVE_64_BITS
uint16_t bpo_end = 16;
#else
uint16_t bpo_end = 8;
#endif

using namespace std;
using namespace v3c;
using namespace treedb;
using namespace meta;

const char * names[] =
{ "apple"
, "banana"
, "pear"
, "plum"
, "grapefruit"
, "orange"
, "melon"
};

char buf[TREEDB_PAGE_SIZE]
__attribute__ ((aligned (TREEDB_ALIGN))) = {0};

char * g_buf() { return buf; }

const char * flags_desc(unsigned int flags)
{
	switch(flags) {
	case eTreedbContainerHasCount:
		return "count";
	case eTreedbContainerHasDepth:
		return "depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth:
		return "count | depth";
	case eTreedbContainerHasCount | eTreedbContainerIsPointer:
		return "pointer | count";
	case eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth
			| eTreedbContainerIsPointer:
		return "pointer | count | depth";
	default:
		printf("=== UNKNOWN FLAGS %u ===\n", flags);
		return "<unknown flags>";
	}
}
const char * xflags_desc(uint16_t flags)
{
	switch(flags) {
	case 0:
		return "";
	case eTreedbContainerHasCount:
		return "count";
	case eTreedbContainerIsPointer:
		return "pointer";
	case eTreedbContainerHasCount | eTreedbContainerIsPointer:
		return "pointer | count";
	default:
		return "<unknown flags>";
	}
}

struct StringListBackendTypes {
	typedef const char * InputType;
#if (0) // BUGFIX
	// Causes "returning a reference to a temporary" problems.
	typedef char * DataType;
#else
	typedef char DataType[0];
#endif
};
struct StringListBackend
: public ListBackendT< StringListBackend, ListRuntime
	, StringListBackendTypes, sizeof(char) >
{
	inline static const DataType & unpack_data(PCVOID pv)
	{ return *(const DataType *)pv; }
	inline static DataType & unpack_data(PVOID pv)
	{ return *(DataType *)pv; }
	inline static const char * unpack_input(PCVOID pv)
	{ return (const char *)pv; }
	inline static HTREEDB_LISTDATA pack_input(const char * p)
	{ return (HTREEDB_LISTDATA)p; }

	uintptr_t size(const char * key) { return strlen(key) + 1; }
	void assign(char * to, const char * from)
	{ strcpy(to, from); }
};

/*! A treedb_list_t delegate.
*/
typedef ListImplT< StringListBackend > StringListImpl;
typedef ListWithImplT< StringListBackend > StringListWithImpl;

int test(uint16_t abytes, uint16_t aflags, uint16_t xbytes, uint16_t xflags)
{
	cout << "############################################################"
		"####################\n";
	if(aflags & eTreedbContainerIsPointer)
		cout << "# "
#ifdef __cplusplus
		"C++ "
#endif
		"String list : allocator [" << flags_desc(aflags) << "], ";
	else
		cout << "# "
#ifdef __cplusplus
		"C++ "
#endif
		"String list : allocator [" << abytes*8 << " : " << flags_desc(aflags) << "], ";
	if(xflags & eTreedbContainerIsPointer)
		cout << "list[" << xflags_desc(xflags) << "].\n";
	else
		cout << "list[" << xbytes*8 << " : " << xflags_desc(xflags) << "].\n";
	cout << "############################################################"
		"####################" << endl;

	/* The storage. */
	StorageOnMemory storage(g_buf(), 4096);

	/* The allocator. */
	AllocatorOnStorageImpl allocator_impl;
	AllocatorOnStorage allocator;
	allocator.m_inst.m_pImpl = & allocator_impl.m_inst;

	/* Even though only the size member is needed from the header, I'm including
	it to compare the numbers with meta-comet. */
	if(!allocator.init_runtime(& allocator_impl.c_type(), abytes, aflags
			, sizeof(TreedbHeader), & storage.base_instance())) {
		cerr << "No matching allocator template found." << endl;
		return 1;
	}
	treedb_const_list_ptr_t plist_template
		= treedb_get_list_template_for(xbytes, xflags);
	if(!plist_template) {
		cerr << "No matching list template found." << endl;
		return 2;
	}
	PackedPayload< StringListImpl > payload;
	uint64_t payload_64;

	if(!allocator.Init
			( plist_template->m_pImpl->m_pDescription->m_root.m_uSize
			, 4096
			, & payload_64))
		return 3;
	if(payload_64 > (uint64_t)((size_t)-1))
		return 4;
	payload.m_value = (size_t)payload_64;

	cout << "Header: " << sizeof(TreedbHeader)
		<< ", Allocator: " << payload.m_value - sizeof(TreedbHeader)
		<< ", Node: " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_node.m_uSize
		<< ", offset: " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_pointer.m_uSize
		<< ", Item " << allocator.description().m_aligned_item_size
		<< ", Node total " <<
				allocator.free_tree().m_pImpl->m_pDescription->m_node.m_uSize
				+ allocator.description().m_aligned_item_size
		<< "\n";

	size_t free_offset, free_size;
	if(!allocator.LargestFree(&free_offset, &free_size)) {
		cerr << "treedb_allocator_LargestFree failed." << endl;
		return 5;
	}

	if(xflags & eTreedbContainerIsPointer)
		payload.m_value = (uintptr_t)allocator.OffsetToPointer(payload.m_value);
	StringListWithImpl runtime(allocator, plist_template, payload);
	runtime.Init();
	unsigned int n;

	treedb_allocator_t_check_result
		( & allocator.base_c_type()
		, allocator.Check()
	);
	if(allocator.error()) {
		cerr << "Allocator check failed : " << allocator.error() << "." << endl;
		return 6;
	}
	for(n = 0; n < sizeof(names)/sizeof(names[0]); ++n) {
		bool res = runtime.push_back(names[n]);
		if(!res) {
			cerr << "[" << n << "] insertion failed." << endl;
			return 7;
		}
#if META_TREEDB_DEBUG
		if(!allocator.error())
			allocator.Check();
#endif
		if(allocator.error()) {
			cerr << "Allocator check failed : " << allocator.error() << "."
				<< endl;
			return 8;
		}
	}
	uintptr_t count = runtime.Count();
	cout << "Inserted " << count << " items.\n";

	PVOID node = runtime.PointerHead();

	for(n = 0; node; ++n, node = runtime.PointerNodeNextPtr(node)) {
		// Catch a gcc/clang problem with release builds.
#if (1) // BUGFIX
		// This creates a temporary!
		//const char * p = runtime.data((PCVOID)node);
		const char * p = runtime.data(node);
#else
		// This works and is precisely how runtime.data() is implemented!
		const char * p
			= runtime.backend().unpack_data(runtime.node_to_data(node));
#endif
		// Catch a gcc/clang problem with release builds.
		if(!p)
			p = "<null>";
		cout << "[" << n << "] " << p << "\n";
		if(strcmp(p, names[n])) {
			cerr << "Error : [" << n << "] " << p
				<< " != " << names[n] << "." << endl;
			return 9;
		}
	}

	runtime.Clear();
	treedb_allocator_t_check_result
		( & allocator.base_c_type()
		, allocator.Check()
	);
	if(allocator.error()) {
		cerr << "Allocator check failed : " << allocator.error() << ".\n";
		return 10;
	}

	size_t new_free_offset, new_free_size;
	if(!allocator.LargestFree(&new_free_offset, &new_free_size)) {
		cerr << "Error : treedb_allocator_LargestFree(2) failed.\n";
		return 11;
	}
	if(free_size != new_free_size) {
		cerr << "Error : free size " << new_free_size
			<< " != original free size " << free_size << ".\n";
		return 12;
	}
	if(free_offset != new_free_offset) {
		cerr << "Error : free offset " << new_free_offset
			<< " != original free offset " << free_offset << ".\n";
		return 13;
	}
	if(count != sizeof(names)/sizeof(names[0])) {
		cerr << "Error : Expected a count of "
		<< sizeof(names)/sizeof(names[0])
		<< ", got " << count << " instead.\n";
		return 14;
	}
	return 0;
}
#include <signal.h>
void sigsegv_handler(int signum)
{
	cout << flush;
	cerr << flush;
	fflush(stdout);
	fflush(stderr);
	cerr << "Caught sigsegv - exiting." << endl;
	raise(signum); // Does nothing?
	exit(1);
}
int main(int argc, char *argv[])
{
	signal(SIGSEGV, & sigsegv_handler);
	(void)argc;
	(void)argv;
	unsigned int aflags[] =
	{ eTreedbContainerHasCount
	, eTreedbContainerHasDepth
	, eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int apflags[] =
	{ eTreedbContainerIsPointer | eTreedbContainerHasCount
	, eTreedbContainerIsPointer | eTreedbContainerHasDepth
	, eTreedbContainerIsPointer | eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int xflags[] =
	{ 0
	, eTreedbContainerHasCount
	};
	unsigned int xpflags[] =
	{ eTreedbContainerIsPointer
	, eTreedbContainerIsPointer | eTreedbContainerHasCount
	};
	unsigned int abytes, aflags_index;
	unsigned int xbytes, xflags_index;
	/* Work through the allocator byte values */
	for(abytes = 2; abytes < bpo_end; abytes <<= 1) {
		/* Work through the allocator flags */
		for(aflags_index = 0
				; aflags_index < sizeof(aflags)/sizeof(aflags[0])
				; ++aflags_index) {
			/* Work through the list byte values */
			for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
				/* Work through the list flags */
				for(xflags_index = 0
						; xflags_index < sizeof(xflags)/sizeof(xflags[0])
						; ++xflags_index) {
					int ret = test(abytes, aflags[aflags_index], xbytes
						, xflags[xflags_index]);
					if(ret)
						return ret;
				}
			}
			/* Work through the pointer list flags */
			for(xflags_index = 0;
					xflags_index < sizeof(xpflags)/sizeof(xpflags[0])
					; ++xflags_index) {
				int ret = test(abytes, aflags[aflags_index], 0
						, xpflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
	}
	/* Work through the pointer allocator flags */
	for(aflags_index = 0;
			aflags_index < sizeof(apflags)/sizeof(apflags[0])
			; ++aflags_index) {
		/* Work through the list byte values */
		for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
			/* Work through the list flags */
			for(xflags_index = 0
					; xflags_index < sizeof(xflags)/sizeof(xflags[0])
					; ++xflags_index) {
				int ret = test(0, apflags[aflags_index], xbytes
							, xflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
		/* Work through the pointer list flags */
		for(xflags_index = 0;
				xflags_index < sizeof(xpflags)/sizeof(xpflags[0])
				; ++xflags_index) {
			int ret = test(abytes, apflags[aflags_index], xbytes
					, xpflags[xflags_index]);
			if(ret)
				return ret;
		}
	}
	cout << "Success!\n";
	return 0;
}
