/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdio.h>
#include <stdint.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include <v3c/time.hpp>
#include <signal.h>
#include <sys/wait.h>
#include <stdlib.h> // system

#define COBJMACROS
#include "int-tree.h"
#include <treedb/comet-guids.h>
#include <v3c/meta-avl.h>

typedef struct meta_fork_int_tree_tag {
	treedb_tree_impl_t m_tree_impl;
	treedb_tree_t m_tree;
	treedb_tree_backend_t m_backend;
	TreedbItemDescription m_description;
} meta_fork_int_tree_t;
#ifdef __cplusplus
#define META_FORK_INT_TREE_INIT
#else
#define META_FORK_INT_TREE_INIT \
{ META_TREEDB_TREE_IMPL_INIT \
, META_TREEDB_TREE_INIT \
, META_TREEDB_TREE_BACKEND_INIT \
, META_TREEDB_ITEM_DESCRIPTION_INIT \
}
#endif
typedef struct meta_fork_tag {
	treedb_storage_on_file_t m_storage;
	treedb_allocator_on_storage_impl_t m_allocator_impl;
	treedb_allocator_on_storage_t m_allocator;
	meta_fork_int_tree_t int_tree;
} meta_fork_t, * meta_fork_ptr_t;

#define META_FORK_INIT \
{ TREEDB_STORAGE_ON_FILE_INIT \
, META_TREEDB_ALLOCATOR_ON_STORAGE_IMPL_INIT \
, META_TREEDB_ALLOCATOR_ON_STORAGE_INIT \
, META_FORK_INT_TREE_INIT \
}

#if 0
static void test()
{
	treedb_storage_on_file_t storage
		= TREEDB_STORAGE_ON_FILE_INIT;
	treedb_allocator_on_storage_impl_t allocator_impl
		= META_TREEDB_ALLOCATOR_ON_STORAGE_IMPL_INIT;
	treedb_allocator_on_storage_t allocator
		= META_TREEDB_ALLOCATOR_ON_STORAGE_INIT;
	TreedbItemDescription description = META_TREEDB_ITEM_DESCRIPTION_INIT;
	treedb_tree_impl_t tree_impl = META_TREEDB_TREE_IMPL_INIT;
	treedb_tree_t tree = META_TREEDB_TREE_INIT;
	treedb_tree_backend_t backend = META_TREEDB_TREE_BACKEND_INIT;
}
#endif

#ifdef __cplusplus
#define TREEDB_CXX_CTOR(name, value) name
#else
#define TREEDB_CXX_CTOR(name, value) name = value
#endif

meta_fork_t TREEDB_CXX_CTOR(han, META_FORK_INIT);
meta_fork_ptr_t runtime = &han;

const char * fork_file = "fork-test.3db";
uintptr_t free_offs, free_size, packed_pointer;

/* Setting this value too high with TREEDB_NBITS==16 runs out of memory (64K),
making this a good test for graceful insertion/malloc failure. */
int fork_length = 100;
int process_id = 0, treedb_fd;
uint16_t bpo;
pid_t pid;

int create_treedb();
int process_treedb();

FileLockState second_byte = FILE_LOCK_INIT;

#if HAVE_64_BITS
uint16_t bpo_end = 16;
#else
uint16_t bpo_end = 8;
#endif

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	int res = 0;
	uint16_t flags = eTreedbContainerHasCount | eTreedbContainerHasDepth;
	treedb_tree_backend_t_Setup
		( & runtime->int_tree.m_backend
		, & IntTree_backend_vtbl
		);
	for(bpo = 2; bpo < bpo_end; bpo = bpo * 2) {
		printf("=========================================================\n");
		printf("Bytes per offset : %d.\n", bpo);
		printf("=========================================================\n");
		printf("Root : 1. Create fork-test.3db.\n");
		if(!treedb_init_allocator_on_storage_from_template
				( & runtime->m_allocator
				, & runtime->m_allocator_impl
				, bpo
				, flags
				, sizeof(TreedbHeader)
				, (treedb_storage_ptr_t)(& runtime->m_storage))
		) {
			printf("treedb_init_allocator_on_storage_from_template() "
				"failed.\n");
			return 1;
		}
		treedb_const_tree_ptr_t ptree_template
			= treedb_get_avl_tree_template_for(bpo, flags);
		if(!ptree_template) {
			printf("No matching tree template found.\n");
			return 3;
		}
		treedb_tree_impl_t_Setup2
			( & runtime->int_tree.m_tree_impl
			, & runtime->m_allocator
			, & runtime->int_tree.m_backend
			, & runtime->int_tree.m_description
			, ptree_template->m_pImpl
			);

		/* We need to align the int32_t on a 4-byte boundary, like the compiler
			does, as the backend expects it.
		*/
		uint16_t nsz
			= V3C_ROUND_UP_T2(uint16_t
				, runtime->int_tree.m_description.m_node.m_uSize
				, sizeof(int32_t)
				);
		printf("Changing aligned_node_size from %zu to %u.\n"
			, (size_t)runtime->int_tree.m_description.m_aligned_node_size
			, nsz);
		runtime->int_tree.m_description.m_aligned_node_size = nsz;

		res = create_treedb();
		if(res)
			return res;

		printf("Root : 2. write-lock the second byte of fork-test.3db.\n");
		treedb_fd = open(fork_file, O_RDWR);
		if(treedb_fd == -1) {
			printf("Root : 2. couldn't open fork-test.3db?\n");
			unlink(fork_file);
			return 1;
		}
		res = file_lock_state_init(treedb_fd, & second_byte, 2, 1
			, eFileLockWriteLocked);
		if(res) {
			printf("Root : 2. couldn't write-lock fork-test.3db?\n");
			close(treedb_fd);
			unlink(fork_file);
			return 1;
		}
		res = process_treedb();
		if(process_id)
			return res;
		close(treedb_fd); /* Loses all locks. */
#if META_TREEDB_DEBUG
		if(!res) {
			printf("bpo %d : Root : 13. Deleting the treedb.\n", bpo);
			unlink(fork_file);
			int n;
			char filename[30];
			for(n = 1; n < fork_length; ++n) {
				sprintf(filename, "fork_test-%.2d.txt", n);
				if(n == 1) {
					printf("bpo %d : Root : 13. Dumping output of sub-process 1"
						".\n\n\n", bpo);
					fflush(stdout);
					char cmd[40];
					sprintf(cmd, "cat %s", filename);
					system(cmd);
					printf("\n\n\n");
					fflush(stdout);
				}
				unlink(filename);
			}
		}else{
			printf("Root : 13. (fail) leaving treedb and trace files.\n");
			return res;
		}
#else
		printf("Root : 13. Deleting the treedb.\n");
		unlink(fork_file);
#endif
	}
	fflush(stdout);
	return res;
}
#define WRITE_LOCK_TREEDB(step) \
	printf("bpo %d : %d : " #step ". Write-locking.\n", bpo, process_id); \
	ret = treedb_storage_t_WriteLock(& runtime->m_storage); \
	if(ret != eTreedbHeapSuccess) { \
		printf("bpo %d : %d : " #step ". treedb_storage_t_WriteLock()"\
			" failed : %s.\n", bpo, process_id, TreedbHeapStatusDesc(ret)); \
		CLOSE_TREEDB; \
		return 1; \
	}

#define READ_LOCK_TREEDB(step) \
	printf("bpo %d : %d : " #step ". Read-locking.\n", bpo, process_id); \
	ret = treedb_storage_t_ReadLock(& runtime->m_storage); \
	if(ret != eTreedbHeapSuccess) { \
		printf("bpo %d : %d : " #step ". treedb_storage_t_ReadLock()"\
			" failed : %s.\n", bpo, process_id, TreedbHeapStatusDesc(ret)); \
		CLOSE_TREEDB; \
		return 1; \
	}
#define UNLOCK_TREEDB(step) \
	printf("bpo %d : %d : " #step ". Unlocking.\n", bpo, process_id); \
	treedb_storage_t_Unlock(& runtime->m_storage)
#define CLOSE_TREEDB \
	treedb_storage_t_Close(& runtime->m_storage)

void wait_child_exit(int num, int fail)
{
	if(pid) {
		printf("bpo %d : %d : %d. %sWaiting for child to exit.\n"
			, bpo, process_id, num, fail ? "(fail) " : "");
		waitpid(pid, 0, 0);
		printf("bpo %d : %d : %d. %sChild exited.\n"
			, bpo, process_id, num, fail ? "(fail) " : "");
	}
}
int process_treedb()
{
	printf("bpo %d : %d : 3. Spawn child process before opening the "
		"treedb.\n", bpo, process_id);
	TreedbHeapStatus ret;
	while(process_id < (fork_length-1)) {
		fflush(stdout);
		pid = fork();
		if(pid == (pid_t)-1) {
			printf("bpo %d : %d: 3. fork() failed.\n", bpo, process_id);
			return 3;
		}else if(!pid) {
			/* The child. */
			close(treedb_fd);
			++process_id;
#if META_TREEDB_DEBUG
			char filename[20];
			sprintf(filename, "fork_test-%.2d.txt", process_id);
			int file = TEMP_FAILURE_RETRY (open(filename
				, O_RDWR | O_CREAT | O_TRUNC, 0644));
			dup2(file, STDOUT_FILENO);
			(void)TEMP_FAILURE_RETRY(close(file));
#endif
			printf("bpo %d : %d : 3. Child process started.\n"
				, bpo, process_id);
		}else{
			/* The parent continues. */
			break;
		}
	}
	TreedbHeapStatus res = treedb_storage_t_Open
		( & runtime->m_storage
		, UUID_treedb_int_tree_schema
		, fork_file
		, eHeapOpenWrite
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0
		);
	if(res != eTreedbHeapSuccess) {
		printf("bpo %d : %d: 3. treedb_storage_t_Open() failed : %s.\n"
			, bpo, process_id, TreedbHeapStatusDesc(res));
		if(process_id < (fork_length-1))
			wait_child_exit(2, 1);
		return 3;
	}
	UNLOCK_TREEDB(3);

	if(process_id) {
		printf("bpo %d : %d : 4. read-locking the second byte.\n"
			, bpo, process_id);
		second_byte.m_fd = runtime->m_storage.m_fls.m_fd;
		if(file_lock_state_update(& second_byte, eFileLockReadLocked)) {
			printf("bpo %d : %d : 4. read-locking the second byte failed.\n"
				, bpo, process_id);
			CLOSE_TREEDB;
			wait_child_exit(4, 1);
			return 4;
		}
	}else{
		printf("bpo %d : Root : 5. Unlocking the second byte.\n", bpo);
		centi_sleep(40);
		if(file_lock_state_update(& second_byte, eFileLockUnlocked)) {
			printf("bpo %d : Root : 5. Unlocking second byte failed.\n", bpo);
			CLOSE_TREEDB;
			wait_child_exit(5, 1);
			return 5;
		}
	}
	if(process_id) {
		WRITE_LOCK_TREEDB(6);
		printf("bpo %d : %d : 6. Inserting %d.\n", bpo, process_id, process_id);
		TreedbInsertResult ret
			= IntTree_Insert(& runtime->int_tree.m_tree, process_id);
		if(ret != eTreedbInsertSuccess) {
			printf("bpo %d : %d: 6. IntTree_Insert() failed : "
				"%s.\n", bpo, process_id, TreedbInsertResultDesc(ret));
			UNLOCK_TREEDB(6);
			CLOSE_TREEDB;
			wait_child_exit(6, 1);
			return 6;
		}
		UNLOCK_TREEDB(6);
	}
	if(process_id < (fork_length-1)) {
		printf("bpo %d : %d : 7. Waiting for %d to be inserted.\n"
			, bpo, process_id, process_id + 1);
		/* Macro madness. */
		int to_remove = process_id + 1;
		for(;;) {
			READ_LOCK_TREEDB(7);
			int res
				= IntTree_Contains(& runtime->int_tree.m_tree, to_remove);
			UNLOCK_TREEDB(7);
			if(res) {
				printf("bpo %d : %d : 7. %d was inserted.\n"
					, bpo, process_id, to_remove);
				break;
			}
			int wres = waitpid(pid, 0, WNOHANG);
			if((wres == -1) || (wres == pid)) {
				printf("bpo %d : %d : 7. waitpid(%d) returned %d.\n"
					, bpo, process_id, pid, wres);
				break;
			}
			/* It seems that omitting the following line leaves too little time
			for other processes to lock the treedb, causing this process to
			hog the treedb. */
			centi_sleep(1);
		}
		WRITE_LOCK_TREEDB(8);
		int res
			= IntTree_Remove(& runtime->int_tree.m_tree, to_remove);
		uintptr_t count = treedb_tree_t_Count(& runtime->int_tree.m_tree);
		if(!res) {
			printf("bpo %d : %d : 8. IntTree_Remove(%d) failed(%zu).\n"
				, bpo, process_id, to_remove, (size_t)count);
			if(process_id)
				treedb_allocator_t_Free(& runtime->m_allocator, packed_pointer);

			UNLOCK_TREEDB(8);
			CLOSE_TREEDB;
			wait_child_exit(8, 1);
			return 8;
		}
		UNLOCK_TREEDB(8);
		printf("bpo %d : %d : 8. Removed %d(%zu).\n"
			, bpo, process_id, process_id + 1, (size_t)count);
	}
	if(process_id) {
		WRITE_LOCK_TREEDB(9);
		printf("bpo %d : %d : 9. Allocating chunk.\n", bpo, process_id);
		packed_pointer = treedb_allocator_t_Alloc(& runtime->m_allocator
			, bpo == 2 ? 1000 : 8000);
		if(!packed_pointer) {
			printf("bpo %d : %d: 9. treedb_allocator_t_Alloc() failed."
				"\n", bpo, process_id);
			UNLOCK_TREEDB(9);
			CLOSE_TREEDB;
			if(process_id < (fork_length-1))
				wait_child_exit(9, 1);
			return 9;
		}
		UNLOCK_TREEDB(9);
	}
	if(process_id < (fork_length-1)) {
		wait_child_exit(10, 0);
	}
	if(process_id) {
		WRITE_LOCK_TREEDB(11);
		printf("bpo %d : %d : 11. Free'ing chunk.\n", bpo, process_id);
		TreedbFreeResult ret
			= treedb_allocator_t_Free(& runtime->m_allocator, packed_pointer);
		UNLOCK_TREEDB(11);
		if(ret != eTreedbFreeSuccess) {
			printf("bpo %d : %d : 11. TREEDB_NS(free)() failed : %s.\n"
				, bpo, process_id, TreedbFreeResultDesc(ret));
			CLOSE_TREEDB;
			return 11;
		}
	}
	int retval = 0;
	if(!process_id) {
		READ_LOCK_TREEDB(12);
		printf("bpo %d : Root : 12. Checking size.\n", bpo);
		retval = treedb_allocator_t_Check(& runtime->m_allocator);
		uintptr_t free_offs2, free_size2;
		if(!treedb_allocator_t_LargestFree(& runtime->m_allocator, & free_offs2
				, & free_size2)) {
			printf("bpo %d : Root : 12. treedb_allocator_t_LargestFree() "
				"failed.\n", bpo);
			UNLOCK_TREEDB(12);
			CLOSE_TREEDB;
			return 12;
		}
		if((free_offs != free_offs2) || (free_size != free_size2)) {
			printf("bpo %d : Root : 9. free size mismatch : free offset(%zu) "
				", expected %zu, free_size(%zu), expected %zu.\n", bpo
				, (size_t)free_offs2, (size_t)free_offs, (size_t)free_size2
				, (size_t)free_size
			);
			UNLOCK_TREEDB(12);
			CLOSE_TREEDB;
			return 12;
		}
		UNLOCK_TREEDB(12);
	}
	CLOSE_TREEDB;
	return retval;
}
int create_treedb()
{
	TreedbHeapStatus res = treedb_storage_t_Create
		( & runtime->m_storage
		, & runtime->m_allocator
		, UUID_treedb_int_tree_schema
		, fork_file
		, eHeapOpenWrite
		, eHeapCreateOwnerRead | eHeapCreateOwnerWrite | eHeapCreateGroupRead
			| eHeapCreateOtherRead
		, eHeapAccessRead | eHeapAccessWrite | eHeapAccessShared
		, 0, TREEDB_PAGE_SIZE
		, 3 * runtime->m_allocator.m_pImpl->m_pDescription->m_pointer.m_uSize
		);
	if(res != eTreedbHeapSuccess) {
		printf("treedb_storage_t_Create() failed : %s.\n"
			, TreedbHeapStatusDesc(res));
		return 2;
	}
	treedb_tree_t_Setup
		( & runtime->int_tree.m_tree
		, & runtime->int_tree.m_tree_impl
		, runtime->m_storage.m_pTreedb->m_ulPayloadOffset
		);
	treedb_tree_t_Init(& runtime->int_tree.m_tree);
	if(!treedb_allocator_t_LargestFree(& runtime->m_allocator, & free_offs
			, & free_size)) {
		printf("treedb_allocator_t_LargestFree() failed.\n");
		UNLOCK_TREEDB(1);
		CLOSE_TREEDB;
		return 1;
	}
	UNLOCK_TREEDB(1);
	CLOSE_TREEDB;
	return 0;
}
