/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include "int-list.h"
#include <v3c/meta-avl.h>
#include <stdio.h>

#if HAVE_64_BITS
uint16_t bpo_end = 16;
#else
uint16_t bpo_end = 8;
#endif

char buf[4096 + 8];
char * g_buf() { return V3C_ROUND_UP_PTR(buf,8); }

const char * flags_desc(uint16_t flags)
{
	switch(flags) {
	case eTreedbContainerHasCount:
		return "count";
	case eTreedbContainerHasDepth:
		return "depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth:
		return "count | depth";
	case eTreedbContainerHasCount | eTreedbContainerIsPointer:
		return "pointer | count";
	case eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | count | depth";
	default:
		return "<unknown flags>";
	}
}
const char * xflags_desc(unsigned int flags)
{
	switch(flags) {
	case 0:
		return "";
	case eTreedbContainerHasCount:
		return "count";
	case eTreedbContainerIsPointer:
		return "pointer";
	case eTreedbContainerHasCount | eTreedbContainerIsPointer:
		return "pointer | count";
	default:
		return "<unknown flags>";
	}
}
int test(uint16_t abytes, uint16_t aflags, uint16_t xbytes, uint16_t xflags)
{
	printf("############################################################"
		"####################\n");
	if(aflags & eTreedbContainerIsPointer)
		printf("# "
#ifdef __cplusplus
		"CXX "
#endif
		"Int list : allocator [%s], ", flags_desc(aflags));
	else
		printf("# "
#ifdef __cplusplus
		"CXX "
#endif
		"Int list : allocator [%u : %s], ", abytes*8, flags_desc(aflags));
	if(xflags & eTreedbContainerIsPointer)
		printf("list[%s].\n", xflags_desc(xflags));
	else
		printf("list[%u : %s].\n", xbytes*8, xflags_desc(xflags));
	printf("############################################################"
		"####################\n");
	fflush(stdout);

	/* The storage. */
	treedb_storage_on_memory_t storage
		= TREEDB_STORAGE_ON_MEMORY_INIT(g_buf(), 4096);

	/* The allocator. */
	treedb_allocator_on_storage_impl_t allocator_impl;
	treedb_allocator_on_storage_t allocator;
	allocator.m_pImpl = & allocator_impl;
	/* Even though only the size member is needed from the header, I'm including
	it to compare the numbers with meta-comet. */
	if(!treedb_init_allocator_on_storage_from_template
			(& allocator, & allocator_impl, abytes, aflags, sizeof(TreedbHeader)
			, (treedb_storage_ptr_t)(& storage))) {
		printf("No matching allocator template found.\n");
		return 1;
	}
	treedb_const_list_ptr_t plist_template
		= treedb_get_list_template_for(xbytes, xflags);
	if(!plist_template) {
		printf("No matching list template found.\n");
		return 2;
	}

	uint64_t payload;
	if(!treedb_allocator_t_Init(& allocator
			, plist_template->m_pImpl->m_pDescription->m_root.m_uSize
			, 4096
			, & payload)) {
		printf("Allocator Init() failed.\n");
		return 3;
	}
	printf("Header: %u, Allocator: %zu, Node: %u, Item %u, Node total %u\n"
		, (unsigned int)sizeof(TreedbHeader)
		, (size_t)(payload - sizeof(TreedbHeader))
		, allocator.m_free.m_pImpl->m_pDescription->m_node.m_uSize
		, allocator.m_pImpl->m_pDescription->m_aligned_item_size
		, allocator.m_free.m_pImpl->m_pDescription->m_aligned_node_size
			+ allocator.m_pImpl->m_pDescription->m_aligned_item_size
	);

	size_t free_offset, free_size;
	if(!treedb_allocator_t_LargestFree(& allocator, &free_offset, &free_size)) {
		printf("treedb_allocator_t_LargestFree failed.\n");
		return 4;
	}

	/* Set up the int list backend. */
	treedb_list_backend_t backend;
	treedb_list_backend_t_Setup(& backend, & IntList_backend_vtbl);

	/* Set up the int list implementation. */
	TreedbItemDescription description;
	treedb_list_impl_t int_list_impl;
	treedb_list_impl_t_Setup
		( & int_list_impl
		, & allocator
		, & backend
		, & description
		, plist_template->m_pImpl
		);

	/* Set up the string list. */
	if(xflags & eTreedbContainerIsPointer)
		payload = (uintptr_t)treedb_allocator_t_OffsetToPointer(&allocator
			, payload);
	treedb_list_t int_list;
	treedb_list_t_Setup
		( & int_list
		, & int_list_impl
		, payload
		);
	treedb_list_t_Init(& int_list);

	unsigned int n;

	IntList_Dump(& int_list);

	treedb_allocator_t_check_result
		( (treedb_allocator_ptr_t)& allocator
		, treedb_allocator_t_Check(& allocator)
	);
	if(allocator.m_errno) {
		printf("Allocator check failed : %d.\n", allocator.m_errno);
		return 6;
	}
	for(n = 0; n < 10; ++n) {
		printf("[%u] inserting %u\n", n, n);
		fflush(stdout);
		HTREEDB_LISTNODE node = IntList_PushBack(& int_list, n);
		if(!node) {
			printf("[%d] insertion failed.\n", n);
			return 7;
		}
#if META_TREEDB_DEBUG
		if(!allocator.m_errno)
			treedb_allocator_t_Check(& allocator);
#endif
		if(allocator.m_errno) {
			printf("Allocator check failed : %d.\n", allocator.m_errno);
			return 8;
		}
	}
	size_t count = treedb_list_t_Count(& int_list);
	printf("Inserted %zu items.\n", count);

	PVOID node = treedb_list_t_PointerHead(& int_list);

	for(n = 0; node; ++n
			, node = treedb_list_t_PointerNodeNextPtr(& int_list, node)) {
		int d = IntList_Data(& int_list, node);
		printf("[%u] %d\n", n, d);
		if(d != (int)n) {
			printf("[%u] != %d.\n", d, n);
			return 9;
		}
	}

	treedb_list_t_Clear(& int_list);
	treedb_allocator_t_check_result
		( (treedb_allocator_ptr_t)& allocator
		, treedb_allocator_t_Check(& allocator)
	);
	if(allocator.m_errno) {
		printf("Allocator check failed : %d.\n", allocator.m_errno);
		return 10;
	}

	size_t new_free_offset, new_free_size;
	if(!treedb_allocator_t_LargestFree(& allocator, &new_free_offset
			, &new_free_size)) {
		printf("Error : treedb_allocator_t_LargestFree(2) failed.\n");
		return 11;
	}
	if(free_size != new_free_size) {
		printf("Error : free size %zu != original free size %zu.\n"
			, new_free_size
			, free_size
		);
		return 12;
	}
	if(free_offset != new_free_offset) {
		printf("Error : free offset %zu != original free offset %zu.\n"
			, new_free_offset
			, free_offset
		);
		return 13;
	}
	if(count != 10) {
		printf("Error : Expected a count of %u, got %zu instead.\n"
			, 10
			, count
		);
		return 14;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	unsigned int aflags[] =
	{ eTreedbContainerHasCount
	, eTreedbContainerHasDepth
	, eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int apflags[] =
	{ eTreedbContainerIsPointer | eTreedbContainerHasCount
	, eTreedbContainerIsPointer | eTreedbContainerHasDepth
	, eTreedbContainerIsPointer | eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int xflags[] =
	{ 0
	, eTreedbContainerHasCount
	};
	unsigned int xpflags[] =
	{ eTreedbContainerIsPointer
	, eTreedbContainerIsPointer | eTreedbContainerHasCount
	};
	unsigned int abytes, aflags_index;
	unsigned int xbytes, xflags_index;
	/* Work through the allocator byte values */
	for(abytes = 2; abytes < bpo_end; abytes <<= 1) {
		/* Work through the allocator flags */
		for(aflags_index = 0
				; aflags_index < sizeof(aflags)/sizeof(aflags[0])
				; ++aflags_index) {
			/* Work through the list byte values */
			for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
				/* Work through the list flags */
				for(xflags_index = 0
						; xflags_index < sizeof(xflags)/sizeof(xflags[0])
						; ++xflags_index) {
					int ret = test(abytes, aflags[aflags_index], xbytes
						, xflags[xflags_index]);
					if(ret)
						return ret;
				}
			}
			/* Work through the pointer list flags */
			for(xflags_index = 0;
					xflags_index < sizeof(xpflags)/sizeof(xpflags[0])
					; ++xflags_index) {
				int ret = test(abytes, aflags[aflags_index], 0
						, xpflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
	}
	/* Work through the pointer allocator flags */
	for(aflags_index = 0;
			aflags_index < sizeof(apflags)/sizeof(apflags[0])
			; ++aflags_index) {
		/* Work through the list byte values */
		for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
			/* Work through the list flags */
			for(xflags_index = 0
					; xflags_index < sizeof(xflags)/sizeof(xflags[0])
					; ++xflags_index) {
				int ret = test(0, apflags[aflags_index], xbytes
							, xflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
		/* Work through the pointer list flags */
		for(xflags_index = 0;
				xflags_index < sizeof(xpflags)/sizeof(xpflags[0])
				; ++xflags_index) {
			int ret = test(abytes, apflags[aflags_index], xbytes
					, xpflags[xflags_index]);
			if(ret)
				return ret;
		}
	}
	printf("Success!\n");
	return 0;
}
