/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include "int-tree.h"
#include <stdio.h>

#if HAVE_64_BITS
uint16_t bpo_end = 16;
#else
uint16_t bpo_end = 8;
#endif

const char * flags_desc(unsigned int flags)
{
	switch(flags) {
	case eTreedbContainerHasCount:
		return "count";
	case eTreedbContainerHasDepth:
		return "depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth:
		return "count | depth";
	case eTreedbContainerHasCount | eTreedbContainerIsPointer:
		return "pointer | count";
	case eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | count | depth";
	default:
		return "<unknown flags>";
	}
}
int test(uint16_t abytes, uint16_t aflags, uint16_t xbytes, uint16_t xflags)
{
	printf("############################################################"
		"####################\n");
	if(aflags & eTreedbContainerIsPointer)
		printf("# "
#ifdef __cplusplus
		"CXX "
#endif
		"Int tree : allocator [%s], ", flags_desc(aflags));
	else
		printf("# "
#ifdef __cplusplus
		"CXX "
#endif
		"Int tree : allocator [%u : %s], ", abytes*8, flags_desc(aflags));
	if(xflags & eTreedbContainerIsPointer)
		printf("tree[%s].\n", flags_desc(xflags));
	else
		printf("tree[%u : %s].\n", xbytes*8, flags_desc(xflags));
	printf("############################################################"
		"####################\n");
	fflush(stdout);

	int tres = setup_IntTreeData(abytes, aflags, xbytes, xflags);
	if(tres)
		return tres;

	unsigned int n;

	IntTree_Dump(& g_data.int_tree);

	treedb_allocator_t_check_result
		( (treedb_allocator_ptr_t) & g_data.allocator
		, treedb_allocator_t_Check(& g_data.allocator)
	);
	if(g_data.allocator.m_errno) {
		printf("Allocator check failed : %d.\n", g_data.allocator.m_errno);
		return 5;
	}
	for(n = 0; n < 10; ++n) {
		printf("[%d] inserting %d\n", n, n);
		fflush(stdout);
		TreedbInsertResult res = IntTree_Insert(& g_data.int_tree, n);
		if(res != eTreedbInsertSuccess) {
			printf("[%d] insertion failed: %s.\n", n
				, TreedbInsertResultDesc(res));
			return 6;
		}
#if META_TREEDB_DEBUG
		if(!allocator.m_errno)
			treedb_allocator_t_Check(& allocator);
#endif
		if(g_data.allocator.m_errno) {
			printf("Allocator check failed : %d.\n", g_data.allocator.m_errno);
			return 7;
		}
		IntTree_Dump(& g_data.int_tree);
	}
	--n;
	TreedbInsertResult res = IntTree_Insert(& g_data.int_tree, n);
	if(res != eTreedbInsertExists) {
		printf("[%d] insertion should have failed: %s.\n", n
			, TreedbInsertResultDesc(res));
		return 8;
	}
	size_t count = treedb_tree_t_Count(& g_data.int_tree);
	printf("Inserted %zu items.\n", count);

	PVOID node = treedb_tree_t_PointerHead(& g_data.int_tree);

	for(n = 0; node; ++n, node
			= treedb_tree_t_PointerNodeNextPtr(& g_data.int_tree, node)) {
		int d = IntTree_Data(& g_data.int_tree, node);
		printf("[%d] %d\n", n, d);
		if(d != (int)n) {
			printf("[%u] != %d.\n", d, n);
			return 9;
		}
	}

	treedb_tree_t_Clear(& g_data.int_tree);
	treedb_allocator_t_check_result
		( (treedb_allocator_ptr_t) & g_data.allocator
		, treedb_allocator_t_Check(& g_data.allocator)
	);
	if(g_data.allocator.m_errno) {
		printf("Allocator check failed : %d.\n", g_data.allocator.m_errno);
		return 10;
	}

	size_t new_free_offset, new_free_size;
	if(!treedb_allocator_t_LargestFree(& g_data.allocator, &new_free_offset
			, &new_free_size)) {
		printf("Error : treedb_allocator_LargestFree(2) failed.\n");
		return 11;
	}
	if(g_data.free_size != new_free_size) {
		printf("Error : free size %zu != original free size %zu.\n"
			, new_free_size
			, g_data.free_size
		);
		return 12;
	}
	if(g_data.free_offset != new_free_offset) {
		printf("Error : free offset %zu != original free offset %zu.\n"
			, new_free_offset
			, g_data.free_offset
		);
		return 13;
	}
	if(count != 10) {
		printf("Error : Expected a count of %zu, got %zu instead.\n"
			, (size_t)10
			, count
		);
		return 14;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	unsigned int flags[] =
	{ eTreedbContainerHasCount
	, eTreedbContainerHasDepth
	, eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int pflags[] =
	{ eTreedbContainerIsPointer | eTreedbContainerHasCount
	, eTreedbContainerIsPointer | eTreedbContainerHasDepth
	, eTreedbContainerIsPointer | eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int abytes, aflags_index;
	unsigned int xbytes, xflags_index;
	/* Work through the allocator byte values */
	for(abytes = 2; abytes < bpo_end; abytes <<= 1) {
		/* Work through the allocator flags */
		for(aflags_index = 0
				; aflags_index < sizeof(flags)/sizeof(flags[0])
				; ++aflags_index) {
			/* Work through the tree byte values */
			for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
				/* Work through the tree flags */
				for(xflags_index = 0
						; xflags_index < sizeof(flags)/sizeof(flags[0])
						; ++xflags_index) {
					int ret = test(abytes, flags[aflags_index], xbytes
							, flags[xflags_index]);
					if(ret)
						return ret;
				}
			}
			/* Work through the pointer tree flags */
			for(xflags_index = 0;
					xflags_index < sizeof(pflags)/sizeof(pflags[0])
					; ++xflags_index) {
				int ret = test(abytes, flags[aflags_index], 0
						, pflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
	}
	/* Work through the pointer allocator flags */
	for(aflags_index = 0;
			aflags_index < sizeof(pflags)/sizeof(pflags[0])
			; ++aflags_index) {
		/* Work through the tree byte values */
		for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
			/* Work through the tree flags */
			for(xflags_index = 0
					; xflags_index < sizeof(flags)/sizeof(flags[0])
					; ++xflags_index) {
				int ret = test(0, pflags[aflags_index], xbytes
							, flags[xflags_index]);
				if(ret)
					return ret;
			}
		}
		/* Work through the pointer tree flags */
		for(xflags_index = 0;
				xflags_index < sizeof(pflags)/sizeof(pflags[0])
				; ++xflags_index) {
			int ret = test(0, pflags[aflags_index], 0, pflags[xflags_index]);
			if(ret)
				return ret;
		}
	}
	printf("Success!\n");
	return 0;
}
