/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <treedb/treedb-header.h>
#include "string-tree.h"
#include <stdio.h>
#include <string.h>

#if HAVE_64_BITS
uint16_t bpo_end = 16;
#else
uint16_t bpo_end = 8;
#endif

const char * names[] =
{ "apple"
, "banana"
, "pear"
, "plum"
, "grapefruit"
, "orange"
, "melon"
};
const char * sorted_names[] =
{ "apple"
, "banana"
, "grapefruit"
, "melon"
, "orange"
, "pear"
, "plum"
};

char buf[4096 + 8];
char * g_buf() { return V3C_ROUND_UP_PTR(buf,8); }

const char * flags_desc(uint16_t flags)
{
	switch(flags) {
	case eTreedbContainerHasCount:
		return "count";
	case eTreedbContainerHasDepth:
		return "depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth:
		return "count | depth";
	case eTreedbContainerHasCount | eTreedbContainerIsPointer:
		return "pointer | count";
	case eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | depth";
	case eTreedbContainerHasCount | eTreedbContainerHasDepth | eTreedbContainerIsPointer:
		return "pointer | count | depth";
	default:
		printf("=== UNKNOWN FLAGS %u ===\n", flags);
		return "<unknown aflags>";
	}
}
int test(uint16_t abytes, uint16_t aflags, uint16_t xbytes, uint16_t xflags)
{
	printf("############################################################"
		"####################\n");
	if(aflags & eTreedbContainerIsPointer)
		printf("# "
#ifdef __cplusplus
		"CXX "
#endif
		"String tree : allocator [%s], ", flags_desc(aflags));
	else
		printf("# "
#ifdef __cplusplus
		"CXX "
#endif
		"String tree : allocator [%u : %s], ", abytes*8, flags_desc(aflags));
	if(xflags & eTreedbContainerIsPointer)
		printf("tree[%s].\n", flags_desc(xflags));
	else
		printf("tree[%u : %s].\n", xbytes*8, flags_desc(xflags));
	printf("############################################################"
		"####################\n");
	fflush(stdout);

	/* The storage. */
	treedb_storage_on_memory_t storage
		= TREEDB_STORAGE_ON_MEMORY_INIT(g_buf(), 4096);

	/* The allocator. */
	treedb_allocator_on_storage_impl_t allocator_impl;
	treedb_allocator_on_storage_t allocator;
	allocator.m_pImpl = & allocator_impl;
	/* Even though only the size member is needed from the header, I'm including
	it to compare the numbers with meta-comet. */
	if(!treedb_init_allocator_on_storage_from_template
			(& allocator, & allocator_impl, abytes, aflags, sizeof(TreedbHeader)
			, (treedb_storage_ptr_t)(& storage))) {
		printf("treedb_init_allocator_on_storage_from_template failed.\n");
		return 1;
	}
	treedb_const_tree_ptr_t ptree_template
		= treedb_get_avl_tree_template_for(xbytes, xflags);
	if(!ptree_template) {
		printf("No matching tree template found.\n");
		return 2;
	}

	uint64_t payload;
	if(!treedb_allocator_t_Init(& allocator
			, ptree_template->m_pImpl->m_pDescription->m_root.m_uSize
			, 4096
			, & payload)) {
		printf("Allocator Init() failed.\n");
		return 3;
	}
	printf("Header: %u, Allocator: %zu, Node: %u, Item %u, Node total %u\n"
		, (unsigned int)sizeof(TreedbHeader)
		, (size_t)(payload - sizeof(TreedbHeader))
		, allocator.m_free.m_pImpl->m_pDescription->m_node.m_uSize
		, allocator.m_pImpl->m_pDescription->m_aligned_item_size
		, allocator.m_free.m_pImpl->m_pDescription->m_aligned_node_size
			+ allocator.m_pImpl->m_pDescription->m_aligned_item_size
	);

	size_t free_offset, free_size;
	if(!treedb_allocator_t_LargestFree(& allocator, &free_offset, &free_size)) {
		printf("treedb_allocator_LargestFree failed.\n");
		return 4;
	}

	/* Set up the string tree backend. */
	treedb_tree_backend_t backend;
	treedb_tree_backend_t_Setup(& backend, & StringTree_backend_vtbl);

	/* Set up the string tree implementation. */
	TreedbItemDescription description;
	treedb_tree_impl_t string_tree_impl;
	treedb_tree_impl_t_Setup2
		( & string_tree_impl
		, & allocator
		, & backend
		, & description
		, ptree_template->m_pImpl
		);

	/* Set up the string tree. */
	if(xflags & eTreedbContainerIsPointer)
		payload = (uintptr_t)treedb_allocator_t_OffsetToPointer(&allocator
			, payload);
	treedb_tree_t string_tree;
	treedb_tree_t_Setup
		( & string_tree
		, & string_tree_impl
		, payload
		);
	treedb_tree_t_Init(& string_tree);

	unsigned int n;

	StringTree_Dump(& string_tree);

	treedb_allocator_t_check_result
		( (treedb_allocator_ptr_t) & allocator
		, treedb_allocator_t_Check(& allocator)
	);
	if(allocator.m_errno) {
		printf("Allocator check failed : %d.\n", allocator.m_errno);
		return 6;
	}
	for(n = 0; n < sizeof(names)/sizeof(names[0]); ++n) {
		printf("[%d] inserting %s\n", n, names[n]);
		fflush(stdout);
		TreedbInsertResult res = StringTree_Insert(& string_tree, names[n]);
		if(res != eTreedbInsertSuccess) {
			printf("[%d] insertion failed: %s.\n", n
				, TreedbInsertResultDesc(res));
			return 7;
		}
#if META_TREEDB_DEBUG
		if(!allocator.m_errno)
			treedb_allocator_t_Check(& allocator);
#endif
		if(allocator.m_errno) {
			printf("Allocator check failed : %d.\n", allocator.m_errno);
			return 8;
		}
		StringTree_Dump(& string_tree);
	}
	--n;
	TreedbInsertResult res = StringTree_Insert(& string_tree, names[n]);
	if(res != eTreedbInsertExists) {
		printf("[%d] insertion should have failed: %s.\n", n
			, TreedbInsertResultDesc(res));
		return 9;
	}
	size_t count = treedb_tree_t_Count(& string_tree);
	printf("Inserted %zu items.\n", count);

	PVOID node = treedb_tree_t_PointerHead(& string_tree);

	for(n = 0; node; ++n
			, node = treedb_tree_t_PointerNodeNextPtr(& string_tree, node)) {
		const char * p = StringTree_Data(& string_tree, node);
		printf("[%u] %s\n", n, p);
		if(strcmp(p, sorted_names[n])) {
			printf("Error : [%u] %s != %s.\n", n, p, sorted_names[n]);
			return 10;
		}
	}

	treedb_tree_t_Clear(& string_tree);
	treedb_allocator_t_check_result
		( (treedb_allocator_ptr_t) & allocator
		, treedb_allocator_t_Check(& allocator)
	);
	if(allocator.m_errno) {
		printf("Error : Allocator check failed : %d.\n", allocator.m_errno);
		return 11;
	}

	size_t new_free_offset, new_free_size;
	if(!treedb_allocator_t_LargestFree(& allocator, &new_free_offset
			, &new_free_size)) {
		printf("Error : treedb_allocator_LargestFree(2) failed.\n");
		return 12;
	}
	if(free_size != new_free_size) {
		printf("Error : free size %zu != original free size %zu.\n"
			, new_free_size
			, free_size
		);
		return 13;
	}
	if(free_offset != new_free_offset) {
		printf("Error : free offset %zu != original free offset %zu.\n"
			, new_free_offset
			, free_offset
		);
		return 14;
	}
	if(count != sizeof(names)/sizeof(names[0])) {
		printf("Error : Expected a count of %u, got %zu instead.\n"
			, (unsigned int)(sizeof(names)/sizeof(names[0]))
			, count
		);
		return 15;
	}
	return 0;
}

int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	unsigned int flags[] =
	{ eTreedbContainerHasCount
	, eTreedbContainerHasDepth
	, eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int pflags[] =
	{ eTreedbContainerIsPointer | eTreedbContainerHasCount
	, eTreedbContainerIsPointer | eTreedbContainerHasDepth
	, eTreedbContainerIsPointer | eTreedbContainerHasCount | eTreedbContainerHasDepth
	};
	unsigned int abytes, aflags_index;
	unsigned int xbytes, xflags_index;
	/* Work through the allocator byte values */
	for(abytes = 2; abytes < bpo_end; abytes <<= 1) {
		/* Work through the allocator flags */
		for(aflags_index = 0
				; aflags_index < sizeof(flags)/sizeof(flags[0])
				; ++aflags_index) {
			/* Work through the tree byte values */
			for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
				/* Work through the tree flags */
				for(xflags_index = 0
						; xflags_index < sizeof(flags)/sizeof(flags[0])
						; ++xflags_index) {
					int ret = test(abytes, flags[aflags_index], xbytes
							, flags[xflags_index]);
					if(ret)
						return ret;
				}
			}
			/* Work through the pointer tree flags */
			for(xflags_index = 0;
					xflags_index < sizeof(pflags)/sizeof(pflags[0])
					; ++xflags_index) {
				int ret = test(abytes, flags[aflags_index], 0
						, pflags[xflags_index]);
				if(ret)
					return ret;
			}
		}
	}
	/* Work through the pointer allocator flags */
	for(aflags_index = 0;
			aflags_index < sizeof(pflags)/sizeof(pflags[0])
			; ++aflags_index) {
		/* Work through the tree byte values */
		for(xbytes = 2; xbytes < bpo_end; xbytes <<= 1) {
			/* Work through the tree flags */
			for(xflags_index = 0
					; xflags_index < sizeof(flags)/sizeof(flags[0])
					; ++xflags_index) {
				int ret = test(0, pflags[aflags_index], xbytes
							, flags[xflags_index]);
				if(ret)
					return ret;
			}
		}
		/* Work through the pointer tree flags */
		for(xflags_index = 0;
				xflags_index < sizeof(pflags)/sizeof(pflags[0])
				; ++xflags_index) {
			int ret = test(0, pflags[aflags_index], 0, pflags[xflags_index]);
			if(ret)
				return ret;
		}
	}
	printf("Success!\n");
	return 0;
}
