/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/*! @file critter.cpp critter.cpp

See @ref ccxxintegration for details.
*/

#include <stdio.h>
#include "critter.h"
#include <stddef.h> /* for offsetof() */

const char * names[] = { "bat", "wolf" };
const char * sounds[] = { "Eek!", "Arooo!" };

////////////////////////////////////////////////////////////////////////////////
// This part goes into the library.
////////////////////////////////////////////////////////////////////////////////
template< typename Interface = INamed_NoData >
class WolfName : public INamed< Interface >
{
public:
	typedef INamed< Interface > Base;
	WolfName() : Base(1) {}
	virtual ~WolfName() {}
	inline const char * name() { return names[Base::m_index]; }
};
/* Re-use WolfName's implementation of INamed plus implement ICritter.
Providing this as a template allows further interfaces to be implemented that
can re-use Wolf's implementation. */
#define V3C_DELEGATE(Cls, x) Cls::x
#define DGTHIS0(x)
#define DGTHIS(x)
#define DGTHIS_MEM(a,b) Base::m_##b
template< typename Interface = ICritter_NoData >
class Wolf : public WolfName< Interface >
{
public:
	typedef WolfName< Interface > Base;
	Wolf() : Base() {}
	virtual ~Wolf() {}
	//inline const char * name() { return WolfName::name(); }
	//TREEDB_INamed_DELEGATE_TO(Base);
	inline const char * sound() { return sounds[Base::m_index]; }
};
#undef V3C_DELEGATE
#undef DGTHIS0
#undef DGTHIS
#undef DGTHIS_MEM
static const char * bat_name(critter_ptr_t This)
{ (void)This; return names[0]; }
static const char * bat_sound(critter_ptr_t This)
{ (void)This; return sounds[0]; }
extern "C" {
static critter_vtbl_t bat_vtbl = { & bat_name, & bat_sound };
static critter_t bat = { & bat_vtbl, 0 };
static Wolf<> wolf;
const_critter_ptr_t get_critter_template(bool b)
{
	if(b) {
		// Inform the compiler not to use optimizations that assume no aliasing.
		union {
			Wolf<> * w;
			critter_ptr_t c;
		} u;
		u.w = & wolf;
		return u.c;
	}else
		return & bat;
}
const_named_ptr_t get_named_template(bool b)
{
	// Inform the compiler not to use optimizations that assume no aliasing.
	union {
		const_named_ptr_t n;
		const_critter_ptr_t c;
	} u;
	u.c = get_critter_template(b);
	return u.n;
}
} // extern "C" {

////////////////////////////////////////////////////////////////////////////////
// This part goes into the program.
////////////////////////////////////////////////////////////////////////////////
int cxxtest()
{
	Named named;
	for(int n = 0; n < 2; ++n) {
		named.m_inst = * get_named_template(n);
		printf("[%d] name: %s.\n", n, named.name());
		if(named.name() != names[n])
			return 1;
		if(named.m_inst.index != n)
			return 2;
	}
	Critter critter;
	for(int n = 0; n < 2; ++n) {
		critter.m_inst = * get_critter_template(n);
		printf("[%d] %s's sound is %s.\n", n, critter.name(), critter.sound());
		if(critter.name() != names[n])
			return 3;
		if(critter.sound() != sounds[n])
			return 4;
		if(critter.m_inst.index != n)
			return 5;
	}
	return 0;
}
int main(int argc, char *argv[])
{
	(void)argc;
	(void)argv;
	return cxxtest();
}
