/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/*! @file critter.h critter.h

See @ref ccxxintegration for details.
*/
extern "C" {
typedef struct named_t named_t, * named_ptr_t;
typedef const named_t * const_named_ptr_t;
typedef struct named_vtbl_t named_vtbl_t;
typedef const named_vtbl_t * const_named_vtbl_ptr_t;

#define named_vtbl_t_entries(Class) \
	const char * (* name)(Class * /*This*/)
struct named_vtbl_t {
	named_vtbl_t_entries(named_t);
};
#define named_t_members \
	int index
struct named_t {
	const_named_vtbl_ptr_t lpVtbl;
	named_t_members;
};

typedef struct critter_t critter_t, * critter_ptr_t;
typedef critter_t const * const_critter_ptr_t;
typedef struct critter_vtbl_t critter_vtbl_t;
typedef const critter_vtbl_t * const_critter_vtbl_ptr_t;

#define critter_vtbl_t_entries(Class) \
	named_vtbl_t_entries(Class); \
	const char * (* sound)(Class * /*This*/)

struct critter_vtbl_t {
	critter_vtbl_t_entries(critter_t);
};
struct critter_t {
	const_critter_vtbl_ptr_t lpVtbl;
	named_t_members;
};

extern const_named_ptr_t get_named_template(bool b);
extern const_critter_ptr_t get_critter_template(bool b);
} // extern "C" {

class INamed_NoData
{
public:
	virtual const char * name() = 0;
protected:
	~INamed_NoData() {}
};
#define INamed_members \
	int m_index
template< typename Base = INamed_NoData >
class INamed : public Base
{
public:
	virtual const char * name() = 0;
/*	virtual int getNumber() { return m_index; }
	virtual int setNumber(int n) { m_index = n; return m_index; }
*/protected:
	INamed(int n) : m_index(n) {}
public:
	INamed_members;
};
#define TREEDB_INamed_NoData_DELEGATE_TO(x) \
	inline const char * name() \
		{ return V3C_DELEGATE(x, name(DGTHIS0(x))); }

#define TREEDB_INamed_DELEGATE_TO(x) \
	inline const char * name() \
		{ return V3C_DELEGATE(x, name(DGTHIS0(x))); } \
	inline int getNumber() { return DGTHIS_MEM(x,index); } \
	inline int setNumber(int n) { DGTHIS_MEM(x,index) = n; return DGTHIS_MEM(x,index); }

class ICritter_NoData : public INamed_NoData
{
public:
	virtual const char * sound() = 0;
protected:
	~ICritter_NoData() {}
};
template< typename Base = ICritter_NoData >
class ICritter : public Base
{
protected:
	~ICritter() {}
	INamed_members;
};
#define TREEDB_ICritter_DELEGATE_TO(x) \
	TREEDB_INamed_DELEGATE_TO(x) \
	inline const char * sound() { return V3C_DELEGATE(x, sound(DGTHIS0(x))); }

#define DGTHIS0(x) &x
#define DGTHIS(x) &x,
#define DGTHIS_MEM(x,m) x.m
#define V3C_DELEGATE(Inst, x) Inst.lpVtbl->x
class Named
{
public:
	TREEDB_INamed_DELEGATE_TO(m_inst);
	named_t m_inst;
};
class Critter
{
public:
	TREEDB_ICritter_DELEGATE_TO(m_inst);
	critter_t m_inst;
};
#undef V3C_DELEGATE
#undef DGTHIS0
#undef DGTHIS
#undef DGTHIS_MEM
