/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include "meta-allocator.h"

#define TREEDB_NS_runtime_ptr_t treedb_allocator_on_storage_ptr_t

#include <treedb/ns-types.h>

#include <treedb/treedb-impl.h>

/* I've tried to always go through the vtable to call other members but I may
have missed one or two. Having another tree implementation would help. :) */

/*============================= "Regular" methods ============================*/

static BOOL STDCALL TREEDB_NS(Init)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, size_t uPayloadSize
, size_t size
, uint64_t * puPayloadOffs
) {
	(void)This;
	uPayloadSize
		= V3C_ROUND_UP_T(uintptr_t, uPayloadSize, sizeof(TREEDB_NS_count_t));
	uintptr_t used
		= pAllocator->m_used.m_handle
		+ pAllocator->m_used.m_pImpl->m_pDescription->m_root.m_uSize
		+ uPayloadSize;
	/* Pointer allocators use real pointers - account for this. */
	if(This->m_pDescription->m_flags & eTreedbContainerIsPointer)
		used -= (uintptr_t)pAllocator->m_pImpl->m_pStorage->m_pBase;

	used = V3C_ROUND_UP_T(size_t, used, sizeof(TREEDB_NS_count_t));
	if((used + pAllocator->m_used.m_pImpl->m_pDescription->m_aligned_node_size)
			> size)
		return 0; /* No room to store the first node. */

	treedb_tree_t_Init(& pAllocator->m_used);
	treedb_tree_t_Init(& pAllocator->m_free);

	if(puPayloadOffs)
		*puPayloadOffs = (uintptr_t)(used - uPayloadSize);
	/* Manually insert a free node that describes the available storage space.*/
	TREEDB_NSA(node_ptr_t) pFirstNode
		= (TREEDB_NSA(node_ptr_t))
			treedb_allocator_impl_t_OffsetToPointer(This, used);
	used += META_TREEDB_ALLOC_node_size;

	META_TREEDB_ALLOC_node_set_length(pFirstNode, size - used);
	META_TREEDB_ALLOC_node_set_prev(pFirstNode, 0);
	META_TREEDB_ALLOC_node_set_next(pFirstNode, 0);
	treedb_tree_t_InsertNodePtr(& pAllocator->m_free, pFirstNode);
#if META_TREEDB_DEBUG
	printf("%s : free node (%zu) : length %zu.\n"
		, __PRETTY_FUNCTION__
		, treedb_allocator_t_PointerToOffset(pAllocator, pFirstNode)
		, (uintptr_t)(size - used));
	fflush(stdout);
#endif
	return 1;
}
static int STDCALL TREEDB_NS(Check)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
) {
	(void)This;
	return TREEDB_NS(check_heap)(pAllocator);
}
	/* helper */
static size_t STDCALL TREEDB_NS(node_offset_data_offset)
( treedb_allocator_on_storage_impl_ptr_t This
, TREEDB_NS(count_t) noffs
) {
	(void)This;
	noffs &= ~((TREEDB_NS(count_t))1);
	if(!noffs)
		return 0;
	return treedb_allocator_impl_t_PointerToOffset
		(This, treedb_allocator_impl_t_PointerUnpack
			(This, noffs + META_TREEDB_ALLOC_node_size
		));
}
	/* helper */
static char * STDCALL TREEDB_NS(node_status)
( treedb_allocator_on_storage_impl_ptr_t This
, TREEDB_NS(item_ptr_t) pn
) {
#if META_TREEDB_DEBUG
	printf("%s : dumping %zu.\n", __PRETTY_FUNCTION__
		, treedb_allocator_impl_t_PointerToOffset(This, pn));
	fflush(stdout);
#endif
	char *p1 = 0, *p2, *p3;
	if(pn->m_offset_next || pn->m_offset_prev) {
		p2 = treedb_string_append(p1, " ");
		if(!p2) {
			free(p1);
			return 0;
		}
		p1 = p2;
	}
	if(pn->m_offset_prev) {
		if(asprintf(& p3, "{<- %zu"
			, TREEDB_NS(node_offset_data_offset)
					(This, pn->m_offset_prev))
				< 0) {
			free(p1);
			return 0;
		}
		p2 = treedb_string_append(p1, p3);
		free(p3);
		if(!p2)
			return 0;
		p1 = p2;
		if(TREEDB_item_isfree(pn->m_offset_prev))
			p2 = treedb_string_append(p1, "[free]}");
		else
			p2 = treedb_string_append(p1, "[used]}");
		if(!p2) {
			free(p1);
			return 0;
		}
		p1 = p2;
		if(pn->m_offset_next) {
			p2 = treedb_string_append(p1, " ");
			if(!p2) {
				free(p1);
				return 0;
			}
			p1 = p2;
		}
	}
	if(pn->m_offset_next) {
		if(asprintf(& p3, "{-> %zu"
			, TREEDB_NS(node_offset_data_offset)
				(This, pn->m_offset_next))
				< 0) {
			free(p1);
			return 0;
		}
		p2 = treedb_string_append(p1, p3);
		free(p3);
		if(!p2)
			return 0;
		p1 = p2;
		if(TREEDB_item_isfree(pn->m_offset_next))
			p2 = treedb_string_append(p1, "[free]}");
		else
			p2 = treedb_string_append(p1, "[used]}");
		if(!p2) {
			free(p1);
			return 0;
		}
		p1 = p2;
	}
	if(!p1)
		p1 = strdup("");
	return p1;
}
	/* helpers */
	typedef struct TREEDB_NS(data_renderer_t)
	TREEDB_NS(data_renderer_t)
	, * TREEDB_NS(data_renderer_ptr_t);
	typedef struct TREEDB_NS(data_renderer_vtbl_t)
	TREEDB_NS(data_renderer_vtbl_t)
	, * TREEDB_NS(data_renderer_vtbl_ptr_t);

	struct TREEDB_NS(data_renderer_vtbl_t) {
		treedb_tree_data_renderer_t_methods(,,TREEDB_NS(data_renderer_t));
	};
	struct TREEDB_NS(data_renderer_t)
	{
		treedb_tree_data_renderer_t_members(TREEDB_NS(data_renderer_vtbl_ptr_t));
		treedb_tree_data_renderer_ptr_t m_pd;
		BOOL is_used;
	};
	/* helper */
static PSTR STDCALL TREEDB_NS(data_Dump)
( TREEDB_NS(data_renderer_ptr_t) This
, treedb_tree_runtime_ptr_t runtime
, PCVOID pv
) {
#if META_TREEDB_DEBUG
	printf("%s : dumping %zu.\n", __PRETTY_FUNCTION__
		, treedb_tree_impl_t_PointerToOffset(runtime->m_pImpl, pv));
	fflush(stdout);
#endif
	TREEDB_NS(item_ptr_t) node = (TREEDB_NS(item_ptr_t))(pv);
	char * p1, * p2, * p3;
	if(This->is_used) {
		p1 = treedb_tree_data_renderer_t_Dump
			(This->m_pd, runtime, node->m_data);
		if(!p1)
			return 0;
		p2 = TREEDB_NS(node_status)
			( (treedb_allocator_on_storage_impl_ptr_t)
				(runtime->m_pImpl->m_pAllocator->m_pImpl)
			, node);
		if(!p2) {
			free(p1);
			return 0;
		}
		int res = asprintf(& p3, "[%zu : %zu] =%s=%s"
				, treedb_tree_impl_t_PointerToOffset(runtime->m_pImpl, node)
					+ runtime->m_pImpl->m_pAllocator->m_pImpl->m_pDescription
						->m_aligned_item_size
				, (size_t)node->m_uLength, p1, p2);
		free(p1);
		free(p2);
		if(res < 0)
			return 0;
		return p3;
	}else{
		p1 = TREEDB_NS(node_status)
			( (treedb_allocator_on_storage_impl_ptr_t)
				(runtime->m_pImpl->m_pAllocator->m_pImpl)
			, node);
		if(!p1)
			return 0;
		int res = asprintf(& p2, "[%zu : %zu]%s"
				, treedb_tree_impl_t_PointerToOffset(runtime->m_pImpl, node)
					+ runtime->m_pImpl->m_pAllocator->m_pImpl->m_pDescription
						->m_aligned_item_size
				, (size_t)node->m_uLength, p1);
		free(p1);
		if(res < 0) {
			free(p2);
			return 0;
		}
		return p2;
	}
}
	TREEDB_NS(data_renderer_vtbl_t) TREEDB_NS(data_renderer_vtbl) = {
		V3C_INVOKE1(treedb_tree_data_renderer_vtbl_t_entries,TREEDB_NS(data_))
	};
static void STDCALL TREEDB_NS(Dump)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, treedb_tree_renderer_ptr_t pr
, treedb_tree_data_renderer_ptr_t pd
) {
	(void)This;
	char * header = 0;
	int res
	= asprintf(& header, "Blocks [node size %zu, item size %zu, total %zu]"
		, (size_t)META_TREEDB_ALLOC_node_size - META_TREEDB_ALLOC_item_size
		, (size_t)META_TREEDB_ALLOC_item_size
		, (size_t)(META_TREEDB_ALLOC_node_size));
	if(res < 0)
		header = "Blocks";
	TREEDB_NS(data_renderer_t) allocator_data_renderer =
	{ (TREEDB_NS(data_renderer_vtbl_ptr_t))& TREEDB_NS(data_renderer_vtbl)
	, pd, 1 };
	treedb_tree_renderer_t_Add(pr, header, 1);
	if(res >= 0)
		free(header);
	treedb_tree_renderer_t_Push(pr);
	treedb_tree_renderer_t_Add(pr, "Used", 0);
	treedb_tree_renderer_t_Push(pr);
	treedb_tree_t_Dump(&pAllocator->m_used, pr
		, (treedb_tree_data_renderer_ptr_t)& allocator_data_renderer);
	treedb_tree_renderer_t_Pop(pr);
	treedb_tree_renderer_t_Add(pr, "Free", 1);
	treedb_tree_renderer_t_Push(pr);
	allocator_data_renderer.is_used = 0;
	treedb_tree_t_Dump(&pAllocator->m_free, pr
		, (treedb_tree_data_renderer_ptr_t)& allocator_data_renderer);
	treedb_tree_renderer_t_Pop(pr);
	treedb_tree_renderer_t_Pop(pr);
}
static BOOL STDCALL TREEDB_NS(LargestFree)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, size_t * poffs
, size_t * psz
) {
	(void)This;
	HTREEDB_TREENODE hn = treedb_tree_t_Tail(& pAllocator->m_free);
	if(!hn)
		return 0;
	if(poffs)
		* poffs = hn;
	if(!psz)
		return 1;
	PVOID pn = META_TREEDB_unpack_pointer(hn);
	* psz = META_TREEDB_ALLOC_node_length(pn);
	return 1;
}
static size_t STDCALL TREEDB_NS(Size)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
) {
	(void)This;
	(void)pAllocator;
	return treedb_storage_t_Size(This->m_pStorage);
}
#define ALLOC_OFFS(mem) \
(unsigned int)offsetof(TREEDB_NS(item_t), mem) + nsz
void STDCALL TREEDB_NS(DumpSchema)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, int fd
) {
#ifdef EXAMPLE
Header: 56, Allocator: 24, Node: 28, Item 12, Node total 40.

Allocator Node     Allocator Item
Of Name     Sz     Of Name          Sz     flags
-- -------- --     -- ------------- --     -----
 0 m_next    4     28 m_offset_prev  4     count
 4 m_prev    4     32 m_offset_next  4     depth
 8 m_parent  4     36 m_uLength      4
12 m_left    4
16 m_right   4
20 m_count   4
24 m_depth   1
--------------------------
40 Aligned total node size
#endif
	(void)pAllocator;
	FILE *fp;
	fp = fdopen(fd, "w");
	if(!fp)
		return;
	unsigned int sz = This->m_pDescription->m_pointer.m_uSize;
	unsigned int nsz = This->m_pDescription->m_aligned_node_size;
	unsigned int isz = This->m_pDescription->m_aligned_item_size;
	unsigned int totalsz = nsz + isz;
	fflush(fp);
	fprintf(fp,
	/* 1*/  "Header: %u, Allocator: %u, Node: %u, Item %u, Node total %u.\n\n"
	        "Allocator Node     Allocator Item\n"
	        "Of Name     Sz     Of Name          Sz     flags\n"
	        "-- -------- --     -- ------------- --     -----\n"
	/* 2*/  " 0 m_next   %2u     %2u m_offset_prev %2u     count\n"
	/* 3*/ "%2u m_prev   %2u     %2u m_offset_next %2u     depth\n"
	/* 4*/ "%2u m_parent %2u     %2u m_uLength     %2u\n"
	/* 5*/ "%2u m_left   %2u\n"
	/* 6*/ "%2u m_right  %2u\n"
	/* 7*/ "%2u m_count  %2u\n"
	/* 8*/ "%2u m_depth   1\n"
	"--------------------------\n"
	/* 9*/ "%2u Aligned total node size\n\n"
	, /* 1*/ (unsigned int)sizeof(TreedbHeader)
	, This->m_pDescription->m_root.m_uSize * 2
	, This->m_pDescription->m_node.m_uSize
	, This->m_pDescription->m_aligned_item_size
	, nsz + isz
	, /* 2*/ sz, nsz, sz
	, /* 3*/ sz, sz, ALLOC_OFFS(m_offset_next), sz
	, /* 4*/ sz*2, sz, ALLOC_OFFS(m_uLength), sz
	, /* 5*/ sz*3, sz
	, /* 6*/ sz*4, sz
	, /* 7*/ sz*5, sz
	, /* 8*/ sz*6
	, /* 9*/ totalsz
	);
	fflush(fp);
	fclose(fp);
}
#undef ALLOC_OFFS
static HTREEDB_HANDLE STDCALL TREEDB_NS(Alloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, uintptr_t cb
) {
	(void)This;
	return META_TREEDB_pack_pointer(TREEDB_NS(malloc)(pAllocator, cb));
}
static HTREEDB_HANDLE STDCALL TREEDB_NS(Calloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, uintptr_t  count
, uintptr_t  cb
) {
	(void)This;
	return META_TREEDB_pack_pointer(TREEDB_NS(calloc)(pAllocator, count, cb));
}
static HTREEDB_HANDLE STDCALL TREEDB_NS(AlignedAlloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, uintptr_t  cb
, USHORT align
, USHORT offs
) {
	(void)This;
	return META_TREEDB_pack_pointer(
		TREEDB_NS(malloc_aligned)(pAllocator, cb, align, offs));
}
static HTREEDB_HANDLE STDCALL TREEDB_NS(AlignedCalloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, uintptr_t  count
, uintptr_t  cb
, USHORT align
, USHORT offs
) {
	(void)This;
	return META_TREEDB_pack_pointer(
		TREEDB_NS(calloc_aligned)(pAllocator, count, cb, align, offs));
}

static PVOID STDCALL TREEDB_NS(PointerAlloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, uintptr_t  cb
) {
	(void)This;
	return TREEDB_NS(malloc)(pAllocator, cb);
}
static PVOID STDCALL TREEDB_NS(PointerCalloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, uintptr_t  count
, uintptr_t  cb
) {
	(void)This;
	return TREEDB_NS(calloc)(pAllocator, count, cb);
}
static PVOID STDCALL TREEDB_NS(PointerAlignedAlloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, uintptr_t  cb
, USHORT align
, USHORT offs
) {
	(void)This;
	return TREEDB_NS(malloc_aligned)(pAllocator, cb, align, offs);
}
static PVOID STDCALL TREEDB_NS(PointerAlignedCalloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, uintptr_t  count
, uintptr_t  cb
, USHORT align
, USHORT offs
) {
	(void)This;
	return TREEDB_NS(calloc_aligned)(pAllocator, count, cb, align, offs);
}

/*============================= "Handle" methods =============================*/

static HTREEDB_HANDLE STDCALL TREEDB_NS(Realloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, HTREEDB_HANDLE hv
, uintptr_t  cb
) {
	(void)This;
	return META_TREEDB_pack_pointer(
		TREEDB_NS(realloc)(pAllocator, META_TREEDB_unpack_pointer(hv), cb));
}
static PVOID STDCALL TREEDB_NS(PointerRealloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, HTREEDB_HANDLE hv
, uintptr_t  cb
) {
	(void)This;
	return TREEDB_NS(realloc)(pAllocator, META_TREEDB_unpack_pointer(hv), cb);
}
static TreedbFreeResult STDCALL TREEDB_NS(Free)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, HTREEDB_HANDLE hv
) {
	(void)This;
	return TREEDB_NS(free)(pAllocator, META_TREEDB_unpack_pointer(hv));
}
static uintptr_t STDCALL TREEDB_NS(GetSize)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, HTREEDB_HANDLE hv
) {
	hv -=
		This->m_pDescription->m_aligned_node_size
		+ This->m_pDescription->m_aligned_item_size;
	
	if(! TREEDB_NSU(search)(pAllocator, META_TREEDB_ALLOC_used_tree, hv))
		return 0;
	PVOID p = META_TREEDB_unpack_pointer(hv);
	return META_TREEDB_ALLOC_node_length(p);
}
static BOOL STDCALL TREEDB_NS(InHeap)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, HTREEDB_HANDLE u
) {
	(void)This;
	(void)pAllocator;
	return treedb_storage_t_InHeap(This->m_pStorage, u);
}
static BOOL STDCALL TREEDB_NS(DidAlloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, HTREEDB_HANDLE u
) {
	(void)This;
	TREEDB_NS_count_t offs = (TREEDB_NS_count_t)u
		- This->m_pDescription->m_aligned_node_size
		- This->m_pDescription->m_aligned_item_size;
	if(! TREEDB_NSU(search)(pAllocator, META_TREEDB_ALLOC_used_tree, offs))
		return 0;
	return 1;
}
static HTREEDB_HANDLE STDCALL TREEDB_NS(AlignedRealloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, HTREEDB_HANDLE hv
, uintptr_t cb
, USHORT align
, USHORT offs
) {
	(void)This;
	return META_TREEDB_pack_pointer(
	TREEDB_NS(realloc_aligned)(pAllocator
		, META_TREEDB_unpack_pointer(hv)
		, cb, align, offs));
}
static PVOID STDCALL TREEDB_NS(PointerAlignedRealloc)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, HTREEDB_HANDLE hv
, uintptr_t  cb
, USHORT align
, USHORT offs
) {
	(void)This;
	return TREEDB_NS(realloc_aligned)(pAllocator
		, META_TREEDB_unpack_pointer(hv)
		, cb, align, offs);
}

/*============================= "Pointer" methods ============================*/

static HTREEDB_HANDLE STDCALL TREEDB_NS(ReallocPtr)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, PVOID pv
, uintptr_t  cb
) {
	(void)This;
	return META_TREEDB_pack_pointer(
		TREEDB_NS(realloc)(pAllocator, pv, cb));
}
static PVOID STDCALL TREEDB_NS(PointerReallocPtr)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, PVOID pv
, uintptr_t  cb
) {
	(void)This;
	return TREEDB_NS(realloc)(pAllocator, pv, cb);
}
static TreedbFreeResult STDCALL TREEDB_NS(FreePtr)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, PVOID pv
) {
	(void)This;
	return TREEDB_NS(free)(pAllocator, pv);
}
static uintptr_t STDCALL TREEDB_NS(GetSizePtr)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, PCVOID pv
) {
	(void)This;
	HTREEDB_TREENODE hv = META_TREEDB_pack_pointer(pv);
	hv -=
		This->m_pDescription->m_aligned_node_size
		+ This->m_pDescription->m_aligned_item_size;
	if(! TREEDB_NSU(search)(pAllocator, META_TREEDB_ALLOC_used_tree, hv))
		return 0;
	PVOID p = META_TREEDB_unpack_pointer(hv);
	return META_TREEDB_ALLOC_node_length(p);
}
static BOOL STDCALL TREEDB_NS(InHeapPtr)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, PCVOID pv
) {
	(void)This;
	(void)pAllocator;
	return treedb_storage_t_InHeapPtr(This->m_pStorage, pv);
}
static BOOL STDCALL TREEDB_NS(DidAllocPtr)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, PCVOID pv
) {
	(void)This;
	HTREEDB_TREENODE hv = META_TREEDB_pack_pointer(pv);
	hv -=
		This->m_pDescription->m_aligned_node_size
		+ This->m_pDescription->m_aligned_item_size;
	if(! TREEDB_NSU(search)(pAllocator, META_TREEDB_ALLOC_used_tree, hv))
		return 0;
	return 1;
}
static HTREEDB_HANDLE STDCALL TREEDB_NS(AlignedReallocPtr)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, PVOID pv
, uintptr_t cb
, USHORT align
, USHORT offs
) {
	(void)This;
	return META_TREEDB_pack_pointer(
		TREEDB_NS(realloc_aligned)(pAllocator, pv, cb, align, offs));
}
static PVOID STDCALL TREEDB_NS(PointerAlignedReallocPtr)
( treedb_allocator_on_storage_impl_ptr_t This
, treedb_allocator_on_storage_ptr_t pAllocator
, PVOID pv
, uintptr_t  cb
, USHORT align
, USHORT offs
) {
	(void)This;
	return TREEDB_NS(realloc_aligned)(pAllocator, pv, cb, align, offs);
}

/*========================= Allocator-neutral methods ========================*/
static const unsigned char * STDCALL TREEDB_NS(Guid)
( treedb_allocator_on_storage_impl_ptr_t This
) {
	(void)This;
	return (const unsigned char *)UUID_treedb_avl_two_tree_allocator_algorithm;
}
static uintptr_t STDCALL TREEDB_NS(PointerToOffset)
( treedb_allocator_on_storage_impl_ptr_t This
, PCVOID pv
) {
	(void)This;
	uintptr_t ret = pv
		? (uintptr_t)(((char *)pv) - This->m_pStorage->m_pBase)
		: 0;
	return ret;
}
static PVOID STDCALL TREEDB_NS(OffsetToPointer)
( treedb_allocator_on_storage_impl_ptr_t This
, HTREEDB_TREENODE hn
) {
	(void)This;
	PVOID ret = hn
		? This->m_pStorage->m_pBase + hn
		: 0;
	return ret;
}

/*------------------- These get the internal representation ------------------*/

#if TREEDB_NBITS != 0
static uintptr_t STDCALL TREEDB_NS(PointerPack)
( treedb_allocator_on_storage_impl_ptr_t This
, PCVOID pv
) {
	(void)This;
	uintptr_t ret = pv
		? (uintptr_t)(((char *)pv) - This->m_pStorage->m_pBase)
		: 0;
	return ret;
}
static PVOID STDCALL TREEDB_NS(PointerUnpack)
( treedb_allocator_on_storage_impl_ptr_t This
, uintptr_t u
) {
	(void)This;
	PVOID ret = u
		? This->m_pStorage->m_pBase + u
		: 0;
	return ret;
}
#else
static uintptr_t STDCALL TREEDB_NS(PointerPack)
( treedb_allocator_on_storage_impl_ptr_t This
, PCVOID pv
) {
	(void)This;
	return (uintptr_t)pv;
}
static PVOID STDCALL TREEDB_NS(PointerUnpack)
( treedb_allocator_on_storage_impl_ptr_t This
, uintptr_t u
) {
	(void)This;
	return (PVOID)u;
}
#endif

/*===================== Allocator-neutral "Handle" methods ===================*/

static HTREEDB_HANDLE STDCALL TREEDB_NS(Read)
( treedb_allocator_on_storage_impl_ptr_t This
, HTREEDB_HANDLE hloc
) {
	(void)This;
	if(!hloc)
		return 0;
	return *(TREEDB_NS_count_t *)META_TREEDB_unpack_pointer(hloc);
}
static PVOID STDCALL TREEDB_NS(PointerRead)
( treedb_allocator_on_storage_impl_ptr_t This
, HTREEDB_HANDLE hloc
) {
	(void)This;
	if(!hloc)
		return 0;
	HTREEDB_HANDLE value = TREEDB_NS(Read)(This, hloc);
	return value ? This->m_pStorage->m_pBase + value : 0;
}
static void STDCALL TREEDB_NS(Write)
( treedb_allocator_on_storage_impl_ptr_t This
, HTREEDB_HANDLE hloc
, HTREEDB_HANDLE hvalue
) {
	(void)This;
	if(!hloc)
		return;
	*(TREEDB_NS_count_t *)META_TREEDB_unpack_pointer(hloc)
		= (TREEDB_NS_count_t)hvalue;
}

/*------------------- My naming convention broke down here :) ----------------*/

static void STDCALL TREEDB_NS(PointerWrite)
( treedb_allocator_on_storage_impl_ptr_t This
, HTREEDB_HANDLE hloc
, PCVOID pv
) {
	TREEDB_NS(Write)(This, hloc, pv ?
		(size_t)(((char *)pv) - This->m_pStorage->m_pBase) : 0);
}
#if TREEDB_NBITS != 0
static void STDCALL TREEDB_NS(AddTo)
( treedb_allocator_on_storage_impl_ptr_t This
, HTREEDB_HANDLE hloc
, uintptr_t u
) {
	(void)This;
	if(!hloc)
		return;
	*((TREEDB_NS_count_t *)META_TREEDB_unpack_pointer(hloc))
		+= (TREEDB_NS_count_t)u;
}
#else
static void STDCALL TREEDB_NS(AddTo)
( treedb_allocator_on_storage_impl_ptr_t This
, HTREEDB_HANDLE hloc
, uintptr_t u
) {
	(void)This;
	*(TREEDB_NS_count_t *)hloc += (TREEDB_NS_count_t)u;
}
#endif

/*===================== Allocator-neutral "Pointer" methods ==================*/

static HTREEDB_HANDLE STDCALL TREEDB_NS(ReadPtr)
( treedb_allocator_on_storage_impl_ptr_t This
, PCVOID pv
) {
	(void)This;
	return *((TREEDB_NS_count_t *)pv);
}
static PVOID STDCALL TREEDB_NS(PointerReadPtr)
( treedb_allocator_on_storage_impl_ptr_t This
, PCVOID pv
) {
	(void)This;
	size_t sz = TREEDB_NS(ReadPtr)(This, pv);
	return sz ? This->m_pStorage->m_pBase + sz : 0;
}
static void STDCALL TREEDB_NS(WritePtr)
( treedb_allocator_on_storage_impl_ptr_t This
, PVOID pv
, HTREEDB_HANDLE hvalue
) {
	(void)This;
	if(!pv)
		return;
	*(TREEDB_NS_count_t *)pv = (TREEDB_NS_count_t)hvalue;
}

/*------------------- My naming convention broke down here :) ----------------*/

static void STDCALL TREEDB_NS(PointerWritePtr)
( treedb_allocator_on_storage_impl_ptr_t This
, PVOID pv
, PCVOID pvvalue
) {
	if(!pv)
		return;
	*(TREEDB_NS_count_t *)pv = META_TREEDB_pack_pointer(pvvalue);
}
static void STDCALL TREEDB_NS(AddToPtr)
( treedb_allocator_on_storage_impl_ptr_t This
, PVOID pv
, uintptr_t u
) {
	(void)This;
	*((TREEDB_NS_count_t *)pv) += (TREEDB_NS_count_t)u;
}
static uint8_t STDCALL TREEDB_NS(EndianIndex)
( treedb_allocator_on_storage_impl_ptr_t This
) {
	(void)This;
	return treedb_storage_t_EndianIndex(This->m_pStorage);
}
static uint8_t STDCALL TREEDB_NS(EndianSwapIndex)
( treedb_allocator_on_storage_impl_ptr_t This
) {
	(void)This;
	return treedb_storage_t_EndianSwapIndex(This->m_pStorage);
}


#if 0
#define ve(x) v->x = & TREEDB_NS(x)
static void TREEDB_NS(test)(treedb_allocator_on_storage_impl_vtbl_ptr_t v)
{
	ve(Init);
	ve(Check);
	ve(Dump);
	ve(LargestFree);
	ve(Size);
	ve(DumpSchema);
	ve(Guid);
	ve(PointerToOffset);
	ve(OffsetToPointer);

	ve(Alloc);
	ve(Calloc);
	ve(AlignedAlloc);
	ve(AlignedCalloc);

	ve(PointerAlloc);
	ve(PointerCalloc);
	ve(PointerAlignedAlloc);
	ve(PointerAlignedCalloc);

	ve(PointerPack);
	ve(PointerUnpack);

	ve(Realloc);
	ve(PointerRealloc);
	ve(Free);
	ve(GetSize);
	ve(InHeap);
	ve(DidAlloc);
	ve(AlignedRealloc);
	ve(PointerAlignedRealloc);
	ve(Read);
	ve(Write);
	ve(PointerRead);
	ve(PointerWrite);
	ve(AddTo);

	ve(ReallocPtr);
	ve(PointerReallocPtr);
	ve(FreePtr);
	ve(GetSizePtr);
	ve(InHeapPtr);
	ve(DidAllocPtr);
	ve(AlignedReallocPtr);
	ve(PointerAlignedReallocPtr);
	ve(ReadPtr);
	ve(WritePtr);
	ve(PointerReadPtr);
	ve(PointerWritePtr);
	ve(AddToPtr);
	ve(EndianIndex);
}
#endif
treedb_allocator_on_storage_impl_vtbl_t
TREEDB_NS(allocator_on_storage_impl_vtbl) = {
	V3C_INVOKE1(treedb_allocator_impl_vtbl_t_entries,TREEDB_NS_)
};
