/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

/*! @file meta-avl-allocator-x.h

I decided to seperate the allocator bits from meta-avl.c to at least try to
show that the allocator really is independent from the tree implementation.

Of course including "meta-avl.h" doesn't help, but we need a tree implementation
from somewhere.

I guess the only way to be sure is to implement another tree type...
*/
#include "meta-avl.h"

/* We just want sizeof(AVL_TMPL_NS(tree_t)) etc. */
#define AVL_TMPL_prefix

#define AVL_TMPL_NS(x) V3C_INVOKE2(V3C_JOIN2,AVL_TMPL_NS_,x)
/* #define AVL_TMPL_input_t void * */

#ifdef DOXYGEN_processing
{
#endif

#include <v3c/meta-avl-defines.h>

#ifdef DOXYGEN_processing
}
#endif

#include <stddef.h> /* for offsetof() */

#define AVL_TMPL_size_t uintptr_t
#define AVL_TMPL_ssize_t intptr_t

#define treedb_Tree(This) \
((AVL_TMPL_NS(tree_ptr_t)) \
	(treedb_tree_t_PointerUnpack(This, This->m_tree_offset)))

#define AVL_TMPL_packed_node_ptr_t AVL_TMPL_count_t
#define AVL_TMPL_allow_duplicates treedb_tree_t_allows_duplicates(runtime)
#define AVL_TMPL_cmp(a,b) treedb_tree_backend_Compare(a,b)
#define AVL_TMPL_unpack_item(i, o)
#define AVL_TMPL_unpack_key(i, o) \
	o = treedb_tree_backend_t_UnpackKey(runtime->m_pImpl->m_pBackend, runtime,i)
#define AVL_TREE treedb_Tree(runtime)
#define AVL_TMPL_unpacked_key AVL_TMPL_const_key_t unpacked_key

#define AVL_TMPL_unpack_pointer(x) \
	treedb_tree_t_PointerUnpack(runtime, x)
#define AVL_TMPL_pack_pointer(x) \
	((AVL_TMPL_count_t)treedb_tree_t_PointerPack(runtime, x))

#define AVL_TMPL_alloc_node(x) \
	(AVL_TMPL_NS(node_ptr_t)) \
	(treedb_tree_backend_t_AllocNode(runtime->m_pImpl->m_pBackend, runtime, x))
#define AVL_TMPL_free_node(x) \
	treedb_tree_backend_t_FreeNode(runtime->m_pImpl->m_pBackend, runtime, x)
#define AVL_TMPL_free(x) treedb_allocator_t_Free(runtime->m_pAllocator, x)

/******************************************************************************/
/* o16                                                                        */
/******************************************************************************/
#define AVL_TMPL_have_count
#define AVL_TMPL_NS_ treedb_avl_tree_o16c_
#define AVL_TMPL_count_t uint16_t
/* Map to the name in meta-allocator-impl.h */
#define treedb_avl_tree_o16c_allocator_on_storage_impl_vtbl \
treedb_allocator_o16_allocator_on_storage_impl_vtbl

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_have_count
#define AVL_TMPL_have_depth
#undef AVL_TMPL_NS_
#define AVL_TMPL_NS_ treedb_avl_tree_o16d_
#define treedb_avl_tree_o16d_allocator_on_storage_impl_vtbl \
treedb_allocator_o16_allocator_on_storage_impl_vtbl

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#define AVL_TMPL_have_count
#undef AVL_TMPL_NS_
#define AVL_TMPL_NS_ treedb_avl_tree_o16cd_
#define treedb_avl_tree_o16cd_allocator_on_storage_impl_vtbl \
treedb_allocator_o16_allocator_on_storage_impl_vtbl

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_have_count
#undef AVL_TMPL_have_depth
#undef AVL_TMPL_NS_
#undef AVL_TMPL_count_t

/******************************************************************************/
/* o32                                                                        */
/******************************************************************************/
#define AVL_TMPL_have_count
#define AVL_TMPL_NS_ treedb_avl_tree_o32c_
#define AVL_TMPL_count_t uint32_t
#define treedb_avl_tree_o32c_allocator_on_storage_impl_vtbl \
treedb_allocator_o32_allocator_on_storage_impl_vtbl

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_have_count
#define AVL_TMPL_have_depth
#undef AVL_TMPL_NS_
#define AVL_TMPL_NS_ treedb_avl_tree_o32d_
#define treedb_avl_tree_o32d_allocator_on_storage_impl_vtbl \
treedb_allocator_o32_allocator_on_storage_impl_vtbl

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#define AVL_TMPL_have_count
#undef AVL_TMPL_NS_
#define AVL_TMPL_NS_ treedb_avl_tree_o32cd_
#define treedb_avl_tree_o32cd_allocator_on_storage_impl_vtbl \
treedb_allocator_o32_allocator_on_storage_impl_vtbl

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_have_depth
#undef AVL_TMPL_have_count
#undef AVL_TMPL_NS_
#undef AVL_TMPL_count_t

#if HAVE_64_BITS
/******************************************************************************/
/* o64                                                                        */
/******************************************************************************/
#define AVL_TMPL_have_count
#define AVL_TMPL_NS_ treedb_avl_tree_o64c_
#define AVL_TMPL_count_t uint64_t
#define treedb_avl_tree_o64c_allocator_on_storage_impl_vtbl \
treedb_allocator_o64_allocator_on_storage_impl_vtbl

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_have_count
#define AVL_TMPL_have_depth
#undef AVL_TMPL_NS_
#define AVL_TMPL_NS_ treedb_avl_tree_o64d_
#define treedb_avl_tree_o64d_allocator_on_storage_impl_vtbl \
treedb_allocator_o64_allocator_on_storage_impl_vtbl

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#define AVL_TMPL_have_count
#undef AVL_TMPL_NS_
#define AVL_TMPL_NS_ treedb_avl_tree_o64cd_
#define treedb_avl_tree_o64cd_allocator_on_storage_impl_vtbl \
treedb_allocator_o64_allocator_on_storage_impl_vtbl

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_have_depth
#undef AVL_TMPL_have_count
#undef AVL_TMPL_NS_
#undef AVL_TMPL_count_t

#endif /* #if HAVE_64_BITS */

/******************************************************************************/
/* pointer                                                                    */
/******************************************************************************/
#define v3c_nbits_u0(x) v3c_nbits_up(x)
#define AVL_POINTER_TREE
#define AVL_TMPL_count_t uintptr_t

#undef AVL_TMPL_unpack_pointer
#undef AVL_TMPL_pack_pointer
#define AVL_TMPL_unpack_pointer(x) ((PVOID)(x))
#define AVL_TMPL_pack_pointer(x) ((AVL_TMPL_count_t)(x))

#define TREEDB_OMIT_AVL_TMPL_pack_tree
#define TREEDB_OMIT_AVL_TMPL_pack_node
#define TREEDB_OMIT_AVL_TMPL_unpack_tree
#define TREEDB_OMIT_AVL_TMPL_unpack_node

#define AVL_TMPL_have_count
#define AVL_TMPL_NS_ treedb_avl_tree_pc_
#define treedb_avl_tree_pc_allocator_on_storage_impl_vtbl \
treedb_allocator_pointer_allocator_on_storage_impl_vtbl

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_have_count
#define AVL_TMPL_have_depth
#undef AVL_TMPL_NS_
#define AVL_TMPL_NS_ treedb_avl_tree_pd_
#define treedb_avl_tree_pd_allocator_on_storage_impl_vtbl \
treedb_allocator_pointer_allocator_on_storage_impl_vtbl

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_have_count
#undef AVL_TMPL_NS_
#define AVL_TMPL_NS_ treedb_avl_tree_pcd_
#define treedb_avl_tree_pcd_allocator_on_storage_impl_vtbl \
treedb_allocator_pointer_allocator_on_storage_impl_vtbl

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-allocator-impl.h"

#ifdef DOXYGEN_processing
}
#endif

treedb_const_allocator_on_storage_ptr_t
treedb_get_allocator_on_storage_template_for
( uint16_t uBytes /* 2, 4, 8  */
, uint16_t flags /* flags & eTreedbContainerIsPointer means ignore uBytes */
) {
	if(flags & eTreedbContainerIsPointer) {
		switch(flags & ~((uint16_t)eTreedbContainerIsPointer)) {
		case eTreedbContainerHasCount:
			return & treedb_avl_tree_pc_allocator_on_storage;
		case eTreedbContainerHasDepth:
			return & treedb_avl_tree_pd_allocator_on_storage;
		case eTreedbContainerHasCount | eTreedbContainerHasDepth:
			return & treedb_avl_tree_pcd_allocator_on_storage;
		default:
			break;
		}
		return 0;
	}
	switch(uBytes) {
	case 2:
		switch(flags) {
		case eTreedbContainerHasCount:
			return & treedb_avl_tree_o16c_allocator_on_storage;
		case eTreedbContainerHasDepth:
			return & treedb_avl_tree_o16d_allocator_on_storage;
		case eTreedbContainerHasCount | eTreedbContainerHasDepth:
			return & treedb_avl_tree_o16cd_allocator_on_storage;
		default:
			break;
		}
		break;
	case 4:
		switch(flags) {
		case eTreedbContainerHasCount:
			return & treedb_avl_tree_o32c_allocator_on_storage;
		case eTreedbContainerHasDepth:
			return & treedb_avl_tree_o32d_allocator_on_storage;
		case eTreedbContainerHasCount | eTreedbContainerHasDepth:
			return & treedb_avl_tree_o32cd_allocator_on_storage;
		default:
			break;
		}
		break;
#if HAVE_64_BITS
	case 8:
		switch(flags) {
		case eTreedbContainerHasCount:
			return & treedb_avl_tree_o64c_allocator_on_storage;
		case eTreedbContainerHasDepth:
			return & treedb_avl_tree_o64d_allocator_on_storage;
		case eTreedbContainerHasCount | eTreedbContainerHasDepth:
			return & treedb_avl_tree_o64cd_allocator_on_storage;
		default:
			break;
		}
		break;
#endif
	default:
		break;
	}
	return 0;
}

BOOL
treedb_init_allocator_on_storage_from_template
( treedb_allocator_on_storage_ptr_t pAllocator
, treedb_allocator_on_storage_impl_ptr_t pAllocatorImpl
, uint16_t uBytes /* 2, 4, 8 */
, uint16_t flags /* flags & eTreedbContainerIsPointer means ignore uBytes */
, uint64_t offset
, treedb_storage_ptr_t pStorage
) {
	if(!pAllocator || !pStorage)
		return 0;
	treedb_const_allocator_on_storage_ptr_t pt
	= treedb_get_allocator_on_storage_template_for
		(uBytes, flags);
	if(!pt)
		return 0;
	*pAllocator = *pt;
	*pAllocatorImpl = * pt->m_pImpl;
	pAllocator->m_pImpl = pAllocatorImpl;
	pAllocatorImpl->m_pStorage = pStorage;
	treedb_allocator_on_storage_t_Setup(pAllocator);
	pAllocator->m_used.m_handle += offset;
	pAllocator->m_free.m_handle += offset;
	return 1;
}

treedb_const_tree_ptr_t
treedb_get_avl_tree_template_for
( uint16_t uBytes /* 2, 4, 8 */
, uint16_t flags /* flags & eTreedbContainerIsPointer means ignore uBytes */
) {
	treedb_const_allocator_on_storage_ptr_t p
	= treedb_get_allocator_on_storage_template_for
		(uBytes, flags);
	if(!p)
		return 0;
	return & p->m_used;
}
treedb_const_tree_ptr_t
treedb_get_avl_tree_template_using_allocator
( treedb_allocator_ptr_t pallocator
) {
	if(!pallocator)
		return 0;
	treedb_const_tree_ptr_t pt
		= treedb_get_avl_tree_template_for
			( pallocator->m_pImpl->m_pDescription->m_pointer.m_uSize
			, pallocator->m_pImpl->m_pDescription->m_flags);
	return pt;
}

BOOL
treedb_init_avl_tree_from_template
( treedb_tree_ptr_t p
, uint16_t uBytes /* 2, 4, 8 */
, uint16_t flags /* flags & eTreedbContainerIsPointer means ignore uBytes */
) {
	if(!p || !p->m_pImpl)
		return 0;
	treedb_const_tree_ptr_t pt
	= treedb_get_avl_tree_template_for(uBytes, flags);
	if(!pt)
		return 0;
	treedb_tree_impl_ptr_t pImpl = p->m_pImpl;
	p->m_handle = 0;
	*pImpl = * pt->m_pImpl;
	return 1;
}

BOOL
treedb_init_avl_tree_using_allocator
( treedb_tree_ptr_t p
, treedb_allocator_ptr_t pallocator
) {
	return treedb_init_avl_tree_from_template
		(p
		, pallocator->m_pImpl->m_pDescription->m_pointer.m_uSize
		, pallocator->m_pImpl->m_pDescription->m_flags);
}
