/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <treedb/avl-template-impl.h>
#include <malloc.h> /* free() */

/* Forward declarations.*/
static uintptr_t STDCALL AVL_TMPL_NS(NodeCount)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREENODE hn
);
static void STDCALL AVL_TMPL_NS(DumpNode)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREENODE hn
, treedb_tree_renderer_ptr_t pr
, treedb_tree_data_renderer_ptr_t pd
);
static void STDCALL AVL_TMPL_NS(DumpNodePtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pn
, treedb_tree_renderer_ptr_t pr
, treedb_tree_data_renderer_ptr_t pd
);
static uintptr_t STDCALL AVL_TMPL_NS(NodeCountPtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pn
);

/* Implementation */
static void STDCALL AVL_TMPL_NS(Init)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
) {
	AVL_TMPL_NS(init)(runtime, AVL_TMPL_unpack_tree(ht));
}
static void STDCALL AVL_TMPL_NS(InitPtr)
( treedb_tree_runtime_ptr_t runtime
, PVOID pt
) {
	AVL_TMPL_NS(init)(runtime, (AVL_TMPL_tree_ptr_t)pt);
}
static void STDCALL AVL_TMPL_NS(Clear)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
) {
	AVL_TMPL_NS(clear)(runtime, (AVL_TMPL_packed_tree_ptr_t)ht);
}
static BOOL STDCALL AVL_TMPL_NS(Clone)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht_to
, HTREEDB_TREE ht_from
) {
	return ! AVL_TMPL_NS(clone)(runtime, (AVL_TMPL_packed_tree_ptr_t)ht_to
		, (AVL_TMPL_packed_tree_ptr_t)ht_from);
}
static HTREEDB_TREE STDCALL AVL_TMPL_NS(Alloc)
( treedb_tree_runtime_ptr_t runtime
) {
	AVL_TMPL_NS(tree_ptr_t) pt = (AVL_TMPL_NS(tree_ptr_t))
		treedb_allocator_t_Alloc(runtime->m_pImpl->m_pAllocator
			, sizeof(AVL_TMPL_NS(tree_t)));
	if(!pt)
		return 0;
	pt->m_head = pt->m_tail = pt->m_top = 0;
	return (HTREEDB_TREE)AVL_TMPL_pack_tree(pt);
}
static void STDCALL AVL_TMPL_NS(Free)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
) {
	if(!ht)
		return;
	treedb_allocator_t_Free(runtime->m_pImpl->m_pAllocator, ht);
}
static uintptr_t STDCALL AVL_TMPL_NS(Count)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
) {
	uintptr_t count = AVL_TMPL_NS(count)(runtime, AVL_TMPL_unpack_tree(ht));
	return count;
}
static TreedbInsertResult STDCALL AVL_TMPL_NS(Insert)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREEKEY hkey
, PVOID * ppn
) {
	return AVL_TMPL_NS(insert)(runtime, (AVL_TMPL_packed_tree_ptr_t)ht, hkey
		, (AVL_TMPL_node_ptr_t *)ppn);
}
static BOOL STDCALL AVL_TMPL_NS(Remove)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREEKEY hkey
) {
	return ! AVL_TMPL_NS(remove)(runtime, (AVL_TMPL_packed_tree_ptr_t)ht, hkey);
}
static BOOL STDCALL AVL_TMPL_NS(RemoveBetween)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREENODE hn_lower
, HTREEDB_TREENODE hn_upper
, TreedbNodeDistanceFlags flags
) {
	return ! AVL_TMPL_NS(remove_between)(runtime
		, (AVL_TMPL_packed_tree_ptr_t)ht
		, (AVL_TMPL_packed_node_ptr_t)hn_lower
		, (AVL_TMPL_packed_node_ptr_t)hn_upper
		, flags);
}
static BOOL STDCALL AVL_TMPL_NS(RemoveMany)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREEKEY lower_key
, HTREEDB_TREEKEY upper_key
, TreedbKeyDistanceFlags flags
) {
	return ! AVL_TMPL_NS(remove_many)(runtime, (AVL_TMPL_packed_tree_ptr_t)ht
		, lower_key, upper_key, flags);
}
static intptr_t STDCALL AVL_TMPL_NS(NodeDistance)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREENODE hn_lower
, HTREEDB_TREENODE hn_upper
, TreedbNodeDistanceFlags flags
) {
	return AVL_TMPL_NS(node_distance)(runtime
		, AVL_TMPL_unpack_tree(ht)
		, AVL_TMPL_unpack_node(hn_lower)
		, AVL_TMPL_unpack_node(hn_upper)
		, flags);
}
static BOOL STDCALL AVL_TMPL_NS(KeyDistance)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREEKEY hlower
, HTREEDB_TREEKEY hupper
, TreedbKeyDistanceFlags flags
, intptr_t * pres
) {
	return ! AVL_TMPL_NS(key_distance)(runtime, AVL_TMPL_unpack_tree(ht)
		, hlower, hupper, flags, pres);
}
static void STDCALL AVL_TMPL_NS(Dump)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, treedb_tree_renderer_ptr_t pr
, treedb_tree_data_renderer_ptr_t pd
) {
	if(!ht)
		return;
	HTREEDB_TREENODE node = (HTREEDB_TREENODE)(AVL_TMPL_unpack_tree(ht)->m_top);
	AVL_TMPL_NS(DumpNode)(runtime, node, pr, pd);
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(Head)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
) {
	(void)runtime;
	return (HTREEDB_TREENODE)(AVL_TMPL_unpack_tree(ht)->m_head);
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(Tail)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
) {
	(void)runtime;
	return (HTREEDB_TREENODE)(AVL_TMPL_unpack_tree(ht)->m_tail);
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(Top)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
) {
	(void)runtime;
	return (HTREEDB_TREENODE)(AVL_TMPL_unpack_tree(ht)->m_top);
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(Index)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, uintptr_t u
) {
	return (HTREEDB_TREENODE)AVL_TMPL_pack_node
	( AVL_TMPL_NS(index)(runtime, AVL_TMPL_unpack_tree(ht), u)
	);
}
static PVOID STDCALL AVL_TMPL_NS(PointerHead)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
) {
	(void)runtime;
	return AVL_TMPL_unpack_node(
		(HTREEDB_TREENODE)(AVL_TMPL_unpack_tree(ht)->m_head));
}
static PVOID STDCALL AVL_TMPL_NS(PointerTail)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
) {
	(void)runtime;
	return AVL_TMPL_unpack_node(
		(HTREEDB_TREENODE)(AVL_TMPL_unpack_tree(ht)->m_tail));
}
static PVOID STDCALL AVL_TMPL_NS(PointerTop)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
) {
	(void)runtime;
	return AVL_TMPL_unpack_node(
		(HTREEDB_TREENODE)(AVL_TMPL_unpack_tree(ht)->m_top));
}
static PVOID STDCALL AVL_TMPL_NS(PointerIndex)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, uintptr_t u
) {
	return
		AVL_TMPL_NS(index)(runtime, AVL_TMPL_unpack_tree(ht), u);
}
/* Handle ------------------------------------------------------------------- */
static void STDCALL AVL_TMPL_NS(DumpNode)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREENODE hn
, treedb_tree_renderer_ptr_t pr
, treedb_tree_data_renderer_ptr_t pd
) {
	if(!hn)
		return;
	AVL_TMPL_NS(node_ptr_t) pn = AVL_TMPL_unpack_node(hn);
	AVL_TMPL_NS(DumpNodePtr)(runtime, pn, pr, pd);
}
static BOOL STDCALL AVL_TMPL_NS(InsertNode)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREENODE hn
) {
	return AVL_TMPL_NS(insert_node)
	( runtime
	, AVL_TMPL_unpack_tree((AVL_TMPL_packed_tree_ptr_t)ht)
	, AVL_TMPL_unpack_node((AVL_TMPL_packed_node_ptr_t)hn)
	) != 0;
}
static void STDCALL AVL_TMPL_NS(InsertNodeBefore)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREENODE hn
, HTREEDB_TREENODE hn_new
) {
	AVL_TMPL_NS(insert_node_before)
	( runtime
	, AVL_TMPL_unpack_tree((AVL_TMPL_packed_tree_ptr_t)ht)
	, AVL_TMPL_unpack_node((AVL_TMPL_packed_node_ptr_t)hn)
	, AVL_TMPL_unpack_node((AVL_TMPL_packed_node_ptr_t)hn_new)
	);
}
static void STDCALL AVL_TMPL_NS(InsertNodeAfter)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREENODE hn
, HTREEDB_TREENODE hn_new
) {
	AVL_TMPL_NS(insert_node_after)
	( runtime
	, AVL_TMPL_unpack_tree((AVL_TMPL_packed_tree_ptr_t)ht)
	, AVL_TMPL_unpack_node((AVL_TMPL_packed_node_ptr_t)hn)
	, AVL_TMPL_unpack_node((AVL_TMPL_packed_node_ptr_t)hn_new)
	);
}
static void STDCALL AVL_TMPL_NS(RelinkNode)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREENODE hn
) {
	AVL_TMPL_NS(relink_node)
	( runtime
	, AVL_TMPL_unpack_tree((AVL_TMPL_packed_tree_ptr_t)ht)
	, AVL_TMPL_unpack_node((AVL_TMPL_packed_node_ptr_t)hn)
	);
}
static void STDCALL AVL_TMPL_NS(UnlinkNode)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREENODE hn
) {
	AVL_TMPL_NS(unlink_node)
	( runtime
	, AVL_TMPL_unpack_tree((AVL_TMPL_packed_tree_ptr_t)ht)
	, AVL_TMPL_unpack_node((AVL_TMPL_packed_node_ptr_t)hn)
	);
}
static void STDCALL AVL_TMPL_NS(DeleteNode)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREENODE hn
) {
	AVL_TMPL_NS(delete_node)
	( runtime
	, (AVL_TMPL_packed_tree_ptr_t)ht
	, AVL_TMPL_unpack_node((AVL_TMPL_packed_node_ptr_t)hn)
	);
}
static TreedbSearchResult STDCALL AVL_TMPL_NS(SearchClosest)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREEKEY key
, HTREEDB_TREENODE * phn
) {
	AVL_TMPL_NS(node_ptr_t) pn;
	TreedbSearchResult res = AVL_TMPL_NS(search_closest)
	(runtime, AVL_TMPL_unpack_tree(ht), key, & pn);
	if(phn)
		*phn = (HTREEDB_TREENODE)AVL_TMPL_pack_node(pn);
	return res;
}
static TreedbSearchResult STDCALL AVL_TMPL_NS(SearchLe)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREEKEY key
, HTREEDB_TREENODE * phn
) {
	AVL_TMPL_NS(node_ptr_t) pn;
	TreedbSearchResult res = AVL_TMPL_NS(search_le)
	(runtime, AVL_TMPL_unpack_tree(ht), key, & pn);
	if(phn)
		*phn = (HTREEDB_TREENODE)AVL_TMPL_pack_node(pn);
	return res;
}
static TreedbSearchResult STDCALL AVL_TMPL_NS(SearchGe)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREEKEY key
, HTREEDB_TREENODE * phn
) {
	AVL_TMPL_NS(node_ptr_t) pn;
	TreedbSearchResult res = AVL_TMPL_NS(search_ge)
	(runtime, AVL_TMPL_unpack_tree(ht), key ,& pn);
	if(phn)
		*phn = (HTREEDB_TREENODE)AVL_TMPL_pack_node(pn);
	return res;
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(Search)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREEKEY key
) {
	return (HTREEDB_TREENODE)AVL_TMPL_pack_node
	( AVL_TMPL_NS(search)
		( runtime
		, AVL_TMPL_unpack_tree(ht)
		, key
		)
	);
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(SearchLt)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREEKEY key
) {
	return (HTREEDB_TREENODE)AVL_TMPL_pack_node
	( AVL_TMPL_NS(search_lt)
		( runtime
		, AVL_TMPL_unpack_tree(ht)
		, key
		)
	);
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(SearchGt)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREEKEY key
) {
	return (HTREEDB_TREENODE)AVL_TMPL_pack_node
	( AVL_TMPL_NS(search_gt)
		( runtime
		, AVL_TMPL_unpack_tree(ht)
		, key
		)
	);
}
static BOOL STDCALL AVL_TMPL_NS(NodeHasNext)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREENODE hn
) {
	(void)runtime;
	return hn ? (AVL_TMPL_unpack_node(hn)->m_next != 0) : 0;
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(NodeNext)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREENODE hn
) {
	(void)runtime;
	return hn ? (HTREEDB_TREENODE)(AVL_TMPL_unpack_node(hn)->m_next) : 0;
}
static PVOID STDCALL AVL_TMPL_NS(PointerNodeNext)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREENODE hn
) {
	(void)runtime;
	return hn ?
		AVL_TMPL_unpack_node(
		(HTREEDB_TREENODE)(AVL_TMPL_unpack_node(hn)->m_next))
		: 0;
}
static BOOL STDCALL AVL_TMPL_NS(NodeHasPrev)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREENODE hn
) {
	(void)runtime;
	return hn ? (AVL_TMPL_unpack_node(hn)->m_prev != 0) : 0;
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(NodePrev)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREENODE hn
) {
	(void)runtime;
	return hn ? (HTREEDB_TREENODE)(AVL_TMPL_unpack_node(hn)->m_prev) : 0;
}
static PVOID STDCALL AVL_TMPL_NS(PointerNodePrev)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREENODE hn
) {
	(void)runtime;
	return hn ?
		AVL_TMPL_unpack_node(
		(HTREEDB_TREENODE)(AVL_TMPL_unpack_node(hn)->m_prev))
		: 0;
}
static uintptr_t STDCALL AVL_TMPL_NS(NodeIndex)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, HTREEDB_TREENODE hn
) {
	return
		AVL_TMPL_NS(node_index)(runtime, AVL_TMPL_unpack_tree(ht)
			, AVL_TMPL_unpack_node(hn));
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(NodeAtIndex)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREENODE hn
, uintptr_t u
) {
	return (HTREEDB_TREENODE)AVL_TMPL_pack_node
	( AVL_TMPL_NS(node_at_index)(runtime, AVL_TMPL_unpack_node(hn), u)
	);
}
static PVOID STDCALL AVL_TMPL_NS(PointerNodeAtIndex)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREENODE hn
, uintptr_t u
) {
	return
		AVL_TMPL_NS(node_at_index)(runtime, AVL_TMPL_unpack_node(hn), u);
}
static uintptr_t STDCALL AVL_TMPL_NS(NodeCount)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREENODE hn
) {
	if(!hn)
		return 0;
	return AVL_TMPL_NS(NodeCountPtr)(runtime, AVL_TMPL_unpack_node(hn));
}
/* Pointer ------------------------------------------------------------------ */
static void STDCALL AVL_TMPL_NS(DumpNodePtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pn
, treedb_tree_renderer_ptr_t pr
, treedb_tree_data_renderer_ptr_t pd
) {
	if(!pn)
		return;
#if META_TREEDB_DEBUG
	HTREEDB_TREENODE hn = (HTREEDB_TREENODE)
		AVL_TMPL_pack_node((AVL_TMPL_node_ptr_t)pn)
		+ runtime->m_pImpl->m_pAllocator->m_pImpl->m_pDescription
			->m_aligned_node_size
		+ runtime->m_pImpl->m_pAllocator->m_pImpl->m_pDescription
			->m_aligned_item_size;
	printf("%s : dumping %zu.\n", __PRETTY_FUNCTION__, hn);
#endif
	AVL_TMPL_node_ptr_t node = (AVL_TMPL_node_ptr_t)pn;

	PSTR s = treedb_tree_data_renderer_t_Dump(pd, runtime
		, treedb_tree_t_NodeToDataPtr(runtime, node));
	if(s) {
		treedb_tree_renderer_t_Add(pr, s, 1);
		free(s);
	}else{
		treedb_tree_renderer_t_Add(pr, "<no memory>", 1);
	}
	treedb_tree_renderer_t_Push(pr);
	if(node->m_left) {
		treedb_tree_renderer_t_Add(pr, "Left", !node->m_right);
		treedb_tree_renderer_t_Push(pr);
		AVL_TMPL_NS(DumpNode)(runtime, (HTREEDB_TREENODE)(node->m_left), pr, pd);
		treedb_tree_renderer_t_Pop(pr);
	}
	if(node->m_right) {
		treedb_tree_renderer_t_Add(pr, "Right", 1);
		treedb_tree_renderer_t_Push(pr);
		AVL_TMPL_NS(DumpNode)(runtime, (HTREEDB_TREENODE)(node->m_right), pr, pd);
		treedb_tree_renderer_t_Pop(pr);
	}
	treedb_tree_renderer_t_Pop(pr);
}
static BOOL STDCALL AVL_TMPL_NS(InsertNodePtr)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PVOID pn
) {
	return AVL_TMPL_NS(insert_node)
	( runtime
	, AVL_TMPL_unpack_tree((AVL_TMPL_packed_tree_ptr_t)ht)
	, (AVL_TMPL_node_ptr_t)pn
	) != 0;
}
static void STDCALL AVL_TMPL_NS(InsertNodeBeforePtr)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PVOID pn
, PVOID pnew_node
) {
	AVL_TMPL_NS(insert_node_before)
	( runtime
	, AVL_TMPL_unpack_tree((AVL_TMPL_packed_tree_ptr_t)ht)
	, (AVL_TMPL_node_ptr_t)pn
	, (AVL_TMPL_node_ptr_t)pnew_node
	);
}
static void STDCALL AVL_TMPL_NS(InsertNodeAfterPtr)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PVOID pn
, PVOID pnew_node
) {
	AVL_TMPL_NS(insert_node_after)
	( runtime
	, AVL_TMPL_unpack_tree((AVL_TMPL_packed_tree_ptr_t)ht)
	, (AVL_TMPL_node_ptr_t)pn
	, (AVL_TMPL_node_ptr_t)pnew_node
	);
}
static void STDCALL AVL_TMPL_NS(RelinkNodePtr)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PVOID pn
) {
	AVL_TMPL_NS(relink_node)
	( runtime
	, AVL_TMPL_unpack_tree((AVL_TMPL_packed_tree_ptr_t)ht)
	, (AVL_TMPL_node_ptr_t)pn
	);
}
static void STDCALL AVL_TMPL_NS(UnlinkNodePtr)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PVOID pn
) {
	AVL_TMPL_NS(unlink_node)
	( runtime
	, AVL_TMPL_unpack_tree((AVL_TMPL_packed_tree_ptr_t)ht)
	, (AVL_TMPL_node_ptr_t)pn
	);
}
static void STDCALL AVL_TMPL_NS(DeleteNodePtr)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PVOID pn
) {
	AVL_TMPL_NS(delete_node)
	( runtime
	, (AVL_TMPL_packed_tree_ptr_t)ht
	, (AVL_TMPL_node_ptr_t)pn
	);
}
static TreedbSearchResult STDCALL AVL_TMPL_NS(SearchClosestPtr)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PCVOID key
, PVOID * ppn
) {
	TreedbSearchResult res = AVL_TMPL_NS(search_closest)
		(runtime, AVL_TMPL_unpack_tree(ht), key, (AVL_TMPL_node_ptr_t *)ppn);
	return res;
}
static TreedbSearchResult STDCALL AVL_TMPL_NS(SearchLePtr)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PCVOID key
, PVOID * ppn
) {
	TreedbSearchResult res = AVL_TMPL_NS(search_le)
		(runtime, AVL_TMPL_unpack_tree(ht), key, (AVL_TMPL_node_ptr_t *)ppn);
	return res;
}
static TreedbSearchResult STDCALL AVL_TMPL_NS(SearchGePtr)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PCVOID key
, PVOID * ppn
) {
	TreedbSearchResult res = AVL_TMPL_NS(search_ge)
	(runtime, AVL_TMPL_unpack_tree(ht), key, (AVL_TMPL_node_ptr_t *)ppn);
	return res;
}
static PVOID STDCALL AVL_TMPL_NS(PointerSearch)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PCVOID key
) {
	return
		AVL_TMPL_NS(search)
		( runtime
		, AVL_TMPL_unpack_tree(ht)
		, key
		);
}
static PVOID STDCALL AVL_TMPL_NS(PointerSearchLt)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PCVOID key
) {
	return
		AVL_TMPL_NS(search_lt)
		( runtime
		, AVL_TMPL_unpack_tree(ht)
		, key
		);
}
static PVOID STDCALL AVL_TMPL_NS(PointerSearchGt)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PCVOID key
) {
	return
		AVL_TMPL_NS(search_gt)
		( runtime
		, AVL_TMPL_unpack_tree(ht)
		, key
		);
}
static BOOL STDCALL AVL_TMPL_NS(NodeHasNextPtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pn
) {
	(void)runtime;
	return pn ? (((AVL_TMPL_node_ptr_t)pn)->m_next != 0) : 0;
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(NodeNextPtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pn
) {
	(void)runtime;
	return pn ? (HTREEDB_TREENODE)(((AVL_TMPL_node_ptr_t)pn)->m_next) : 0;
}
static PVOID STDCALL AVL_TMPL_NS(PointerNodeNextPtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pn
) {
	(void)runtime;
	return pn ?
		AVL_TMPL_unpack_node(((AVL_TMPL_node_ptr_t)pn)->m_next)
		: 0;
}
static BOOL STDCALL AVL_TMPL_NS(NodeHasPrevPtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pn
) {
	(void)runtime;
	return pn ? (((AVL_TMPL_node_ptr_t)pn)->m_prev != 0) : 0;
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(NodePrevPtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pn
) {
	(void)runtime;
	return pn ? (HTREEDB_TREENODE)(((AVL_TMPL_node_ptr_t)pn)->m_prev) : 0;
}
static PVOID STDCALL AVL_TMPL_NS(PointerNodePrevPtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pn
) {
	(void)runtime;
	return pn ?
		AVL_TMPL_unpack_node(((AVL_TMPL_node_ptr_t)pn)->m_prev)
		: 0;
}
static uintptr_t STDCALL AVL_TMPL_NS(NodeIndexPtr)
( treedb_tree_runtime_ptr_t runtime
, HTREEDB_TREE ht
, PCVOID pn
) {
	return
		AVL_TMPL_NS(node_index)(runtime, AVL_TMPL_unpack_tree(ht)
			, (AVL_TMPL_node_ptr_t)pn);
}
static HTREEDB_TREENODE STDCALL AVL_TMPL_NS(NodeAtIndexPtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pn
, uintptr_t u
) {
	return (HTREEDB_TREENODE)AVL_TMPL_pack_node
	( AVL_TMPL_NS(node_at_index)(runtime, (AVL_TMPL_node_ptr_t)pn, u)
	);
}
static PVOID STDCALL AVL_TMPL_NS(PointerNodeAtIndexPtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pn
, uintptr_t u
) {
	return AVL_TMPL_NS(node_at_index)(runtime, (AVL_TMPL_node_ptr_t)pn, u);
}
#ifdef AVL_TMPL_has_count
static uintptr_t STDCALL AVL_TMPL_NS(NodeCountPtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pv
) {
	(void)runtime;
	if(!pv)
		return 0;
	AVL_TMPL_node_ptr_t pn = (AVL_TMPL_node_ptr_t)pv;
	return pn->m_count;
}
#else
static uintptr_t STDCALL AVL_TMPL_NS(NodeCountPtr)
( treedb_tree_runtime_ptr_t runtime
, PCVOID pv
) {
	if(!pv)
		return 0;
	AVL_TMPL_node_ptr_t pn = (AVL_TMPL_node_ptr_t)pv;
	HTREEDB_TREENODE pl = (HTREEDB_TREENODE)(pn->m_left);
	HTREEDB_TREENODE pr = (HTREEDB_TREENODE)(pn->m_right);
	uintptr_t n = 1;
	if(pl)
		n += AVL_TMPL_NS(NodeCount)(runtime, pl);
	if(pr)
		n += AVL_TMPL_NS(NodeCount)(runtime, pr);
	META_TREEDB_printf("%s [%zu] returning %zu.\n"
		, __PRETTY_FUNCTION__
		, treedb_allocator_t_PointerToOffset(runtime->m_pImpl->m_pAllocator
			, treedb_tree_t_NodeToDataPtr(runtime, pn))
		, n);
	return n;
}
#endif


#if 0
#define ve(x) v->x = & AVL_TMPL_NS(x)
#define xve(x) v->x = 0

static void AVL_TMPL_NS(test)(treedb_tree_impl_vtbl_t * v)
{
	ve(Clear);
	ve(Clone);
	ve(Alloc);
	ve(Free);
	ve(Count);
	ve(Insert);
	ve(Remove);
	ve(RemoveBetween);
	ve(RemoveMany);
	ve(NodeDistance);
	ve(KeyDistance);
	ve(Dump);
	ve(Head);
	ve(Tail);
	ve(Top);
	ve(Index);
	ve(PointerHead);
	ve(PointerTail);
	ve(PointerTop);
	ve(PointerIndex);

	ve(Init);
	ve(DumpNode);
	ve(InsertNode);
	ve(InsertNodeBefore);
	ve(InsertNodeAfter);
	ve(UnlinkNode);
	ve(DeleteNode);

	ve(SearchClosest);
	ve(SearchLe);
	ve(SearchGe);
	ve(Search);
	ve(SearchLt);
	ve(SearchGt);
	ve(NodeHasNext);
	ve(NodeNext);
	ve(PointerNodeNext);
	ve(NodeHasPrev);
	ve(NodePrev);
	ve(PointerNodePrev);
	ve(NodeIndex);
	ve(NodeAtIndex);
	ve(PointerNodeAtIndex);
	ve(NodeCount);

	ve(InitPtr);
	ve(DumpNodePtr);
	ve(InsertNodePtr);
	ve(InsertNodeBeforePtr);
	ve(InsertNodeAfterPtr);
	ve(UnlinkNodePtr);
	ve(DeleteNodePtr);

	ve(SearchClosestPtr);
	ve(SearchLePtr);
	ve(SearchGePtr);
	ve(PointerSearch);
	ve(PointerSearchLt);
	ve(PointerSearchGt);
	ve(NodeHasNextPtr);
	ve(NodeNextPtr);
	ve(PointerNodeNextPtr);
	ve(NodeHasPrevPtr);
	ve(NodePrevPtr);
	ve(PointerNodePrevPtr);
	ve(NodeIndexPtr);
	ve(NodeAtIndexPtr);
	ve(PointerNodeAtIndexPtr);
	ve(NodeCountPtr);
}
#endif
const treedb_tree_impl_vtbl_t AVL_TMPL_NS(vtbl) = {
	V3C_INVOKE1(treedb_tree_impl_vtbl_t_entries,AVL_TMPL_NS_)
};
