/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include "meta-avl.h"
#include <stddef.h> /* for offsetof() */

void treedb_tree_t_Setup
( treedb_tree_ptr_t p
, treedb_tree_impl_ptr_t pImpl
, size_t uPayload
) {
	p->m_pImpl = pImpl;
	p->m_handle = uPayload;
}
void treedb_tree_impl_t_Setup
( treedb_tree_impl_ptr_t pm
, treedb_allocator_ptr_t pa
, treedb_tree_backend_ptr_t pb
, TreedbItemDescription * pd
, treedb_const_tree_impl_ptr_t ptemplate
) {
	pm->m_pVtbl = ptemplate->m_pVtbl;
	pm->m_pAllocator = pa;
	pm->m_pBackend = pb;
	pm->m_pDescription = pd;
	*pd = *(ptemplate->m_pDescription);
	pd->m_aligned_node_size = V3C_ROUND_UP_T(uint16_t, pd->m_aligned_node_size
		, pa->m_pImpl->m_pDescription->m_node.m_uAlign);
}
void treedb_tree_backend_t_Setup
( treedb_tree_backend_ptr_t pb
, PVOID pVtbl
) {
	*((treedb_tree_backend_vtbl_ptr_t *)(& pb->m_pVtbl))
		= (treedb_tree_backend_vtbl_ptr_t)pVtbl;
}

#define AVL_TMPL_prefix static inline
#define AVL_TMPL_NS(x) V3C_INVOKE2(V3C_JOIN2,AVL_TMPL_NS_,x)

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-defines.h"

#ifdef DOXYGEN_processing
}
#endif

#define AVL_TMPL_size_t uintptr_t
#define AVL_TMPL_ssize_t intptr_t

#define AVL_TMPL_packed_node_ptr_t AVL_TMPL_count_t
#define AVL_TMPL_packed_tree_ptr_t AVL_TMPL_count_t

#define AVL_TMPL_allow_duplicates treedb_tree_t_allows_duplicates(runtime)
#define AVL_TMPL_cmp(a,b) \
	treedb_tree_backend_t_Compare(runtime->m_pImpl->m_pBackend, runtime, a, b)
#define AVL_TMPL_unpack_key(pNode) \
	treedb_tree_backend_t_UnpackKey(runtime->m_pImpl->m_pBackend, runtime \
		, pNode)

#define AVL_TMPL_alloc_node(x) \
	(AVL_TMPL_NS(node_ptr_t))(treedb_tree_backend_t_AllocNode \
	(runtime->m_pImpl->m_pBackend, runtime, x))
#define AVL_TMPL_free_node(x) \
	AVL_TMPL_free_node_pointer(x)

#define AVL_TMPL_free_packed_node(x) \
	treedb_tree_backend_t_FreeNode(runtime->m_pImpl->m_pBackend, runtime \
	, AVL_TMPL_unpack_node(x))
#define AVL_TMPL_free_node_pointer(x) \
	treedb_tree_backend_t_FreeNode(runtime->m_pImpl->m_pBackend, runtime, x)
#define AVL_TMPL_free(x) \
	treedb_allocator_t_FreePtr(runtime->m_pImpl->m_pAllocator, x)
/* This can't be PointerPack - we might be in a pointer allocator. */
#define AVL_TMPL_pack_pointer(x) \
	treedb_tree_t_PointerToOffset(runtime, x)
#define AVL_TMPL_unpack_pointer(x) \
	treedb_tree_t_OffsetToPointer(runtime, x)

#define AVL_TMPL_runtime_ptr_t treedb_tree_runtime_ptr_t

#ifdef __cplusplus
extern "C" {
#endif

/******************************************************************************/
/* o16                                                                        */
/******************************************************************************/
#define AVL_TMPL_count_t uint16_t
#define AVL_TMPL_PACK(x) treedb_pack_16(x)

/* count */
#define AVL_TMPL_has_count
#define AVL_TMPL_NS_ treedb_avl_tree_o16c_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_NS_

/* depth */
#undef AVL_TMPL_has_count
#define AVL_TMPL_has_depth
#define AVL_TMPL_NS_ treedb_avl_tree_o16d_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_NS_

/* count + depth */
#define AVL_TMPL_has_count
#define AVL_TMPL_NS_ treedb_avl_tree_o16cd_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_NS_

#undef AVL_TMPL_has_count
#undef AVL_TMPL_has_depth

/******************************************************************************/
/* o32                                                                        */
/******************************************************************************/
/* 16->32 */
#undef AVL_TMPL_count_t
#define AVL_TMPL_count_t uint32_t

/* count */
#define AVL_TMPL_has_count
#define AVL_TMPL_NS_ treedb_avl_tree_o32c_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_NS_

/* depth */
#undef AVL_TMPL_has_count
#define AVL_TMPL_has_depth
#define AVL_TMPL_NS_ treedb_avl_tree_o32d_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_NS_

/* count + depth */
#define AVL_TMPL_has_count
#define AVL_TMPL_NS_ treedb_avl_tree_o32cd_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_NS_

#undef AVL_TMPL_has_count
#undef AVL_TMPL_has_depth

#if HAVE_64_BITS
/******************************************************************************/
/* o64                                                                        */
/******************************************************************************/
/* 32->64 */
#undef AVL_TMPL_count_t
#define AVL_TMPL_count_t uint64_t

/* count */
#define AVL_TMPL_has_count
#define AVL_TMPL_NS_ treedb_avl_tree_o64c_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_NS_

/* depth */
#undef AVL_TMPL_has_count
#define AVL_TMPL_has_depth
#define AVL_TMPL_NS_ treedb_avl_tree_o64d_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_NS_

/* count + depth */
#define AVL_TMPL_has_count
#define AVL_TMPL_NS_ treedb_avl_tree_o64cd_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_NS_

#undef AVL_TMPL_has_count
#undef AVL_TMPL_has_depth

#endif /* #if HAVE_64_BITS */

/******************************************************************************/
/* pointer                                                                    */
/******************************************************************************/
/* 32/64 -> pointer */
#undef AVL_TMPL_count_t
#define AVL_TMPL_count_t uintptr_t

#define AVL_POINTER_TREE
#define v3c_nbits_u0(x) v3c_nbits_up(x)

#undef AVL_TMPL_packed_tree_ptr_t
#undef AVL_TMPL_packed_node_ptr_t
#define AVL_TMPL_packed_tree_ptr_t AVL_TMPL_tree_ptr_t
#define AVL_TMPL_packed_node_ptr_t AVL_TMPL_node_ptr_t

#undef AVL_TMPL_unpack_tree
#undef AVL_TMPL_pack_tree
#undef AVL_TMPL_unpack_node
#undef AVL_TMPL_pack_node
#define AVL_TMPL_unpack_tree(x) ((AVL_TMPL_NS(tree_ptr_t))(x))
#define AVL_TMPL_pack_tree(x) (x)
#define AVL_TMPL_unpack_node(x) ((AVL_TMPL_NS(node_ptr_t))(x))
#define AVL_TMPL_pack_node(x) (x)

#undef AVL_TMPL_unpack_pointer
#undef AVL_TMPL_pack_pointer
#define AVL_TMPL_unpack_pointer(x) ((PVOID)(x))
#define AVL_TMPL_pack_pointer(x) (x)

#define AVL_TMPL_omit_pack_tree
#define AVL_TMPL_omit_pack_node
#define AVL_TMPL_omit_unpack_tree
#define AVL_TMPL_omit_unpack_node

/* count */
#define AVL_TMPL_has_count
#define AVL_TMPL_NS_ treedb_avl_tree_pc_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_NS_

/* depth */
#undef AVL_TMPL_has_count
#define AVL_TMPL_has_depth
#define AVL_TMPL_NS_ treedb_avl_tree_pd_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_NS_

/* count + depth */
#define AVL_TMPL_has_count
#define AVL_TMPL_NS_ treedb_avl_tree_pcd_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-avl-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef AVL_TMPL_count_t

#undef AVL_TMPL_has_count
#undef AVL_TMPL_has_depth

#undef AVL_TMPL_packed_node_ptr_t
#undef AVL_TMPL_packed_tree_ptr_t

#undef AVL_TMPL_unpack_tree
#undef AVL_TMPL_unpack_node
#undef AVL_TMPL_pack_tree
#undef AVL_TMPL_pack_node

#undef AVL_TMPL_omit_pack_tree
#undef AVL_TMPL_omit_pack_node
#undef AVL_TMPL_omit_unpack_tree
#undef AVL_TMPL_omit_unpack_node

#undef AVL_TMPL_unpack_pointer
#undef AVL_TMPL_pack_pointer

#ifdef __cplusplus
} /* extern "C" { */
#endif
