/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include "meta-treedb.h"
#include <stddef.h> /* for offsetof() */

void treedb_list_t_Setup
( treedb_list_ptr_t p
, treedb_list_impl_ptr_t pImpl
, size_t uPayload
) {
	p->m_pImpl = pImpl;
	p->m_handle = uPayload;
}
void treedb_list_impl_t_Setup
( treedb_list_impl_ptr_t pm
, treedb_allocator_on_storage_ptr_t pa
, treedb_list_backend_ptr_t pb
, TreedbItemDescription * pd
, treedb_const_list_impl_ptr_t ptemplate
) {
	pm->m_pVtbl = ptemplate->m_pVtbl;
	pm->m_pAllocator = (treedb_allocator_ptr_t)pa;
	pm->m_pBackend = pb;
	*pd = *(ptemplate->m_pDescription);
	pm->m_pDescription = pd;
	pd->m_aligned_node_size = V3C_ROUND_UP_T(uint16_t, pd->m_aligned_node_size
		, pa->m_pImpl->m_pDescription->m_pointer.m_uAlign);
}
void treedb_list_backend_t_Setup
( treedb_list_backend_ptr_t pb
, PVOID pVtbl
) {
	*((treedb_list_backend_vtbl_ptr_t *)(& pb->m_pVtbl))
		= (treedb_list_backend_vtbl_ptr_t)pVtbl;
}

#ifdef DOXYGEN_processing
{
#endif

#include "meta-list-defines.h"

#ifdef DOXYGEN_processing
}
#endif

#define LIST_TMPL_prefix static inline
#define LIST_TMPL_NS(x) V3C_INVOKE2(V3C_JOIN2,LIST_TMPL_NS_,x)
#define LIST_TMPL_input_t HTREEDB_LISTDATA

#define LIST_TMPL_size_t uintptr_t
#define treedb_List(This) \
	((LIST_TMPL_NS(list_ptr_t))(treedb_list_OffsetToPointer(This \
		, This->m_list_offset)))

#define LIST_TMPL_packed_node_ptr_t LIST_TMPL_count_t
#define LIST_TMPL_packed_list_ptr_t LIST_TMPL_count_t

#define LIST_TMPL_alloc_node(x) \
	(LIST_TMPL_NS(node_ptr_t))(treedb_list_backend_t_AllocNode \
	(runtime->m_pImpl->m_pBackend, runtime, x))
#define LIST_TMPL_free_node(x) \
	LIST_TMPL_FREE_NODE_POINTER(x)

#define LIST_TMPL_free_packed_node(x) \
	treedb_list_backend_t_FreeNode(runtime->m_pImpl->m_pBackend, runtime \
	, LIST_TMPL_unpack_node(x))
#define LIST_TMPL_free_node_pointer(x) \
	treedb_list_backend_t_FreeNode(runtime->m_pImpl->m_pBackend, runtime, x)
#define LIST_TMPL_free(x) \
	treedb_allocator_t_Free(runtime->m_pImpl->m_pAllocator, x)
#define LIST_TMPL_pack_pointer(x) treedb_list_t_PointerToOffset(runtime, x)
#define LIST_TMPL_unpack_pointer(x) treedb_list_t_OffsetToPointer(runtime, x)

#ifdef __cplusplus
extern "C" {
#endif

/******************************************************************************/
/* o16                                                                        */
/******************************************************************************/
/* 16 */
#define LIST_TMPL_count_t uint16_t

/* */
#define LIST_TMPL_NS_ treedb_list_o16_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-list-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef LIST_TMPL_NS_

/* count */
#define LIST_TMPL_have_count
#define LIST_TMPL_NS_ treedb_list_o16c_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-list-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef LIST_TMPL_have_count
#undef LIST_TMPL_NS_

/******************************************************************************/
/* o32                                                                        */
/******************************************************************************/
/* 16->32 */
#undef LIST_TMPL_count_t
#define LIST_TMPL_count_t uint32_t

/* */
#define LIST_TMPL_NS_ treedb_list_o32_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-list-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef LIST_TMPL_NS_

/* count */
#define LIST_TMPL_have_count
#define LIST_TMPL_NS_ treedb_list_o32c_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-list-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef LIST_TMPL_have_count
#undef LIST_TMPL_NS_

#if HAVE_64_BITS
/******************************************************************************/
/* o64                                                                        */
/******************************************************************************/
/* 32->64 */
#undef LIST_TMPL_count_t
#define LIST_TMPL_count_t uint64_t

/* */
#define LIST_TMPL_NS_ treedb_list_o64_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-list-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef LIST_TMPL_NS_

/* count */
#define LIST_TMPL_have_count
#define LIST_TMPL_NS_ treedb_list_o64c_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-list-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef LIST_TMPL_have_count
#undef LIST_TMPL_NS_

#endif /* #if HAVE_64_BITS */

/******************************************************************************/
/* pointer                                                                    */
/******************************************************************************/
/* 64->pointer */
#undef LIST_TMPL_count_t
#define LIST_TMPL_count_t uintptr_t

#undef LIST_TMPL_packed_node_ptr_t
#undef LIST_TMPL_packed_list_ptr_t
#define LIST_TMPL_packed_node_ptr_t LIST_TMPL_NS(node_ptr_t)
#define LIST_TMPL_packed_list_ptr_t LIST_TMPL_NS(list_ptr_t)

#undef LIST_TMPL_unpack_list
#undef LIST_TMPL_pack_list
#undef LIST_TMPL_unpack_node
#undef LIST_TMPL_pack_node

#define LIST_TMPL_omit_unpack_list
#define LIST_TMPL_omit_pack_list
#define LIST_TMPL_omit_unpack_node
#define LIST_TMPL_omit_pack_node

#define LIST_TMPL_unpack_list(x) ((LIST_TMPL_NS(list_ptr_t))(x))
#define LIST_TMPL_pack_list(x) ((HTREEDB_LIST)(x))
#define LIST_TMPL_unpack_node(x) ((LIST_TMPL_NS(node_ptr_t))(x))
#define LIST_TMPL_pack_node(x) ((HTREEDB_LISTNODE)(x))

#define LIST_TMPL_POINTER

#undef LIST_TMPL_unpack_pointer
#undef LIST_TMPL_pack_pointer
#define LIST_TMPL_unpack_pointer(x) ((PVOID)(x))
#define LIST_TMPL_pack_pointer(x) (x)

#define TREEDB_omit_list_unpack_list
#define TREEDB_omit_list_unpack_node
#define TREEDB_omit_list_pack_list
#define TREEDB_omit_list_pack_node

/* */
#define LIST_TMPL_NS_ treedb_list_p_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-list-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef LIST_TMPL_NS_

/* count */
#define LIST_TMPL_have_count
#define LIST_TMPL_NS_ treedb_list_pc_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-list-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef LIST_TMPL_have_count
#undef LIST_TMPL_NS_

#undef LIST_TMPL_count_t
#undef LIST_TMPL_POINTER

#undef LIST_TMPL_unpack_list
#undef LIST_TMPL_pack_list
#undef LIST_TMPL_unpack_node
#undef LIST_TMPL_pack_node

#undef LIST_TMPL_POINTER

#undef LIST_TMPL_unpack_pointer
#undef LIST_TMPL_pack_pointer

#undef LIST_TMPL_packed_node_ptr_t
#undef LIST_TMPL_packed_list_ptr_t

extern treedb_const_list_ptr_t
treedb_get_list_template_for
( int16_t uBytesPerOffset /* 2, 4, 8 */
, uint16_t flags
) {
	if(flags & eTreedbContainerIsPointer) {
		switch(flags & ~((uint16_t)eTreedbContainerIsPointer)) {
		case 0:
			return & treedb_list_p_list;
		case eTreedbContainerHasCount:
			return & treedb_list_pc_list;
		default:
			break;
		}
		return 0;
	}
	switch(uBytesPerOffset) {
	case 2:
		switch(flags) {
		case 0:
			return & treedb_list_o16_list;
		case eTreedbContainerHasCount:
			return & treedb_list_o16c_list;
		default:
			break;
		}
		break;
	case 4:
		switch(flags) {
		case 0:
			return & treedb_list_o32_list;
		case eTreedbContainerHasCount:
			return & treedb_list_o32c_list;
		default:
			break;
		}
		break;
#if HAVE_64_BITS
	case 8:
		switch(flags) {
		case 0:
			return & treedb_list_o64_list;
		case eTreedbContainerHasCount:
			return & treedb_list_o64c_list;
		default:
			break;
		}
		break;
#endif
	default:
		break;
	}
	return 0;
}

extern treedb_const_list_ptr_t
treedb_get_list_template_using_allocator
( treedb_allocator_ptr_t pallocator
) {
	if(!pallocator)
		return 0;
	int flags = 0;
	if(pallocator->m_pImpl->m_pDescription->m_flags & eTreedbContainerHasCount)
		flags |= eTreedbContainerHasCount;
	if(pallocator->m_pImpl->m_pDescription->m_flags & eTreedbContainerIsPointer)
		flags |= eTreedbContainerIsPointer;

	treedb_const_list_ptr_t pt
		= treedb_get_list_template_for
			( pallocator->m_pImpl->m_pDescription->m_pointer.m_uSize
			, flags);
	return pt;
}

#ifdef __cplusplus
} /* extern "C" { */
#endif
