/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef v3c_meta_treedb_allocator_hpp_included
#define v3c_meta_treedb_allocator_hpp_included

#ifndef v3c_meta_treedb_allocator_h_included
#include "meta-treedb-allocator.h"
#endif

#ifndef v3c_meta_treedb_meta_avl_h_included
#include "meta-avl.h"
#endif

namespace v3c {
namespace treedb {
namespace meta {

#undef I3dbCLASS
#undef I3dbCLASS0
#define I3dbCLASS(ClassPtr)
#define I3dbCLASS0(ClassPtr)

/*! treedb_allocator_impl_t vtable-only overlay
*/
class IAllocatorImpl_VtableOverlay
{
public:
	treedb_allocator_impl_t_methods
	(,, treedb_allocator_impl_t,);
protected:
	//! This can't be virtual. If it were it would mess up derived interfaces.
	~IAllocatorImpl_VtableOverlay() {}
};

#undef I3dbCLASS
#undef I3dbCLASS0
#define I3dbCLASS(AllocatorPtr) AllocatorPtr,
#define I3dbCLASS0(AllocatorPtr) AllocatorPtr

template< typename Interface = IAllocatorImpl_VtableOverlay >
class IAllocatorImplT : public Interface
{
public:
	inline treedb_allocator_impl_ptr_t base_c_type_ptr()
	{ return (treedb_allocator_impl_ptr_t)this; }
	inline treedb_allocator_impl_const_ptr_t base_c_type_ptr() const
	{ return (treedb_allocator_impl_const_ptr_t)this; }

	inline treedb_allocator_impl_t & base_c_type()
	{ return *base_c_type_ptr(); }
	inline const treedb_allocator_impl_t & base_c_type() const
	{ return *base_c_type_ptr(); }

	const TreedbAllocatorDescription & description() const
	{ return * m_pDescription; }
	TreedbAllocator_impl_members;
};
template< typename Interface = IAllocatorImpl_VtableOverlay >
class ITwoTreeAllocatorImplT
: public IAllocatorImplT< Interface >
{
};

template< typename Interface = IAllocatorImpl_VtableOverlay >
class IAllocatorOnStorageImplT
: public ITwoTreeAllocatorImplT< Interface >
{
public:
	inline treedb_allocator_on_storage_impl_ptr_t & c_type()
	{ return *(treedb_allocator_on_storage_impl_ptr_t *)this; }
	inline const treedb_allocator_on_storage_impl_ptr_t & c_type() const
	{ return *(treedb_allocator_on_storage_impl_ptr_t *)this; }

	Storage * m_pStorage;
};

////////////////////////////////////////////////////////////////////////////////
// Wrappers
////////////////////////////////////////////////////////////////////////////////
#define DGTHIS0(x) (x)->m_pImpl
#define DGTHIS(x) (x)->m_pImpl,
#define DGTHIS_MEM(x,m) (x)->m
#define V3C_DELEGATE(pInst, x) (pInst)->m_pImpl->m_pVtbl->x

#define A
#define A0
#define Ax & m_inst,
#define A0x0 , & m_inst
#define Ax0 & m_inst

// Default argument defined in meta-treedb.hpp as treedb_allocator_t
template< typename Instance >
class AllocatorT
{
public:
	inline treedb_allocator_t & base_c_type()
	{
		// g++ way of abiding with strict aliasing rules.
		union { Instance * i; treedb_allocator_t * j; } u;
		u.i = & m_inst;
		return * u.j;
	}
	inline const treedb_allocator_t & base_c_type() const
	{
		// g++ way of abiding with strict aliasing rules.
		union { Instance * i; treedb_allocator_t * j; } u;
		u.i = & m_inst;
		return * u.j;
	}
	inline Instance & c_type()
	{ return m_inst; }
	inline const Instance & c_type() const
	{ return m_inst; }

	const TreedbAllocatorDescription & description() const
	{ return * m_inst.m_pImpl->m_pDescription; }
	int error() const { return m_inst.m_errno; }
	TREEDB_Allocator_DELEGATE_TO(& m_inst);

	inline operator Instance &() { return c_type(); }
	inline operator const Instance &() const { return c_type(); }

	Instance m_inst;
};

// Default argument defined in meta-treedb.hpp as treedb_two_tree_allocator_t
template< typename Instance >
class TwoTreeAllocatorT
: public AllocatorT< Instance >
{
public:
	typedef AllocatorT< Instance > Base;
	inline operator Base & () { return *this; }
	inline operator const Base & () const { return *this; }
	treedb_tree_t & used_tree() { return Base::m_inst.m_used; }
	treedb_tree_t & free_tree() { return Base::m_inst.m_free; }
};
typedef TwoTreeAllocatorT<> TwoTreeAllocator;

// Default argument defined in meta-treedb.hpp as treedb_allocator_on_storage_t
template< typename Instance >
class AllocatorOnStorageT
: public TwoTreeAllocatorT< Instance >
{
public:
	typedef TwoTreeAllocatorT< Instance > Base;
	inline operator Base & () { return *this; }
	inline operator const Base & () const { return *this; }
	bool init_runtime
	( treedb_allocator_on_storage_impl_ptr_t pAllocatorImpl
	, uint16_t uBytesPerOffset /* 2, 4, 8 */
	, uint16_t flags, uintptr_t offset
	, treedb_storage_ptr_t pStorage
	) {
		return treedb_init_allocator_on_storage_from_template
		( & Base::c_type(), pAllocatorImpl, uBytesPerOffset, flags, offset
		, pStorage);
	}

	inline operator Instance &() { return this->c_type(); }
	inline operator const Instance &() const { return this->c_type(); }
};
#undef A
#undef A0
#undef Ax
#undef A0x0
#undef Ax0

#undef DGTHIS0
#undef DGTHIS
#undef DGTHIS_MEM
#undef V3C_DELEGATE

#define A Allocator * pAllocator,
#define A0 Allocator * pAllocator
#define Ax pAllocator,
#define A0x0 , pAllocator
#define Ax0 pAllocator

#define DGTHIS0(x) (x)
#define DGTHIS(x) (x),
#define DGTHIS_MEM(x,m) (x)->m
#define V3C_DELEGATE(pInst, x) (pInst)->m_pVtbl->x

template< typename Instance = treedb_allocator_impl_t >
class AllocatorImplT
{
public:
	inline treedb_allocator_impl_t & base_c_type()
	{ return *(treedb_allocator_impl_t *)(& m_inst); }
	inline const treedb_allocator_impl_t & base_c_type() const
	{ return *(const treedb_allocator_impl_t *)(& m_inst); }

	inline Instance & c_type()
	{ return *(Instance *)(& m_inst); }
	inline const Instance & c_type() const
	{ return *(Instance *)(& m_inst); }

	const TreedbAllocatorDescription & description() const
	{ return * m_inst.m_pDescription; }
	int errno() const { return m_inst.m_errno; }
	TREEDB_Allocator_DELEGATE_TO(& m_inst);
	Instance m_inst;
};

template< typename Instance = treedb_two_tree_allocator_impl_t>
class TwoTreeAllocatorImplT
: public AllocatorImplT< Instance >
{
};
typedef TwoTreeAllocatorImplT<> TwoTreeAllocatorImpl;

template< typename Instance = treedb_allocator_on_storage_impl_t >
class AllocatorOnStorageImplT
: public TwoTreeAllocatorImplT< Instance >
{
};

typedef AllocatorImplT<> AllocatorImpl;
typedef TwoTreeAllocatorImplT<> TwoTreeAllocatorImpl;
typedef AllocatorOnStorageImplT<> AllocatorOnStorageImpl;

#undef A
#undef A0
#undef Ax
#undef A0x0

#undef DGTHIS0
#undef DGTHIS
#undef DGTHIS_MEM
#undef V3C_DELEGATE

} // namespace meta
} // namespace treedb
} // namespace v3c

#endif /* #ifndef v3c_meta_treedb_allocator_hpp_included */
