/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include "meta-treedb-array.h"
#include <memory.h>

void treedb_array_t_Setup
( treedb_array_ptr_t p
, treedb_array_impl_ptr_t pImpl
, uintptr_t uPayload
) {
	p->m_pImpl = pImpl;
	p->m_handle = uPayload;
}
void treedb_array_impl_t_Setup
( treedb_array_impl_ptr_t pm
, treedb_allocator_on_storage_ptr_t pa
, treedb_array_backend_ptr_t pb
, TreedbArrayDescription * pd
, treedb_const_array_impl_ptr_t ptemplate
) {
	pm->m_pVtbl = ptemplate->m_pVtbl;
	pm->m_pAllocator = (treedb_allocator_ptr_t)pa;
	pm->m_pBackend = pb;
	*pd = *(ptemplate->m_pBackend->m_pDescription);
	pb->m_pDescription = pd;
}
void treedb_array_backend_t_Setup
( treedb_array_backend_ptr_t pb
, PVOID pVtbl
, TreedbArrayDescriptionConstPtr pDescription
) {
	*((treedb_array_backend_vtbl_ptr_t *)(& pb->m_pVtbl))
		= (treedb_array_backend_vtbl_ptr_t)pVtbl;
	pb->m_pDescription = pDescription;
}

#include <stddef.h> /* for offsetof() */
#include <v3c/comdefs.h> /* for PSTR */

#define ARRAY_TMPL_runtime_ptr_t treedb_array_runtime_ptr_t
#define ARRAY_TMPL_prefix static inline
#define ARRAY_TMPL_NS(x) V3C_INVOKE2(V3C_JOIN2,ARRAY_TMPL_NS_,x)

#ifdef DOXYGEN_processing
{
#endif

#include "meta-treedb-array-defines.h"

#ifdef DOXYGEN_processing
}
#endif

static PVOID impl_alloc
( treedb_array_runtime_ptr_t runtime
, size_t size
) {
	PVOID p = treedb_allocator_t_PointerAlloc(runtime->m_pImpl->m_pAllocator
		, size);
	META_TREEDB_printf("treedb_array_impl_alloc : allocated %zu bytes at "
		"%zu.\n"
		, treedb_allocator_t_GetSizePtr(runtime->m_pImpl->m_pAllocator, p)
		, treedb_array_impl_t_PointerToOffset(runtime->m_pImpl, p)
	);
	return p;
}
static void impl_memmove
( treedb_allocator_ptr_t pa
, void * to
, const void * from
, size_t size)
{
	(void)pa;
	META_TREEDB_printf("### memmove(%zu, %zu, %zu) ###\n"
		, (size_t)treedb_allocator_t_PointerToOffset(pa, to)
		, (size_t)treedb_allocator_t_PointerToOffset(pa, from)
		, size
	);
	memmove(to, from, size);
}
#define ARRAY_TMPL_IS_POINTER 0

#ifdef __cplusplus
extern "C" {
#endif

/******************************************************************************/
/* 16                                                                        */
/******************************************************************************/
#define ARRAY_TMPL_count_t uint16_t
#define ARRAY_TMPL_data_t char
#define ARRAY_TMPL_data_ptr_t PSTR
#define ARRAY_TMPL_input_t PCVOID

#define ARRAY_TMPL_NS_ treedb_array_o16_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-treedb-array-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef ARRAY_TMPL_NS_

#include <treedb/array-template-cleardefs.h>

/******************************************************************************/
/* 32                                                                        */
/******************************************************************************/
#define ARRAY_TMPL_count_t uint32_t
#define ARRAY_TMPL_data_t char
#define ARRAY_TMPL_data_ptr_t PSTR

#define ARRAY_TMPL_NS_ treedb_array_o32_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-treedb-array-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef ARRAY_TMPL_NS_

#include <treedb/array-template-cleardefs.h>

#if HAVE_64_BITS
/******************************************************************************/
/* 64                                                                        */
/******************************************************************************/
#define ARRAY_TMPL_count_t uint64_t
#define ARRAY_TMPL_data_t char
#define ARRAY_TMPL_data_ptr_t PSTR

#define ARRAY_TMPL_NS_ treedb_array_o64_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-treedb-array-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef ARRAY_TMPL_NS_

#include <treedb/array-template-cleardefs.h>

#endif /* #if HAVE_64_BITS */

/******************************************************************************/
/* pointer                                                                    */
/******************************************************************************/
#undef ARRAY_TMPL_IS_POINTER
#define ARRAY_TMPL_IS_POINTER eTreedbContainerIsPointer

#define ARRAY_TMPL_count_t uintptr_t
#define ARRAY_TMPL_data_t char
#define ARRAY_TMPL_data_ptr_t PSTR
#define ARRAY_TMPL_packed_data_ptr_t PSTR

#define ARRAY_TMPL_packed_array_ptr_t ARRAY_TMPL_array_ptr_t

#define ARRAY_TMPL_omit_unpack_array
#define ARRAY_TMPL_omit_pack_array
#define ARRAY_TMPL_omit_unpack_data
#define ARRAY_TMPL_omit_pack_data

#define ARRAY_TMPL_unpack_array(x) ((ARRAY_TMPL_array_ptr_t)x)
#define ARRAY_TMPL_pack_array(x) (x)
#define ARRAY_TMPL_unpack_data(x) ((ARRAY_TMPL_data_ptr_t)x)
#define ARRAY_TMPL_pack_data(x) (x)

#define ARRAY_TMPL_pack_pointer(x) (x)
#define ARRAY_TMPL_unpack_pointer(x) (x)

#define ARRAY_TMPL_NS_ treedb_array_p_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-treedb-array-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef ARRAY_TMPL_NS_

#include <treedb/array-template-cleardefs.h>

treedb_const_array_ptr_t
treedb_get_array_template_for
( uint16_t uBytes /* 2, 4, 8 */
, uint16_t flags /* flags & eTreedbContainerIsPointer means ignore uBytes */
) {
	if(flags & eTreedbContainerIsPointer) {
		return & treedb_array_p_instance;
	}
	switch(uBytes) {
	case 2:
		return & treedb_array_o16_instance;
	case 4:
		return & treedb_array_o32_instance;
#if HAVE_64_BITS
	case 8:
		return & treedb_array_o64_instance;
#endif
	default:
		break;
	}
	return 0;
}

treedb_const_array_ptr_t
treedb_get_array_template_using_allocator
( treedb_allocator_ptr_t pallocator
) {
	if(!pallocator)
		return 0;
	int flags = 0;
	if(pallocator->m_pImpl->m_pDescription->m_flags & eTreedbContainerIsPointer)
		flags |= eTreedbContainerIsPointer;
	treedb_const_array_ptr_t pt
		= treedb_get_array_template_for
			( pallocator->m_pImpl->m_pDescription->m_pointer.m_uSize
			, flags);
	return pt;
}

#ifdef __cplusplus
} /* extern "C" { */
#endif
