/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef v3c_meta_treedb_array_hpp_included
  #define v3c_meta_treedb_array_hpp_included

  #ifndef v3c_meta_treedb_array_h_included
	#define COBJMACROS
	#include "meta-treedb-array.h"
  #endif

  #ifndef v3c_meta_treedb_hpp_included
	#define COBJMACROS
	#include "meta-treedb.hpp"
  #endif

  #include <v3c/utils.h>

namespace v3c {
namespace treedb {
namespace meta {

class ArrayRuntime;
/*! treedb_array_impl_t vtable-only overlay
*/
class IArrayImpl_VtableOverlay
{
public:
#include "meta-treedb-cxx-push-c.h"
	treedb_array_impl_t_methods(,,ArrayRuntime);
#include "meta-treedb-cxx-pop-c.h"
protected:
	//! This can't be virtual. If it were it would mess up derived interfaces.
	~IArrayImpl_VtableOverlay() {}
};

/*! treedb_array_impl_t overlay
*/
template< typename Interface = IArrayImpl_VtableOverlay >
class ITreedbArrayImplT : public Interface
{
public:
	inline Allocator & allocator()
	{ return *(Allocator *)(m_pAllocator); }
	inline const TreedbArrayDescription & description() const
	{ return * m_pBackend->m_pDescription; }

	treedb_array_impl_t_members(treedb_array_impl_vtbl_ptr_t);
};

/*! treedb_array_backend_t overlay
*/
template< typename RuntimeType = ArrayRuntime >
class IArrayBackendT
{
public:
	treedb_array_backend_t_methods(,,,RuntimeType);
	TreedbArrayDescriptionConstPtr m_pDescription;
protected:
	//! OK this can be virtual.
	~IArrayBackendT() {}
};

////////////////////////////////////////////////////////////////////////////////
// Wrappers
////////////////////////////////////////////////////////////////////////////////
#define DGTHIS_MEM(x,m) x.m
#define DGTHIS0(x) treedb_array_t_to_runtime(& (x))
#define DGTHIS(x) treedb_array_t_to_runtime(& (x)),
#define V3C_DELEGATE(Inst, x) Inst.m_pImpl->m_pVtbl->x

class ArrayImpl;

/*! treedb_array_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class Array
{
public:
	/* Convenience accessors. */
	inline const treedb_array_impl_t & impl() const
	{ return * m_inst.m_pImpl; }
	inline const treedb_array_backend_t & backend() const
	{ return * m_inst.m_pImpl->m_pBackend; }
	inline const TreedbArrayDescription & description() const
	{ return * backend().m_pDescription; }
	inline Allocator & allocator()
	{ return *(Allocator *)(impl().m_pAllocator); }

	inline HTREEDB_ARRAY raw_handle() const { return m_inst.m_handle; }
	inline HTREEDB_ARRAY & raw_handle() { return m_inst.m_handle; }

	TREEDB_IArray_DELEGATE_TO(m_inst);
	treedb_array_t m_inst;
};

#undef V3C_DELEGATE
#undef DGTHIS
#undef DGTHIS0

/*! treedb_tree_runtime_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class ArrayRuntime
{
public:
	inline ArrayRuntime() { m_inst.m_pImpl = 0; }
	inline ArrayRuntime(ArrayImpl * p)
		{ m_inst.m_pImpl = (treedb_array_impl_ptr_t)p; }
	/* Convenience accessors. */
	inline const TreedbArrayDescription & description() const
	{ return * m_inst.m_pImpl->m_pBackend->m_pDescription; }
	inline Allocator & allocator()
	{ return *(Allocator *)(m_inst.m_pImpl->m_pAllocator); }

	treedb_array_runtime_t m_inst;
};

#define DGTHIS0(x) (&x)
#define DGTHIS(x) (&x),
#define V3C_DELEGATE(Inst, x) Inst.m_pVtbl->x

/*! treedb_array_impl_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class ArrayImpl
{
public:
	inline ArrayImpl() : m_inst() {}
	inline void setup
	( AllocatorOnStorage & allocator
	, treedb_array_backend_ptr_t pBackend
	, TreedbArrayDescription * pd
	, treedb_const_array_impl_ptr_t ptemplate
	) {
		treedb_array_impl_t_Setup(& m_inst, & allocator.c_type(), pBackend, pd
			, ptemplate);
	}
	inline ArrayImpl
	( AllocatorOnStorage & allocator
	, treedb_array_backend_ptr_t pBackend
	, TreedbArrayDescription * pd
	, treedb_const_array_impl_ptr_t ptemplate
	) {
		setup(allocator, pBackend, pd, ptemplate);
	}
	/* Convenience accessors. */
	inline const TreedbArrayDescription & description() const
	{ return * m_inst.m_pBackend->m_pDescription; }
	inline Allocator & allocator()
	{ return *(Allocator *)(m_inst.m_pAllocator); }

	TREEDB_IArrayImpl_DELEGATE_TO(m_inst);
	treedb_array_impl_t m_inst;
};

class ArrayBackend
{
public:
	TREEDB_IArrayBackend_DELEGATE_TO(m_inst);
	treedb_array_backend_t m_inst;
};

#undef V3C_DELEGATE
#undef DGTHIS0
#undef DGTHIS
#undef DGTHIS_MEM

////////////////////////////////////////////////////////////////////////////////
// Templates
////////////////////////////////////////////////////////////////////////////////
template< typename Impl >
class ArrayRuntimeT
: public ArrayRuntime
{
public:
	inline ArrayRuntimeT() : ArrayRuntime() {}
	inline ArrayRuntimeT(Impl * p) : ArrayRuntime((Impl *)p) {}
	inline Impl & impl() { return *(Impl *)m_inst.m_pImpl; }
};

template< typename Backend_ >
class ArrayImplT
: public ArrayImpl
{
public:
	typedef Backend_ Backend;
	typedef typename Backend::DataType DataType;
	inline void setup
	( AllocatorOnStorage & allocator
	, treedb_const_array_impl_ptr_t ptemplate
	) {
		ArrayImpl::setup(allocator, (treedb_array_backend_ptr_t)(& m_backend)
			, & m_backend.m_description, ptemplate);
		m_backend.setup();
	}
	ArrayImplT()
	: ArrayImpl()
	, m_backend()
	{}
	inline ArrayImplT
	( AllocatorOnStorage & allocator, treedb_const_array_impl_ptr_t ptemplate)
	: ArrayImpl()
	, m_backend()
	{
		setup(allocator, ptemplate);
	}
	const DataType * entries(HTREEDB_ARRAY ha) const
	{ return (DataType *)Entries(ha); }
	DataType * entries(HTREEDB_ARRAY ha)
	{ return (DataType *)Entries(ha); }

	Backend m_backend;
};

template < typename Backend_, typename Impl_ = ArrayImplT< Backend_ > >
class ArrayWithImplT;

template < typename Impl_ >
class ArrayT : public Array
{
public:
	typedef Impl_ Impl;
	typedef ArrayT< Impl > ArrayType;
	typedef typename Impl::Backend Backend;
	typedef typename Backend::DataType DataType;
	typedef typename Backend::InputType InputType;
	typedef PackedPayload< Impl > HandleType;

	inline HandleType handle() const
		{ return HandleType(m_inst.m_handle); }
	inline HandleType & handle()
		{ return * (HandleType *)(& m_inst.m_handle); }

	ArrayT() {}
	void setup(Impl_ & timpl, HandleType payload)
	{ treedb_array_t_Setup(& m_inst, & timpl.m_inst, payload.m_value); }
	ArrayT(Impl & timpl, HandleType payload)
	{ setup(timpl, payload); }
	ArrayT(ArrayWithImplT< Backend, Impl > & o, HandleType payload);

	Impl & impl() { return (Impl &)Array::impl(); }
	bool push_front(const InputType & d)
	{ return Array::PushFront(Impl::Backend::pack_input(d)); }
	bool push_back(const InputType & d)
	{ return Array::PushBack(Impl::Backend::pack_input(d)); }
	bool pop_front()
	{ return Array::PopFront(); }
	bool pop_back()
	{ return Array::PopBack(); }

	const DataType * entries() const
	{ return (DataType *)const_cast< ArrayT<Impl_> * >(this)->Entries(); }
	DataType * entries()
	{ return (DataType *)Entries(); }
	const DataType & operator[](uintptr_t x) const
	{ return entries()[x]; }
	DataType & operator[](uintptr_t x)
	{ return entries()[x]; }

	uintptr_t count() const
	{ return const_cast< ArrayT< Impl > * >(this)->Count(); }

	bool insert(const InputType & d, uintptr_t where)
	{ return Array::Insert(Impl::Backend::pack_input(d), where); }
	bool remove(uintptr_t where)
	{ return Array::Remove(where); }

	DataType * begin() { return entries(); }
	DataType * end() { return begin() + count(); }
	const DataType * begin() const { return entries(); }
	const DataType * end() const { return begin() + count(); }

	bool clone(ArrayT< Impl_ > & from)
	{ return Array::Clone(from.m_inst.m_handle); }
};

template < typename Backend_, typename Impl_ /* = ArrayImplT< Backend_ > */>
class ArrayWithImplT : public ArrayT< Impl_ >
{
public:
	typedef Backend_ Backend;
	typedef Impl_ Impl;
	typedef ArrayT< Impl_ > ArrayType;
	ArrayWithImplT()
	: ArrayType()
	, m_impl()
	{}
	inline Backend & backend() { return *(Backend *)m_impl->m_pBackend; }
	void setup
	( treedb_const_array_ptr_t ptemplate
	, AllocatorOnStorage & allocator
	, PackedPayload< Impl > payload
	) {
		m_impl.setup(ptemplate->m_pImpl, allocator);
		ArrayType::setup(m_impl, payload);
	}
	ArrayWithImplT(AllocatorOnStorage & allocator
		, treedb_const_array_ptr_t ptemplate, PackedPayload< Impl > payload)
	: ArrayType()
	, m_impl(allocator, ptemplate->m_pImpl)
	{ ArrayType::setup(m_impl, payload); }
	Impl m_impl;
};
/** Be sure the implementation stays around when using this.
*/
template < typename Impl >
ArrayT< Impl >::ArrayT
( ArrayWithImplT< ArrayT< Impl >::Backend, Impl > & o
, PackedPayload< Impl > payload
) {
	this->setup(o.impl(), payload);
}

template
< typename Runtime
, typename BackendTypes
, uint16_t size
, uint16_t align
>
class ArrayBackendT : public IArrayBackendT< Runtime >
{
public:
	typedef IArrayBackendT< Runtime > Base;
	typedef typename BackendTypes::DataType DataType;
	typedef typename BackendTypes::InputType InputType;

	void setup()
	{
		Base::m_pDescription = & m_description;
		m_description.m_entry.m_uSize = size;
		m_description.m_entry.m_uAlign = align;
	}

	ArrayBackendT() {} // Do not initialise m_description!
	virtual BOOL STDCALL FreeEntries
	( Runtime * runtime
	, HTREEDB_ARRAY ha
	, uintptr_t where
	, uintptr_t count
	) {
		(void)runtime;
		(void)ha;
		(void)where;
		(void)count;
		return false; /* No pointer data to free. */
	}
	TreedbArrayDescription m_description;
};

} // namespace meta
} // namespace treedb
} // namespace v3c

#endif /* #ifndef v3c_meta_treedb_array_hpp_included */
