/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stddef.h>
#include <v3c/bits.h>
#define COBJMACROS
#include "meta-treedb.h"
#include "meta-avl.h"

#define TREEDB_NS(x) treedb_##x
#define treedb_treedb_t TreedbHeader
#define treedb_treedb_ptr_t TreedbHeaderPtr
#define treedb_runtime_t treedb_storage_on_file_t
#define treedb_runtime_ptr_t treedb_storage_on_file_ptr_t

#define TREEDB_FUNC0(x) treedb_storage_##x(runtime)
#define treedb_storage_heap_close(x) treedb_storage_t_Close(x)

#define TREEDB_NS_prefix static
#define TREEDB_STORAGE_base runtime->m_pBase
#define TREEDB_STORAGE_size runtime->m_uSize
#define TREEDB_STORAGE_set_size(sz) runtime->m_uSize = (sz)
#define TREEDB_STORAGE_max_size \
((runtime->m_pTreedb->m_uBytesPerOffset < 8) ? (((uint64_t)1UL << (runtime->m_pTreedb->m_uBytesPerOffset * 8))-1UL) : (uint64_t)-1UL)
#define TREEDB_STORAGE_header runtime->m_pTreedb
#undef treedb_storage_t_Start
#define treedb_storage_t_Start(This) (This)->m_pBase
#define TREEDB_OMIT_FILE_STORAGE_header_check
#define TREEDB_OMIT_FILE_STORAGE_heap_open
#define TREEDB_OMIT_FILE_STORAGE_heap_create
#define TREEDB_OMIT_post_free_impl
#define TREEDB_HAS_RESIZABLE_HEAP

/* Route the calls through the vtable. */
#define TREEDB_heap_read_lock() treedb_storage_t_ReadLock(runtime)
#define TREEDB_heap_write_lock() treedb_storage_t_WriteLock(runtime)
#define TREEDB_heap_unlock() treedb_storage_t_Unlock(runtime)
#define TREEDB_heap_close() treedb_storage_t_Close(runtime)
#define TREEDB_heap_init(name, open_flags, create_permissions, access_flags \
	, offset, page_sz) \
	treedb_storage_on_file_init(runtime, name, open_flags \
		, create_permissions, access_flags, offset, page_sz)

/* Rename for clarity */
#define treedb_heap_init treedb_storage_on_file_init

#include <treedb/file-storage-impl.h>

static TreedbHeapStatus STDCALL treedb_ReadLock
( treedb_storage_on_file_ptr_t runtime
) {
	return TREEDB_NS(heap_read_lock)(runtime);
}
static TreedbHeapStatus STDCALL treedb_WriteLock
( treedb_storage_on_file_ptr_t runtime
) {
	return TREEDB_NS(heap_write_lock)(runtime);
}
static TreedbHeapStatus STDCALL treedb_Unlock
( treedb_storage_on_file_ptr_t runtime
) {
	return TREEDB_NS(heap_unlock)(runtime);
}
static StorageLockType STDCALL treedb_LockType
( treedb_storage_on_file_ptr_t runtime
) {
	return runtime->m_active_writer_thread
		? eStorageWriteLocked
		: TREEDB_STORAGE_base
			? eStorageReadLocked
			: eStorageUnlocked;
}
static BOOL STDCALL treedb_IsOpen
( treedb_storage_on_file_ptr_t runtime
) {
	return runtime->m_fls.m_fd != -1;
}
static BOOL STDCALL treedb_CanWrite
( treedb_storage_on_file_ptr_t runtime
) {
	return (runtime->m_open_flags & eHeapOpenWrite) &&
		(runtime->m_map_protect & eHeapAccessWrite);
}
static BOOL STDCALL treedb_CanExec
( treedb_storage_on_file_ptr_t runtime
) {
	return (runtime->m_map_protect & eHeapAccessExec);
}
static TreedbHeapStatus STDCALL treedb_Close
( treedb_storage_on_file_ptr_t runtime
) {
	return TREEDB_NS(heap_close)(runtime);
}
static PVOID STDCALL treedb_MoreCore
( treedb_storage_on_file_ptr_t runtime
, size_t cb
) {
	return TREEDB_NS(impl_page_alloc)(runtime, cb);
}
static TreedbFreeResult STDCALL treedb_FreeCore
( treedb_storage_on_file_ptr_t runtime
, size_t cb
) {
	return TREEDB_NS(impl_page_free)(runtime, cb);
}

static BOOL STDCALL treedb_Resizable
( treedb_storage_on_file_ptr_t This
) {
	(void)This;
	return 1;
}
/* Set the freed bytes to zero? */
static void STDCALL treedb_PostFree
( treedb_storage_on_file_ptr_t This
, PVOID pv
, size_t cb
) {
	(void)This;
	(void)pv;
	(void)cb;
}
static TreedbHeapStatus treedb_heap_header_check
( treedb_storage_on_file_ptr_t runtime
) {
	TreedbHeapStatus ret = eTreedbHeapSuccess;
	do {
		size_t top = lseek(runtime->m_fls.m_fd, 0, SEEK_END);
		lseek(runtime->m_fls.m_fd, 0, SEEK_SET);

		if(top < sizeof(TreedbHeader)) {
			ret = eTreedbHeapBadTop;
			break;
		}/*
		if(uuid_compare(runtime->m_pTreedb->m_uAlgorithm
				, UUID_treedb_avl_algorithm)) {
			ret = eTreedbHeapOpenUnknownAllocatorAlgorithm;
			break;
		}*/
		if(strcmp(runtime->m_pTreedb->m_szMagic, "3db")) {
			ret = eTreedbHeapOpenBadMagic;
			break;
		}
		switch(runtime->m_pTreedb->m_uBytesPerOffset) {
		case 2: case 4: case 8:
			break;
		default:
			ret = eTreedbHeapOpenUnhandledOffsetSize;
			break;
		}
		if(ret == eTreedbHeapOpenUnhandledOffsetSize)
			break;
		/* Make sure we can at least access the stored top of storage. */
		/* The "6" comes from two trees(3 offs each). */
		if((V3C_ROUND_UP_T2(unsigned int, sizeof(TreedbHeader)
			, runtime->m_pTreedb->m_uBytesPerOffset)
				+ (runtime->m_pTreedb->m_uBytesPerOffset) * 6)
				> top) {
			ret = eTreedbHeapBadTop;
			break;
		}
		if(top % TREEDB_PAGE_SIZE) {
			ret = eTreedbHeapBadTop;
			break;
		}
		/* We use this to try to get a matching allocator, so we don't check
		here. */
#if 0
		if(runtime->m_pTreedb->m_uBytesPerOffset
				!= sizeof(TREEDB_NS(count_t))) {
			ret = eTreedbHeapOpenUnhandledOffsetSize;
			break;
		}
#endif
		if(runtime->m_pTreedb->m_uEndianIndex
				!= v3c_native_endian_index()) {
			ret = eTreedbHeapOpenUnhandledEndianIndex;
			break;
		}
		/* We use this to try to get a matching allocator, so we don't check
		here. */
#if 0
		if(TREEDB_STORAGE_HEADER.m_uFeatures != 3) {
			ret = eTreedbHeapOpenUnhandledFeatures; // We need count and depth.
			break;
		}
#endif
	} while(0);
	return ret;
}

TreedbHeapStatus STDCALL treedb_Create
( treedb_storage_on_file_ptr_t This
, treedb_allocator_ptr_t pAllocator
, const uuid_t uSchema
, const char * name
, TreedbHeapOpenFlags open_flags
, int create_permissions
, int access_flags
, size_t offset
, size_t size
, size_t uPayloadSize
) {
	TreedbHeapStatus ret = treedb_storage_on_file_init(This, name, open_flags
		, create_permissions, access_flags, offset, size);
	if(ret != eTreedbHeapSuccess)
		return ret;
	if(!treedb_allocator_t_Init(pAllocator, uPayloadSize, size
			, & This->m_pTreedb->m_ulPayloadOffset)) {
		treedb_Unlock(This);
		treedb_Close(This);
		return eTreedbHeapPayloadSizeTooLarge;
	}
	strcpy(This->m_pTreedb->m_szMagic, "3db");
	uuid_copy(This->m_pTreedb->m_uAllocatorAlgorithm
		, treedb_allocator_t_Guid(pAllocator));
	uuid_copy(This->m_pTreedb->m_uSchema, uSchema);
	This->m_pTreedb->m_uFeatures
		= pAllocator->m_pImpl->m_pDescription->m_flags / 4;
	This->m_pTreedb->m_uEndianIndex = v3c_native_endian_index();
	This->m_pTreedb->m_uBytesPerOffset
		= pAllocator->m_pImpl->m_pDescription->m_pointer.m_uSize;
	return eTreedbHeapSuccess;
}
TREEDB_NS_prefix TreedbHeapStatus STDCALL treedb_OpenUnknown
( treedb_storage_on_file_ptr_t This
, const char * name
, TreedbHeapOpenFlags open_flags
, int access_flags
, size_t offset
) {
	TreedbHeapStatus ret = treedb_heap_just_open
		(This, name, open_flags, access_flags, offset);
	if(ret != eTreedbHeapSuccess)
		return ret;
	ret = treedb_heap_header_check(This);
	if(ret != eTreedbHeapSuccess) {
		treedb_Unlock(This);
		treedb_Close(This);
		return ret;
	}
	This->m_uEndianIndex
		= This->m_pTreedb->m_uEndianIndex;
	This->m_uEndianSwapIndex
		= v3c_native_endian_index() ^ This->m_uEndianIndex;
	return ret;
}
TREEDB_NS_prefix TreedbHeapStatus STDCALL treedb_Open
( treedb_storage_on_file_ptr_t This
, const uuid_t uSchema
, const char * name
, TreedbHeapOpenFlags open_flags
, int access_flags
, size_t offset
) {
	TreedbHeapStatus res = treedb_OpenUnknown(This, name, open_flags
		, access_flags, offset);
	if(res != eTreedbHeapSuccess)
		return res;
	if(uuid_compare(This->m_pTreedb->m_uSchema, uSchema)) {
		treedb_Unlock(This);
		treedb_Close(This);
		return eTreedbHeapOpenUnknownSchema;
	}/*
	if(uuid_compare
			(This->m_pTreedb->m_uAllocatorAlgorithm, uAllocatorAlgorithm)) {
		treedb_Unlock(This);
		treedb_Close(This);
		return eTreedbHeapOpenUnknownAllocatorAlgorithm;
	}*/
	return res;
}
TREEDB_NS_prefix uint8_t STDCALL treedb_EndianIndex
( treedb_storage_on_file_ptr_t This
) {
	return This->m_uEndianIndex;
}
TREEDB_NS_prefix uint8_t STDCALL treedb_EndianSwapIndex
( treedb_storage_on_file_ptr_t This
) {
	return This->m_uEndianSwapIndex;
}
#if 0
#define TREEDB_NS(x) treedb_##x
static void TREEDB_NS(test)(treedb_storage_on_file_vtbl_t * v)
{
	v->ReadLock = TREEDB_NS(ReadLock);
	v->WriteLock = TREEDB_NS(WriteLock);
	v->Unlock = TREEDB_NS(Unlock);
	v->LockType = TREEDB_NS(LockType);
	v->IsOpen = TREEDB_NS(IsOpen);
	v->CanWrite = TREEDB_NS(CanWrite);
	v->CanExec = TREEDB_NS(CanExec);
	v->Close = TREEDB_NS(Close);
	v->Open = TREEDB_NS(Open);
	v->Create = TREEDB_NS(Create);
	v->MoreCore = TREEDB_NS(MoreCore);
	v->FreeCore = TREEDB_NS(FreeCore);
	v->Resizable = TREEDB_NS(Resizable);
	v->PostFree = TREEDB_NS(PostFree);
	v->EndianIndex = TREEDB_NS(EndianIndex);
}
#endif
treedb_storage_on_file_vtbl_t treedb_storage_on_file_vtbl = {
	treedb_storage_vtbl_t_entries(treedb_)
};
