/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#include "meta-treedb-handle-array.h"
#include <memory.h>

void treedb_handle_array_t_Setup
( treedb_handle_array_ptr_t p
, treedb_handle_array_impl_ptr_t pImpl
, uintptr_t uPayload
) {
	p->m_pImpl = pImpl;
	p->m_handle = uPayload;
}
void treedb_handle_array_impl_t_Setup
( treedb_handle_array_impl_ptr_t pm
, treedb_allocator_on_storage_ptr_t pa
, treedb_handle_array_backend_ptr_t pb
, TreedbHandleArrayDescription * pd
, treedb_const_handle_array_impl_ptr_t ptemplate
) {
	pm->m_pVtbl = ptemplate->m_pVtbl;
	pm->m_pAllocator = (treedb_allocator_ptr_t)pa;
	pm->m_pBackend = pb;
	*pd = *(ptemplate->m_pBackend->m_pDescription);
	pb->m_pDescription = pd;

	pd->m_max_align = v3c_max(pd->m_entry.m_uAlign
		, pa->m_pImpl->m_pDescription->m_pointer.m_uAlign);
	pd->m_entry_offset = V3C_ROUND_UP_T(uint16_t, pd->m_root.m_uSize
		, pd->m_max_align);
}
void treedb_handle_array_backend_t_Setup
( treedb_handle_array_backend_ptr_t pb
, PVOID pVtbl
, TreedbHandleArrayDescriptionConstPtr pDescription
) {
	*((treedb_handle_array_backend_vtbl_ptr_t *)(& pb->m_pVtbl))
		= (treedb_handle_array_backend_vtbl_ptr_t)pVtbl;
	pb->m_pDescription = pDescription;
}

#include <stddef.h> /* for offsetof() */

#define ARRAY_TMPL_runtime_ptr_t treedb_handle_array_runtime_ptr_t
#define ARRAY_TMPL_prefix static inline
#define ARRAY_TMPL_NS(x) V3C_INVOKE2(V3C_JOIN2,ARRAY_TMPL_NS_,x)
#define ARRAY_TMPL_entry_size \
	runtime->m_pImpl->m_pBackend->m_pDescription->m_entry.m_uSize

#define ARRAY_TMPL_data_t char
#define ARRAY_TMPL_input_t PCVOID

#define ARRAY_TMPL_alloc(count) impl_alloc(runtime, count)

#ifdef DOXYGEN_processing
{
#endif

#include "meta-treedb-handle-array-defines.h"

#ifdef DOXYGEN_processing
}
#endif

static PVOID impl_alloc
( treedb_handle_array_runtime_ptr_t runtime
, uintptr_t count
) {
	uintptr_t size
		= runtime->m_pImpl->m_pBackend->m_pDescription->m_entry_offset
			+ count * ARRAY_TMPL_entry_size;

	META_TREEDB_printf("treedb_handle_array_impl_alloc : allocating %zu bytes."
		"\n" , size);
	PVOID p = treedb_allocator_t_PointerAlignedAlloc
		(runtime->m_pImpl->m_pAllocator, size
			, runtime->m_pImpl->m_pBackend->m_pDescription->m_max_align
			, runtime->m_pImpl->m_pBackend->m_pDescription->m_entry_offset);
	META_TREEDB_printf("treedb_handle_array_impl_alloc : allocated %zu bytes at"
		" %p[%zu].\n"
		, treedb_allocator_t_GetSizePtr(runtime->m_pImpl->m_pAllocator, p)
		, p
		, treedb_handle_array_t_PointerToOffset(runtime, p)
	);
	return p;
}

#define ARRAY_TMPL_IS_POINTER 0

#ifdef __cplusplus
extern "C" {
#endif

/******************************************************************************/
/* 16                                                                        */
/******************************************************************************/
#define ARRAY_TMPL_count_t uint16_t
#define ARRAY_TMPL_data_t char
#define ARRAY_TMPL_data_ptr_t PSTR
#define ARRAY_TMPL_input_t PCVOID

#define ARRAY_TMPL_NS_ treedb_handle_array_o16_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-treedb-handle-array-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef ARRAY_TMPL_NS_

#include <treedb/handle-array-template-cleardefs.h>

/******************************************************************************/
/* 32                                                                        */
/******************************************************************************/
/* 16->32 */
#undef ARRAY_TMPL_count_t

#define ARRAY_TMPL_count_t uint32_t

#define ARRAY_TMPL_NS_ treedb_handle_array_o32_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-treedb-handle-array-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef ARRAY_TMPL_NS_

#include <treedb/handle-array-template-cleardefs.h>

#if HAVE_64_BITS
/******************************************************************************/
/* 64                                                                        */
/******************************************************************************/
/* 32->64 */
#undef ARRAY_TMPL_count_t

#define ARRAY_TMPL_count_t uint64_t

#define ARRAY_TMPL_NS_ treedb_handle_array_o64_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-treedb-handle-array-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef ARRAY_TMPL_NS_

#include <treedb/handle-array-template-cleardefs.h>

#endif /* #if HAVE_64_BITS */

/******************************************************************************/
/* pointer                                                                    */
/******************************************************************************/
#undef ARRAY_TMPL_IS_POINTER
#define ARRAY_TMPL_IS_POINTER eTreedbContainerIsPointer

#define ARRAY_TMPL_count_t uintptr_t
#define ARRAY_TMPL_data_t char
#define ARRAY_TMPL_data_ptr_t PSTR
#define ARRAY_TMPL_packed_data_ptr_t PSTR

#define ARRAY_TMPL_packed_ptr_t ARRAY_TMPL_ptr_t
#define ARRAY_TMPL_packed2_ptr_t ARRAY_TMPL_ptr_ptr_t

#define ARRAY_TMPL_omit_unpack
#define ARRAY_TMPL_omit_pack
#define ARRAY_TMPL_omit_unpack2

#define ARRAY_TMPL_unpack(x) ((ARRAY_TMPL_ptr_t)x)
#define ARRAY_TMPL_pack(x) (x)
#define ARRAY_TMPL_unpack2(x) (*(ARRAY_TMPL_ptr_t *)(x))

#define ARRAY_TMPL_pack_pointer(x) (x)
#define ARRAY_TMPL_unpack_pointer(x) (x)

#define ARRAY_TMPL_NS_ treedb_handle_array_p_

#ifdef DOXYGEN_processing
{
#endif

#include "meta-treedb-handle-array-impl.h"

#ifdef DOXYGEN_processing
}
#endif

#undef ARRAY_TMPL_NS_

#include <treedb/handle-array-template-cleardefs.h>

/******************************************************************************/

extern treedb_const_handle_array_ptr_t
treedb_get_handle_array_template_for
( uint16_t uBytes /* 2, 4, 8 */
, uint16_t flags /* flags & eTreedbContainerIsPointer means ignore uBytes */
) {
	if(flags & eTreedbContainerIsPointer) {
		return & treedb_handle_array_p_instance;
	}
	switch(uBytes) {
	case 2:
		return & treedb_handle_array_o16_instance;
	case 4:
		return & treedb_handle_array_o32_instance;
#if HAVE_64_BITS
	case 8:
		return & treedb_handle_array_o64_instance;
#endif
	default:
		break;
	}
	return 0;
}

extern treedb_const_handle_array_ptr_t
treedb_get_handle_array_template_using_allocator
( treedb_allocator_ptr_t pallocator
) {
	if(!pallocator)
		return 0;
	int flags = 0;
	if(pallocator->m_pImpl->m_pDescription->m_flags & eTreedbContainerIsPointer)
		flags |= eTreedbContainerIsPointer;
	treedb_const_handle_array_ptr_t pt
		= treedb_get_handle_array_template_for
			( pallocator->m_pImpl->m_pDescription->m_pointer.m_uSize
			, flags);
	return pt;
}

#ifdef __cplusplus
} /* extern "C" { */
#endif
