/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef v3c_meta_treedb_handle_array_hpp_included
  #define v3c_meta_treedb_handle_array_hpp_included

  #ifndef v3c_meta_treedb_handle_array_h_included
	#define COBJMACROS
	#include "meta-treedb-handle-array.h"
  #endif

  #ifndef v3c_meta_treedb_hpp_included
	#define COBJMACROS
	#include "meta-treedb.hpp"
  #endif

  #include <v3c/utils.h>

namespace v3c {
namespace treedb {
namespace meta {

class HandleArrayRuntime;
/*! treedb_array_impl_t vtable-only overlay
*/
class IHandleArrayImpl_VtableOverlay
{
public:
#include "meta-treedb-cxx-push-c.h"
	treedb_handle_array_impl_t_methods(,,HandleArrayRuntime);
#include "meta-treedb-cxx-pop-c.h"
protected:
	//! This can't be virtual. If it were it would mess up derived interfaces.
	~IHandleArrayImpl_VtableOverlay() {}
};

/*! treedb_array_impl_t overlay
*/
template< typename Interface = IHandleArrayImpl_VtableOverlay >
class ITreedbHandleArrayImplT : public Interface
{
public:
	inline Allocator & allocator()
	{ return *(Allocator *)(m_pAllocator); }
	inline const TreedbHandleArrayDescription & description() const
	{ return * m_pBackend->m_pDescription; }

	treedb_handle_array_impl_t_members(treedb_handle_array_impl_vtbl_ptr_t);
};

/*! treedb_handle_array_backend_t overlay
*/
template< typename HandleArrayRuntimeType = HandleArrayRuntime >
class IHandleArrayBackendT
{
public:
	treedb_handle_array_backend_t_methods(,,,HandleArrayRuntimeType);
	TreedbHandleArrayDescriptionConstPtr m_pDescription;
protected:
	//! OK this can be virtual.
	~IHandleArrayBackendT() {}
};

////////////////////////////////////////////////////////////////////////////////
// Wrappers
////////////////////////////////////////////////////////////////////////////////
#define DGTHIS_MEM(x,m) x.m
#define DGTHIS0(x) treedb_handle_array_t_to_runtime(& (x))
#define DGTHIS(x) treedb_handle_array_t_to_runtime(& (x)),
#define V3C_DELEGATE(Inst, x) Inst.m_pImpl->m_pVtbl->x

class HandleArrayImpl;

/*! treedb_array_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class HandleArray
{
public:
	/* Convenience accessors. */
	inline const treedb_handle_array_impl_t & impl() const
	{ return * m_inst.m_pImpl; }
	inline const treedb_handle_array_backend_t & backend() const
	{ return * m_inst.m_pImpl->m_pBackend; }
	inline const TreedbHandleArrayDescription & description() const
	{ return * backend().m_pDescription; }
	inline Allocator & allocator()
	{ return *(Allocator *)(impl().m_pAllocator); }

	inline HHTREEDB_ARRAY raw_handle() const { return m_inst.m_handle; }
	inline HHTREEDB_ARRAY & raw_handle() { return m_inst.m_handle; }

	TREEDB_IHandleArray_DELEGATE_TO(m_inst);

	inline BOOL STDCALL Alloc(uintptr_t reserve, uintptr_t increment)
	{
		return treedb_handle_array_impl_t_Alloc(
			treedb_handle_array_t_to_runtime(& m_inst), m_inst.m_handle, reserve
			, increment);
	}
	inline void STDCALL Free()
	{
		treedb_handle_array_impl_t_Free(
			treedb_handle_array_t_to_runtime(& m_inst), m_inst.m_handle);
	}
	treedb_handle_array_t m_inst;
};

#undef V3C_DELEGATE
#undef DGTHIS
#undef DGTHIS0

/*! treedb_handle_array_runtime_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class HandleArrayRuntime
{
public:
	inline HandleArrayRuntime() { m_inst.m_pImpl = 0; }
	inline HandleArrayRuntime(HandleArrayImpl * p)
		{ m_inst.m_pImpl = (treedb_handle_array_impl_ptr_t)p; }
	/* Convenience accessors. */
	inline const TreedbHandleArrayDescription & description() const
	{ return * m_inst.m_pImpl->m_pBackend->m_pDescription; }
	inline Allocator & allocator()
	{ return *(Allocator *)(m_inst.m_pImpl->m_pAllocator); }

	treedb_handle_array_runtime_t m_inst;
};

#define DGTHIS0(x) (&x)
#define DGTHIS(x) (&x),
#define V3C_DELEGATE(Inst, x) Inst.m_pVtbl->x

/*! treedb_array_impl_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class HandleArrayImpl
{
public:
	inline HandleArrayImpl() : m_inst() {}
	inline void setup
	( AllocatorOnStorage & allocator
	, treedb_handle_array_backend_ptr_t pBackend
	, TreedbHandleArrayDescription * pd
	, treedb_const_handle_array_impl_ptr_t ptemplate
	) {
		treedb_handle_array_impl_t_Setup(& m_inst
			, & allocator.c_type(), pBackend, pd, ptemplate);
	}
	inline HandleArrayImpl
	( AllocatorOnStorage & allocator
	, treedb_handle_array_backend_ptr_t pBackend
	, TreedbHandleArrayDescription * pd
	, treedb_const_handle_array_impl_ptr_t ptemplate
	) {
		setup(allocator, pBackend, pd, ptemplate);
	}
	/* Convenience accessors. */
	inline const TreedbHandleArrayDescription & description() const
	{ return * m_inst.m_pBackend->m_pDescription; }
	inline Allocator & allocator()
	{ return *(Allocator *)(m_inst.m_pAllocator); }

	TREEDB_IHandleArrayImpl_DELEGATE_TO(m_inst);
	treedb_handle_array_impl_t m_inst;
};

class HandleArrayBackend
{
public:
	TREEDB_IHandleArrayBackend_DELEGATE_TO(m_inst);
	treedb_handle_array_backend_t m_inst;
};

#undef V3C_DELEGATE
#undef DGTHIS0
#undef DGTHIS
#undef DGTHIS_MEM

////////////////////////////////////////////////////////////////////////////////
// Templates
////////////////////////////////////////////////////////////////////////////////
template< typename Impl >
class HandleArrayRuntimeT
: public HandleArrayRuntime
{
public:
	inline HandleArrayRuntimeT() : HandleArrayRuntime() {}
	inline HandleArrayRuntimeT(Impl * p) : HandleArrayRuntime((Impl *)p) {}
	inline Impl & impl() { return *(Impl *)m_inst.m_pImpl; }
};

template< typename Backend_ >
class HandleArrayImplT
: public HandleArrayImpl
{
public:
	typedef Backend_ Backend;
	typedef typename Backend::DataType DataType;
	inline void setup
	( AllocatorOnStorage & allocator
	, treedb_const_handle_array_impl_ptr_t ptemplate
	) {
		HandleArrayImpl::setup(allocator
			, (treedb_handle_array_backend_ptr_t)(& m_backend)
			, & m_backend.m_description, ptemplate);
		m_backend.setup();
	}
	HandleArrayImplT()
	: HandleArrayImpl()
	, m_backend()
	{}
	inline HandleArrayImplT
	( AllocatorOnStorage & allocator
	, treedb_const_handle_array_impl_ptr_t ptemplate)
	: HandleArrayImpl()
	, m_backend()
	{
		setup(allocator, ptemplate);
	}
	const DataType * entries(HHTREEDB_ARRAY hha) const
	{ return (DataType *)Entries(hha); }
	DataType * entries(HHTREEDB_ARRAY hha)
	{ return (DataType *)Entries(hha); }

	Backend m_backend;
};

template < typename Backend_, typename Impl_ = HandleArrayImplT< Backend_ > >
class HandleArrayWithImplT;

template < typename Impl_ >
class HandleArrayT : public HandleArray
{
public:
	typedef Impl_ Impl;
	typedef HandleArrayT< Impl > HandleArrayType;
	typedef typename Impl::Backend Backend;
	typedef typename Backend::DataType DataType;
	typedef typename Backend::InputType InputType;
	typedef PackedPayload< Impl > HandleType;

	inline HandleType handle() const
		{ return HandleType(m_inst.m_handle); }
	inline HandleType & handle()
		{ return * (HandleType *)(& m_inst.m_handle); }

	HandleArrayT() {}
	void setup(Impl_ & timpl, HandleType payload)
	{ treedb_handle_array_t_Setup(& m_inst, & timpl.m_inst, payload.m_value); }
	HandleArrayT(Impl & timpl, HandleType payload)
	{ setup(timpl, payload); }
	HandleArrayT(HandleArrayWithImplT< Backend, Impl > & o, HandleType payload);

	Impl & impl() { return (Impl &)HandleArray::impl(); }
	bool push_front(const InputType & d)
	{ return HandleArray::PushFront(Impl::Backend::pack_input(d)); }
	bool push_back(const InputType & d)
	{ return HandleArray::PushBack(Impl::Backend::pack_input(d)); }
	bool pop_front()
	{ return HandleArray::PopFront(); }
	bool pop_back()
	{ return HandleArray::PopBack(); }

	const DataType * entries() const
	{ return (DataType *)const_cast< HandleArrayT<Impl_> * >(this)->Entries(); }
	DataType * entries()
	{ return (DataType *)Entries(); }
	const DataType & operator[](uintptr_t x) const
	{ return entries()[x]; }
	DataType & operator[](uintptr_t x)
	{ return entries()[x]; }

	uintptr_t count() const
	{ return const_cast< HandleArrayT<Impl_> * >(this)->Count(); }

	bool insert(const InputType & d, uintptr_t where)
	{ return HandleArray::Insert(Impl::Backend::pack_input(d), where); }
	bool remove(uintptr_t where)
	{ return HandleArray::Remove(where); }

	DataType * begin() { return entries(); }
	DataType * end() { return begin() + count(); }
	const DataType * begin() const { return entries(); }
	const DataType * end() const { return begin() + count(); }

	bool clone(HandleArrayT< Impl_ > & from)
	{ return HandleArray::Clone(from.m_inst.m_handle); }
};

template < typename Backend_, typename Impl_ /* = HandleArrayImplT< Backend_ >*/ >
class HandleArrayWithImplT : public HandleArrayT< Impl_ >
{
public:
	typedef Backend_ Backend;
	typedef Impl_ Impl;
	typedef HandleArrayT< Impl_ > HandleArrayType;
	HandleArrayWithImplT()
	: HandleArrayType()
	, m_impl()
	{}
	inline Backend & backend() { return *(Backend *)m_impl->m_pBackend; }
	void setup
	( treedb_const_handle_array_ptr_t ptemplate
	, AllocatorOnStorage & allocator
	, PackedPayload< Impl > payload
	) {
		m_impl.setup(ptemplate->m_pImpl, allocator);
		HandleArrayType::setup(m_impl, payload);
	}
	HandleArrayWithImplT(AllocatorOnStorage & allocator
		, treedb_const_handle_array_ptr_t ptemplate
			, PackedPayload< Impl > payload)
	: HandleArrayType()
	, m_impl(allocator, ptemplate->m_pImpl)
	{ HandleArrayType::setup(m_impl, payload); }
	Impl m_impl;
};

template < typename Impl >
HandleArrayT< Impl >::HandleArrayT
( HandleArrayWithImplT< HandleArrayT< Impl >::Backend, Impl > & o
, PackedPayload< Impl > payload
) {
	this->setup(o.impl(), payload);
}


template
< typename Runtime
, typename BackendTypes
, uint16_t size
, uint16_t align
>
class HandleArrayBackendT : public IHandleArrayBackendT< Runtime >
{
public:
	typedef IHandleArrayBackendT< Runtime > Base;
	typedef typename BackendTypes::DataType DataType;
	typedef typename BackendTypes::InputType InputType;

	void setup()
	{
		Base::m_pDescription = & m_description;
		m_description.m_entry.m_uSize = size;
		m_description.m_entry.m_uAlign = align;
	}

	HandleArrayBackendT() {} // Do not initialise m_description!
	virtual BOOL STDCALL FreeEntries
	( Runtime * runtime
	, HHTREEDB_ARRAY hha
	, uintptr_t where
	, uintptr_t count
	) {
		(void)runtime;
		(void)hha;
		(void)where;
		(void)count;
		return false; /* No pointer data to free. */
	}
	TreedbHandleArrayDescription m_description;
};

} // namespace meta
} // namespace treedb
} // namespace v3c

#endif /* #ifndef v3c_meta_treedb_handle_array_hpp_included */
