/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef v3c_meta_treedb_list_hpp_included
  #define v3c_meta_treedb_list_hpp_included

  #ifndef v3c_meta_treedb_list_h_included
	#define COBJMACROS
	#include "meta-treedb-list.h"
  #endif

  #include <v3c/utils.h>

namespace v3c {
namespace treedb {
namespace meta {

class ListRuntime;
/*! treedb_list_impl_t vtable-only overlay
*/
class IListImpl_VtableOverlay
{
public:
#include "meta-treedb-cxx-push-c.h"
	treedb_list_impl_t_methods(,,ListRuntime);
#include "meta-treedb-cxx-pop-c.h"
protected:
	//! This can't be virtual. If it were it would mess up derived interfaces.
	~IListImpl_VtableOverlay() {}
};

/*! treedb_list_impl_t overlay
*/
template< typename Interface = IListImpl_VtableOverlay >
class ITreedbListImplT : public Interface
{
public:
	inline Allocator & allocator()
	{ return *(Allocator *)(m_pAllocator); }
	inline const TreedbItemDescription & description() const
	{ return * m_pDescription; }
	inline char * node_to_data(PVOID pn)
	{ return (char *)pn + description().m_aligned_node_size; }
	inline char * data_to_node(PVOID pv)
	{ return (char *)pv - description().m_aligned_node_size; }

	inline const char * node_to_data(PCVOID pn)
	{ return (const char *)pn + description().m_aligned_node_size; }
	inline const char * data_to_node(PCVOID pv)
	{ return (const char *)pv - description().m_aligned_node_size; }

	treedb_list_impl_t_members;
};

/*! treedb_list_backend_t overlay

Because treedb_list_backend_t doesn't have data members (besides the vtable)
the step of defining a VtableOverlay is skipped here.
*/
template< typename ListImplType = ITreedbListImplT<> >
class IListBackendT
{
public:
#define treedb_list_ptr_t ListHandleType *
	treedb_list_backend_t_methods(,,,ListImplType);
#undef treedb_list_ptr_t
protected:
	//! OK this can be virtual.
	~IListBackendT() {}
};

////////////////////////////////////////////////////////////////////////////////
// Wrappers
////////////////////////////////////////////////////////////////////////////////
#define DGTHIS_MEM(x,m) x.m
#define DGTHIS0(x) treedb_list_t_to_runtime(& (x))
#define DGTHIS(x) treedb_list_t_to_runtime(& (x)),
#define V3C_DELEGATE(Inst, x) Inst.m_pImpl->m_pVtbl->x

class ListImpl;

/*! treedb_list_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class List
{
public:
	/* Convenience accessors. */
	inline const treedb_list_impl_t & impl() const
	{ return * m_inst.m_pImpl; }
	inline const TreedbItemDescription & description() const
	{ return * impl().m_pDescription; }
	inline Allocator & allocator()
	{ return *(Allocator *)(impl().m_pAllocator); }

	inline char * node_to_data(PVOID pn)
	{ return (char *)pn + description().m_aligned_node_size; }
	inline char * data_to_node(PVOID pv)
	{ return (char *)pv - description().m_aligned_node_size; }

	inline const char * node_to_data(PCVOID pn)
	{ return (const char *)pn + description().m_aligned_node_size; }
	inline const char * data_to_node(PCVOID pv)
	{ return (const char *)pv - description().m_aligned_node_size; }

	inline HTREEDB_LIST raw_handle() const { return m_inst.m_handle; }
	inline HTREEDB_LIST & raw_handle() { return m_inst.m_handle; }

	TREEDB_IList_DELEGATE_TO(m_inst);
	treedb_list_t m_inst;
};

#undef V3C_DELEGATE
#undef DGTHIS
#undef DGTHIS0

/*! treedb_list_runtime_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class ListRuntime
{
public:
	enum { eImReallyAListRuntime = 0 };
	inline ListRuntime() { m_inst.m_pImpl = 0; }
	inline ListRuntime(ListImpl * p)
		{ m_inst.m_pImpl = (treedb_list_impl_ptr_t)p; }
	/* Convenience accessors. */
	inline const TreedbItemDescription & description() const
	{ return * m_inst.m_pImpl->m_pDescription; }
	inline Allocator & allocator()
	{ return *(Allocator *)(m_inst.m_pImpl->m_pAllocator); }

	treedb_list_runtime_t m_inst;
};

#define DGTHIS0(x) (&x)
#define DGTHIS(x) (&x),
#define V3C_DELEGATE(Inst, x) Inst.m_pVtbl->x

/*! treedb_list_impl_t wrapper

This class CANNOT have a vtable! It overlays it's m_inst member.
*/
class ListImpl
{
public:
	inline ListImpl() : m_inst() {}
	inline void setup
	( AllocatorOnStorage & allocator
	, treedb_list_backend_ptr_t pBackend
	, TreedbItemDescription * pd
	, treedb_const_list_impl_ptr_t ptemplate
	) {
		treedb_list_impl_t_Setup(& m_inst, & allocator.c_type(), pBackend, pd
			, ptemplate);
	}
	inline ListImpl
	( AllocatorOnStorage & allocator
	, treedb_list_backend_ptr_t pBackend
	, TreedbItemDescription * pd
	, treedb_const_list_impl_ptr_t ptemplate
	) {
		setup(allocator, pBackend, pd, ptemplate);
	}
	/* Convenience accessors. */
	inline const TreedbItemDescription & description() const
	{ return * m_inst.m_pDescription; }
	inline Allocator & allocator()
	{ return *(Allocator *)(m_inst.m_pAllocator); }

	inline char * node_to_data(PVOID pn)
	{ return (char *)pn + description().m_aligned_node_size; }
	inline char * data_to_node(PVOID pv)
	{ return (char *)pv - description().m_aligned_node_size; }

	inline const char * node_to_data(PCVOID pn)
	{ return (const char *)pn + description().m_aligned_node_size; }
	inline const char * data_to_node(PCVOID pv)
	{ return (const char *)pv - description().m_aligned_node_size; }

	TREEDB_IListImpl_DELEGATE_TO(m_inst);
	treedb_list_impl_t m_inst;
};

class ListBackend
{
public:
	TREEDB_IListBackend_DELEGATE_TO(m_inst);
	treedb_list_backend_t m_inst;
};

#undef V3C_DELEGATE
#undef DGTHIS0
#undef DGTHIS
#undef DGTHIS_MEM

////////////////////////////////////////////////////////////////////////////////
// Templates
////////////////////////////////////////////////////////////////////////////////
template< typename Impl >
class ListRuntimeT
: public ListRuntime
{
public:
	inline ListRuntimeT() : ListRuntime() {}
	inline ListRuntimeT(Impl * p) : ListRuntime((Impl *)p) {}
	inline Impl & impl() { return *(Impl *)m_inst.m_pImpl; }
};

template< typename Backend_ >
class ListImplT
: public ListImpl
{
public:
	typedef Backend_ Backend;
	typedef typename Backend::DataType DataType;

	ListImplT() {}
	inline void setup
	( AllocatorOnStorage & allocator
	, treedb_const_list_impl_ptr_t ptemplate
	) {
		ListImpl::setup(allocator, (treedb_list_backend_ptr_t)(& m_backend)
			, & m_description, ptemplate);
	}
	inline ListImplT
	( AllocatorOnStorage & allocator
	, treedb_const_list_impl_ptr_t ptemplate
	)
	: ListImpl
		( allocator
		, (treedb_list_backend_ptr_t)(& m_backend)
		, & m_description
		, ptemplate
		)
	{}
	inline Backend & backend() { return m_backend; }

	// Accessors
	const DataType & data(PCVOID pv)
	{
		return Backend::unpack_data(node_to_data((PVOID)pv));
	}
	DataType & data(PVOID pv)
	{
		return Backend::unpack_data(node_to_data(pv));
	}

	Backend m_backend;
	TreedbItemDescription m_description;
};

template < typename Backend_, typename Impl_ = ListImplT< Backend_ > >
class ListWithImplT;

template < typename Impl_ >
class ListT : public List
{
public:
	typedef Impl_ Impl;
	typedef ListT< Impl > ListType;
	typedef typename Impl::Backend Backend;
	typedef typename Backend::DataType DataType;
	typedef typename Backend::InputType InputType;
	typedef PackedPayload< Impl > HandleType;

	inline HandleType handle() const
		{ return HandleType(m_inst.m_handle); }
	inline HandleType & handle()
		{ return * (HandleType *)(& m_inst.m_handle); }

	ListT() {}
	void setup
	( Impl_ & timpl
	, HandleType payload
	) { treedb_list_t_Setup(& m_inst, & timpl.m_inst, payload.m_value); }
	ListT(Impl & timpl, HandleType payload)
	{ setup(timpl, payload); }
	ListT(ListWithImplT< Backend, Impl > & o, HandleType payload);
	Impl & impl() { return (Impl &)List::impl(); }
	inline Backend & backend() { return impl().backend(); }
	bool push_front(const InputType & d)
	{ return List::PushFront(Impl::Backend::pack_input(d)); }
	bool push_back(const InputType & d)
	{ return List::PushBack(Impl::Backend::pack_input(d)); }
	bool pop_front()
	{ return List::PopFront(); }
	bool pop_back()
	{ return List::PopBack(); }

	// Accessors
	const DataType & data(PCVOID pv) { return impl().data(pv); }
	DataType & data(PVOID pv) { return impl().data(pv); }

	const DataType & head() const
	{ return data((PCVOID)this->Head()); }
	const DataType & tail() const
	{ return data((PCVOID)this->Tail()); }

	DataType & head()
	{ return data((PVOID)this->PointerHead()); }
	DataType & tail()
	{ return data((PVOID)this->PointerTail()); }

	const DataType & front() const
	{ return data((PCVOID)this->PointerHead()); }
	const DataType & back() const
	{ return data((PCVOID)this->PointerTail()); }

	DataType & front()
	{ return data((PVOID)this->PointerHead()); }
	DataType & back()
	{ return data((PVOID)this->PointerTail()); }
};

template < typename Backend_, typename Impl_ /* = ListImplT< Backend_ > */ >
class ListWithImplT : public ListT< Impl_ >
{
public:
	typedef Backend_ Backend;
	typedef Impl_ Impl;
	typedef ListT< Impl_ > ListType;
	ListWithImplT()
	: ListType()
	, m_impl()
	{}
	void setup
	( treedb_const_list_ptr_t ptemplate
	, AllocatorOnStorage & allocator
	, PackedPayload< Impl > payload
	) {
		m_impl.setup(ptemplate->m_pImpl, allocator);
		ListType::setup(m_impl, ptemplate, payload);
	}
	ListWithImplT(AllocatorOnStorage & allocator
		, treedb_const_list_ptr_t ptemplate
		, PackedPayload< Impl > payload)
	: ListType()
	, m_impl(allocator, ptemplate->m_pImpl)
	{ ListType::setup(m_impl, payload); }
	Impl m_impl;
};

template < typename Impl >
ListT< Impl >::ListT
( ListWithImplT< ListT< Impl >::Backend, Impl > & o
, PackedPayload< Impl > payload
) {
	this->setup(o.impl(), & o.m_inst, payload);
}


template
< typename Derived
, typename Runtime
, typename BackendTypes
, uint16_t align
>
class ListBackendT : public IListBackendT< Runtime >
{
public:
	enum
	{ eIHaveARealListRuntime = Runtime::eImReallyAListRuntime
	};
	typedef typename BackendTypes::DataType DataType;
	typedef typename BackendTypes::InputType InputType;

	virtual PVOID STDCALL AllocNode
	( Runtime * runtime
	, HTREEDB_LISTDATA input
	) {
		Derived * This = static_cast<Derived *>(this);
		uintptr_t offset
			= runtime->description().m_aligned_node_size;
		uintptr_t data_size = This->size(Derived::unpack_input(input));
		if(!data_size)
			return 0;
		char * ret = (char *)runtime->allocator().PointerAlignedAlloc
					(offset + data_size, align, offset);
		if(!ret)
			return 0;
		char *p = ret + offset;
	#if META_TREEDB_DEBUGX
		cout << __FILE__ << '[' << __LINE__ << "] L2ListBackendT::AllocNode : "
			"writing \"";
		This->dump(cout, Derived::unpack_input(input));
		cout
			<< "\" to "
			<< impl->allocator().PointerToOffset(ret)
			<< " + " << offset << " = "
			<< impl->allocator().PointerToOffset(p)
			<< "." << endl;
	#endif
		This->assign(Derived::unpack_data(p), Derived::unpack_input(input));
		return ret;
	}
	virtual void STDCALL FreeNode
	( Runtime * runtime
	, PVOID pn
	) {
		runtime->allocator().FreePtr(pn);
	}
	virtual PVOID STDCALL alloc_node
	( Runtime * runtime
	, const InputType & input
	) {
		return AllocNode(runtime, Derived::pack_input(input));
	}
	void dump(std::ostream &, const InputType &) {}
};
template
< typename Derived
, typename Runtime
, typename BackendTypes
, uint16_t align
>
class FixedListBackendT : public ListBackendT< Derived, Runtime
, BackendTypes, align >
{
public:
	typedef typename BackendTypes::DataType DataType;
	typedef typename BackendTypes::InputType InputType;

	uintptr_t size(const DataType &) { return sizeof(DataType); }
	void assign(DataType & to, const InputType & from)
	{ to = from; }
};

} // namespace meta
} // namespace treedb
} // namespace v3c

#endif /* #ifndef v3c_meta_treedb_list_hpp_included */
