/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef v3c_meta_treedb_h_included
  #include "meta-treedb.h"
#endif

#ifndef v3c_meta_treedb_storage_h_included
  #define v3c_meta_treedb_storage_h_included

  #ifndef META_TREEDB_IN_V3C_DCOM
	#include <uuid.h>
	#include <sys/mman.h>
	#include <treedb/comet.h>
	#include <treedb/treedb-header.h>
	#include <v3c/file-lock.h>
	#include <pthread.h>
  #endif

  #define TREEDB_SHARED

  #define OMIT_TREEDB_STORAGE_RUNTIME_BASE

  #include <treedb/storage-defines.h>

  #undef OMIT_TREEDB_STORAGE_RUNTIME_BASE

  #ifdef __cplusplus
extern "C" {
  #endif

/******************************************************************************/
/* Delegates                                                                  */
/******************************************************************************/
  #define TREEDB_IStorage_DELEGATE_TO(x) \
inline TreedbHeapStatus STDCALL ReadLock() \
	{ return V3C_DELEGATE(x, ReadLock(DGTHIS0(x))); } \
inline TreedbHeapStatus STDCALL WriteLock() \
	{ return V3C_DELEGATE(x, WriteLock(DGTHIS0(x))); } \
inline TreedbHeapStatus STDCALL Unlock() \
	{ return V3C_DELEGATE(x, Unlock(DGTHIS0(x))); } \
inline StorageLockType STDCALL LockType() \
	{ return V3C_DELEGATE(x, LockType(DGTHIS0(x))); } \
inline bool STDCALL IsOpen() \
	{ return V3C_DELEGATE(x, IsOpen(DGTHIS0(x))); } \
inline TreedbHeapStatus STDCALL Close() \
	{ return V3C_DELEGATE(x, Close(DGTHIS0(x))); } \
inline TreedbHeapStatus STDCALL Open \
	( const uuid_t uSchema \
	, const char * name \
	, TreedbHeapOpenFlags open_flags \
	, int access_flags \
	, uint64_t offset \
	) \
	{ return V3C_DELEGATE(x, Open(DGTHIS(x) uSchema, name \
		, open_flags, access_flags, offset)); } \
inline TreedbHeapStatus STDCALL OpenUnknown \
	( const char * name \
	, TreedbHeapOpenFlags open_flags \
	, int access_flags \
	, uint64_t offset \
	) \
	{ return V3C_DELEGATE(x, OpenUnknown(DGTHIS(x) name \
		, open_flags, access_flags, offset)); } \
inline TreedbHeapStatus STDCALL Create( \
	treedb_allocator_on_storage_ptr_t pAllocator \
	, const uuid_t uSchema \
	, const char * name \
	, TreedbHeapOpenFlags open_flags \
	, int create_permissions \
	, int access_flags \
	, uint64_t offset \
	, uint64_t size \
	, uint16_t uPayloadSize \
	) \
	{ return V3C_DELEGATE(x, Create(DGTHIS(x) \
		(treedb_allocator_ptr_t)pAllocator, uSchema, name \
		, open_flags, create_permissions, access_flags, offset, size \
		, uPayloadSize)); } \
inline PVOID STDCALL MoreCore( size_t cb) \
	{ return V3C_DELEGATE(x, MoreCore(DGTHIS(x) cb)); } \
inline TreedbFreeResult STDCALL FreeCore(size_t cb) \
	{ return V3C_DELEGATE(x, FreeCore(DGTHIS(x) cb)); } \
inline BOOL STDCALL Resizable() \
	{ return V3C_DELEGATE(x, Resizable(DGTHIS0(x))); } \
inline void STDCALL PostFree( PVOID pv, size_t cb) \
	{ return V3C_DELEGATE(x, PostFree(DGTHIS(x) pv, cb)); }

/******************************************************************************/
/* forward declarations                                                       */
/******************************************************************************/
  #ifndef MetaTreedb_storage_typedefs_DEFINED
	#define MetaTreedb_storage_typedefs_DEFINED

typedef struct treedb_storage_t treedb_storage_t, * treedb_storage_ptr_t;
typedef struct treedb_storage_on_memory_t
	treedb_storage_on_memory_t
, * treedb_storage_on_memory_ptr_t;
typedef struct treedb_storage_on_file_t treedb_storage_on_file_t
, * treedb_storage_on_file_ptr_t;

/* v-table pointers should really be const pointers, but it's not a big deal.

NEVER,NEVER,NEVER think that you can omit these v-tables in c++!!!
Why:
1. It causes a compile error when you try to instantiate an abstract class
2. We obtain the implementation (along with the v-table) with a function call
   and c++ doesn't copy over the (hidden) v-table with respect to c++ objects
*/
typedef struct treedb_storage_vtbl_t treedb_storage_vtbl_t
, * treedb_storage_vtbl_ptr_t;
typedef struct treedb_storage_on_memory_vtbl_t
	treedb_storage_on_memory_vtbl_t
, * treedb_storage_on_memory_vtbl_ptr_t;
typedef struct treedb_storage_on_file_vtbl_t
	treedb_storage_on_file_vtbl_t
, * treedb_storage_on_file_vtbl_ptr_t;

  #endif /* #ifndef MetaTreedb_storage_typedefs_DEFINED */

typedef enum StorageLockType
{ eStorageReadLocked
, eStorageWriteLocked
, eStorageUnlocked
} StorageLockType;

/******************************************************************************/
/* treedb_storage_t                                                           */
/******************************************************************************/
  #define treedb_storage_t_methods(Imp, Prefix, Class) \
	STDMETHOD##Imp##_(TreedbHeapStatus, Prefix##ReadLock) \
		(THIS0(Class)) PURE; \
	STDMETHOD##Imp##_(TreedbHeapStatus, Prefix##WriteLock) \
		(THIS0(Class)) PURE; \
	STDMETHOD##Imp##_(TreedbHeapStatus, Prefix##Unlock) \
		(THIS0(Class)) PURE; \
	STDMETHOD##Imp##_(StorageLockType, Prefix##LockType) \
		(THIS0(Class)) PURE; \
	STDMETHOD##Imp##_(BOOL, Prefix##IsOpen) \
		(THIS0(Class)) PURE; \
	STDMETHOD##Imp##_(BOOL, Prefix##CanWrite) \
		(THIS0(Class)) PURE; \
	STDMETHOD##Imp##_(BOOL, Prefix##CanExec) \
		(THIS0(Class)) PURE; \
	STDMETHOD##Imp##_(TreedbHeapStatus, Prefix##Close) \
		(THIS0(Class)) PURE; \
	STDMETHOD##Imp##_(TreedbHeapStatus, Prefix##Open) \
		( THIS(Class) \
		  const uuid_t /*uSchema*/ \
		, const char * /*name*/ \
		, TreedbHeapOpenFlags /*open_flags*/ \
		, int /*access_flags*/ \
		, size_t /*offset*/ \
		) PURE; \
	STDMETHOD##Imp##_(TreedbHeapStatus, Prefix##OpenUnknown) \
		( THIS(Class) \
		  const char * /*name*/ \
		, TreedbHeapOpenFlags /*open_flags*/ \
		, int /*access_flags*/ \
		, size_t /*offset*/ \
		) PURE; \
	STDMETHOD##Imp##_(TreedbHeapStatus, Prefix##Create) \
		( THIS(Class) \
		  treedb_allocator_ptr_t /*pAllocator*/ \
		, const uuid_t /*uSchema*/ \
		, const char * /*name*/ \
		, TreedbHeapOpenFlags /*open_flags*/ \
		, int /*create_permissions*/ \
		, int /*access_flags*/ \
		, size_t /*offset*/ \
		, size_t /*size*/ \
		, size_t /*uPayloadSize*/ \
		) PURE; \
	STDMETHOD##Imp##_(PVOID, Prefix##MoreCore) \
		(THIS(Class) size_t /*cb*/) PURE; \
	STDMETHOD##Imp##_(TreedbFreeResult, Prefix##FreeCore) \
		(THIS(Class) size_t /*cb*/) PURE; \
	STDMETHOD##Imp##_(BOOL, Prefix##Resizable) \
		(THIS0(Class)) PURE; \
	STDMETHOD##Imp##_(void, Prefix##PostFree) \
		(THIS(Class) PVOID /*pv*/, size_t /*cb*/) PURE; \
	\
	STDMETHOD##Imp##_(uint8_t, Prefix##EndianIndex) \
		( THIS0(Class) ) PURE; \
	STDMETHOD##Imp##_(uint8_t, Prefix##EndianSwapIndex) \
		( THIS0(Class) ) PURE

  #ifdef COBJMACROS

	#define treedb_storage_t_ReadLock(This) V3C_IFACE_CALL(ReadLock, This)
	#define treedb_storage_t_WriteLock(This) V3C_IFACE_CALL(WriteLock, This)
	#define treedb_storage_t_Unlock(This) V3C_IFACE_CALL(Unlock, This)
	#define treedb_storage_t_Close(This) V3C_IFACE_CALL(Close, This)
	#define treedb_storage_t_OpenUnknown(This, n,of,af,o2) \
	  V3C_IFACE_CALL(OpenUnknown, This, n,of,af,o2)
	#define treedb_storage_t_Open(This, s,n,of,af,o2) \
	  V3C_IFACE_CALL(Open, This, s,n,of,af,o2)
	#ifdef __cplusplus
	  #define treedb_storage_t_Create(This, a,s,n,of,cf,af,o,s2,p2) \
		V3C_IFACE_CALL(Create, This, a,s,n,of,cf,af,o,s2,p2)
	#else
	  #define treedb_storage_t_Create(This, a,s,n,of,cf,af,o,s2,p2) \
		V3C_IFACE_CALL(Create, This, (treedb_allocator_ptr_t)(a) \
		  ,s,n,of,cf,af,o,s2,p2)
	#endif
	#define treedb_storage_t_MoreCore(This, cb) \
	  V3C_IFACE_CALL(MoreCore, This, cb)
	#define treedb_storage_t_FreeCore(This, cb) \
	  V3C_IFACE_CALL(FreeCore, This, cb)
	#define treedb_storage_t_Start(This) \
	  (This)->m_pBase
	#define treedb_storage_t_End(This) \
	  ((This)->m_pBase + (This)->m_uSize)
	#define treedb_storage_t_Size(This) \
	  (This)->m_uSize
	#define treedb_storage_t_Resizable(This, pv) \
	  V3C_IFACE_CALL(Resizable, This)
	#define treedb_storage_t_PostFree(This, pv, cb) \
	  V3C_IFACE_CALL(PostFree, This, pv, cb)
	#define treedb_storage_t_InHeap(This, u) \
	  (u < (This)->m_uSize)
	#define treedb_storage_t_InHeapPtr(This, pv) \
	  (((const char *)pv >= (This)->m_pBase) \
		&& ((const char *)pv < ((This)->m_pBase + (This)->m_uSize)))
	#define treedb_storage_t_PointerUnpack(This, u) \
	  ((u) ? (This)->m_pBase + (u) : 0)
	#define treedb_storage_t_PointerPack(This, p) \
	  ((p) ? (uintptr_t)((char *)(p) - (This)->m_pBase) : (uintptr_t)0)

	#define treedb_storage_t_EndianIndex(This) \
	  V3C_IFACE_CALL(EndianIndex, This)
	#define treedb_storage_t_EndianSwapIndex(This) \
	  V3C_IFACE_CALL(EndianSwapIndex, This)

  #endif /* #ifdef COBJMACROS */

  #define treedb_storage_vtbl_t_entries_default(Prefix) \
	& Prefix##ReadLock \
	, & Prefix##WriteLock \
	, & Prefix##Unlock \
	, & Prefix##LockType \
	, & Prefix##IsOpen \
	, & Prefix##CanWrite \
	, & Prefix##CanExec \
	, & Prefix##Close \
	, & Prefix##Open \
	, & Prefix##OpenUnknown \
	, & Prefix##Create \
	, & Prefix##MoreCore \
	, & Prefix##FreeCore \
	, & Prefix##Resizable \
	, & Prefix##PostFree \
	\
	, & Prefix##EndianIndex \
	, & Prefix##EndianSwapIndex

  #ifndef treedb_storage_vtbl_t_entries
	#define treedb_storage_vtbl_t_entries(x) \
	  treedb_storage_vtbl_t_entries_default(x)
  #endif

  #ifndef MetaTreedb_treedb_storage_t_DEFINED
	#define MetaTreedb_treedb_storage_t_DEFINED

	#define treedb_storage_t_members(vtbl_ptr_t) \
	  vtbl_ptr_t m_pVtbl; \
	  char * m_pBase; \
	  uintptr_t m_uSize;

    #ifdef __cplusplus
	  #include "meta-treedb-cxx-push-c.h"
	#endif
struct treedb_storage_vtbl_t
{
	treedb_storage_t_methods(,,treedb_storage_t);
};
	#ifdef __cplusplus
	  #include "meta-treedb-cxx-pop-c.h"
	#endif

struct treedb_storage_t
{
	#ifdef __cplusplus
	  #define DGTHIS0(x) this
	  #define DGTHIS(x)  this,
	  #define DGTHIS_MEM(x,m) x.m
	  #define V3C_DELEGATE(Inst, x) m_pVtbl->x

	TREEDB_IStorage_DELEGATE_TO();
	  #undef DGTHIS0
	  #undef DGTHIS
	  #undef DGTHIS_MEM
	  #undef V3C_DELEGATE
	#endif
	treedb_storage_t_members(treedb_storage_vtbl_ptr_t);
};

  #endif /* #ifndef MetaTreedb_treedb_storage_t_DEFINED */

/******************************************************************************/
/* treedb_storage_on_memory_t                                                 */
/******************************************************************************/
  #ifndef MetaTreedb_treedb_storage_on_memory_t_DEFINED
	#define MetaTreedb_treedb_storage_on_memory_t_DEFINED

	#ifdef __cplusplus
	  #include "meta-treedb-cxx-push-c.h"
	#endif
struct treedb_storage_on_memory_vtbl_t
{
	treedb_storage_t_methods(,,treedb_storage_on_memory_t);
};
	#ifdef __cplusplus
	  #include "meta-treedb-cxx-pop-c.h"
	#endif

struct treedb_storage_on_memory_t
{
	#ifdef __cplusplus
	  #define DGTHIS0(x) this
	  #define DGTHIS(x)  this,
	  #define DGTHIS_MEM(x,m) x.m
	  #define V3C_DELEGATE(Inst, x) m_pVtbl->x

	TREEDB_IStorage_DELEGATE_TO();
	  #undef DGTHIS0
	  #undef DGTHIS
	  #undef DGTHIS_MEM
	  #undef V3C_DELEGATE
	#endif
	treedb_storage_t_members(treedb_storage_on_memory_vtbl_ptr_t);
};

	#define TREEDB_STORAGE_ON_MEMORY_INIT(p, s) \
	  { & treedb_storage_on_memory_vtbl, p, s }

  #endif /* #ifndef MetaTreedb_treedb_storage_on_memory_t_DEFINED */

extern treedb_storage_on_memory_vtbl_t
	treedb_storage_on_memory_vtbl;

/******************************************************************************/
/* treedb_storage_on_file_t                                                   */
/******************************************************************************/
extern void treedb_storage_on_file_t_Setup
( treedb_storage_on_file_ptr_t ps);

  #ifndef MetaTreedb_treedb_storage_on_file_t_DEFINED
	#define MetaTreedb_treedb_storage_on_file_t_DEFINED

	#ifdef __cplusplus
	  #include "meta-treedb-cxx-push-c.h"
	#endif
struct treedb_storage_on_file_vtbl_t
{
	treedb_storage_t_methods(,,treedb_storage_on_file_t);
};
	#ifdef __cplusplus
	  #include "meta-treedb-cxx-pop-c.h"
	#endif

	#undef TREEDB_STORAGE_RUNTIME_BASE_MEMBERS
	#define TREEDB_STORAGE_RUNTIME_BASE_MEMBERS \
	  union { \
		volatile char * m_pBase; \
		volatile TreedbHeaderPtr m_pTreedb; \
	  }; \
	  size_t m_uSize; \
	  uint8_t m_uEndianIndex, m_uEndianSwapIndex;

	#undef TREEDB_STORAGE_RUNTIME_BASE_MEMBERS_INIT

	#define TREEDB_STORAGE_RUNTIME_BASE_MEMBERS_INIT {}, 0, 0, 0

extern treedb_storage_on_file_vtbl_t treedb_storage_on_file_vtbl;

struct treedb_storage_on_file_t
{
	#ifdef __cplusplus
	inline treedb_storage_on_file_t()
	: m_pVtbl(& treedb_storage_on_file_vtbl)
	, m_uSize(0), m_uEndianIndex(0), m_uEndianSwapIndex(0)
	{}
	  #define DGTHIS0(x) this
	  #define DGTHIS(x)  this,
	  #define DGTHIS_MEM(x,m) x.m
	  #define V3C_DELEGATE(Inst, x) m_pVtbl->x

	TREEDB_IStorage_DELEGATE_TO();
	  #undef DGTHIS0
	  #undef DGTHIS
	  #undef DGTHIS_MEM
	  #undef V3C_DELEGATE
	#endif
	treedb_storage_on_file_vtbl_ptr_t m_pVtbl;
	TREEDB_STORAGE_RUNTIME_MEMBERS;
};
	#define TREEDB_STORAGE_ON_FILE_INIT_BARE \
	  & treedb_storage_on_file_vtbl, TREEDB_STORAGE_RUNTIME_INIT_BARE
	#define TREEDB_STORAGE_ON_FILE_INIT { TREEDB_STORAGE_ON_FILE_INIT_BARE }

  #endif /* #ifndef MetaTreedb_treedb_storage_on_file_t_DEFINED */

  #ifdef __cplusplus
} /* extern "C" { */
  #endif

#endif /* #ifndef v3c_meta_treedb_storage_h_included */
