/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef v3c_meta_treedb_storage_hpp_included
#define v3c_meta_treedb_storage_hpp_included

#ifndef v3c_meta_treedb_storage_h_included
#include "meta-treedb-storage.h"
#endif

namespace v3c {
namespace treedb {
namespace meta {

/*! treedb_storage_t vtable-only overlay
*/
class IStorage_VtableOverlay
{
public:
	treedb_storage_t_methods(,,treedb_storage_t);
};

template< typename Interface = IStorage_VtableOverlay >
class IStorageT : public Interface
{
public:
	inline IStorageT(char *b)
	: m_pBase(b){}

	inline operator treedb_storage_t &()
	{ return *(treedb_storage_ptr_t)this; }
	inline operator const treedb_storage_t &() const
	{ return *(treedb_storage_ptr_t)this; }

	char * m_pBase;
};
template< typename Interface = IStorage_VtableOverlay >
class IStorageOnMemoryT : public IStorageT< Interface >
{
public:
	inline IStorageOnMemoryT(char *b, uintptr_t sz)
	: IStorageT< Interface >(b)
	, m_uSize(sz)
	{}

	inline operator treedb_storage_on_memory_t &()
	{ return *(treedb_storage_on_memory_ptr_t)this; }
	inline operator const treedb_storage_on_memory_t &() const
	{ return *(treedb_storage_on_memory_ptr_t)this; }
	uintptr_t m_uSize;
};

////////////////////////////////////////////////////////////////////////////////
// Wrappers
////////////////////////////////////////////////////////////////////////////////
#define DGTHIS0(x) &x
#define DGTHIS(x) &x,
#define DGTHIS_MEM(x,m) x.m
#define V3C_DELEGATE(Inst, x) Inst.m_pVtbl->x

template< typename Instance = treedb_storage_t >
class StorageT
{
public:
	inline StorageT(char *b)
	{
		m_inst.m_pBase = b;
	}
	inline treedb_storage_t & base_instance()
	{
		// g++ way of abiding with strict aliasing rules.
		union { Instance * i; treedb_storage_t * j; } u;
		u.i = & m_inst;
		return * u.j;
	}
	inline const treedb_storage_t & base_instance() const
	{
		// g++ way of abiding with strict aliasing rules.
		union { const Instance * i; const treedb_storage_t * j; } u;
		u.i = & m_inst;
		return * u.j;
	}

	inline Instance & instance()
	{ return m_inst; }
	inline const Instance & instance() const
	{ return m_inst; }
	TREEDB_IStorage_DELEGATE_TO(m_inst);

	inline char * Start() { return m_inst.m_pBase; }
	inline char * End() { return m_inst.m_pBase + m_inst.m_uSize; }
	inline uintptr_t Size() { return m_inst.m_uSize; }

	inline char * OffsetToPointer(uintptr_t u)
	{ return (char *)(u ? m_inst.m_pBase + u : 0); }
	inline uintptr_t PointerToOffset(PCVOID p)
	{ return p ? (uintptr_t)((char *)p - m_inst.m_pBase) : (uintptr_t)0; }

	Instance m_inst;
};
typedef StorageT<> Storage;

template< typename Instance = treedb_storage_on_memory_t >
class StorageOnMemoryT
: public StorageT<Instance>
{
public:
	typedef StorageT<Instance> Base;
	inline StorageOnMemoryT(char *b, size_t l)
	: Base(b)
	{
		Base::m_inst.m_uSize = l;
		Base::m_inst.m_pVtbl = & treedb_storage_on_memory_vtbl;
	}
};
typedef StorageOnMemoryT<> StorageOnMemory;

template< typename Instance = treedb_storage_on_file_t >
class StorageOnFileT
: public StorageT< Instance >
{
public:
	typedef StorageT< Instance > Base;
	inline StorageOnFileT()
	: Base(0)
	{
		treedb_storage_on_file_t_Setup(& Base::instance());
	}
	uint64_t payload_offset()
	{ return Base::m_inst.m_pTreedb->m_ulPayloadOffset; }
};

typedef StorageOnFileT<> StorageOnFile;

#undef V3C_DELEGATE
#undef DGTHIS0
#undef DGTHIS
#undef DGTHIS_MEM

} // namespace meta
} // namespace treedb
} // namespace v3c

#endif /* #ifndef v3c_meta_treedb_storage_hpp_included */
