/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef v3c_meta_treedb_h_included
  #define v3c_meta_treedb_h_included

  #include <treedb/treedb.h>

  #ifndef META_TREEDB_DEBUG
	#define META_TREEDB_DEBUG 0
  #endif

  #if (META_TREEDB_DEBUG & 1)
	#define META_TREEDB_printf(fmt, ...) \
		fprintf(stderr, fmt, ## __VA_ARGS__); fflush(stderr)
	#define META_TREEDB_cout cerr
  #else
	#define META_TREEDB_printf(fmt, ...)
	#define META_TREEDB_cout if(0) cerr
  #endif

  #if (META_TREEDB_DEBUG & 2)
	#define META_TREEDB_printf2(fmt, ...) \
		fprintf(stderr, fmt, ## __VA_ARGS__); fflush(stderr)
	#define META_TREEDB_cout2 cerr
  #else
	#define META_TREEDB_printf2(fmt, ...)
	#define META_TREEDB_cout2 if(0) cerr
  #endif

  #if (META_TREEDB_DEBUG & 4)
	#define META_TREEDB_printf4(fmt, ...) \
		fprintf(stderr, fmt, ## __VA_ARGS__); fflush(stderr)
	#define META_TREEDB_cout4 cerr
  #else
	#define META_TREEDB_printf4(fmt, ...)
	#define META_TREEDB_cout4 if(0) cerr
  #endif

  #ifdef __cplusplus
extern "C" {
  #endif


  #ifndef MetaTreedb_TreedbItemDescriptionEntry_DEFINED
typedef struct TreedbItemDescriptionEntry_tag {
	int32_t m_uSize;   /* < 0 means  (1 - m_uSize) * sizeof(count_t) units */
	uint32_t m_uAlign; /* m_uSize < 0 means use count_t alignment */
} TreedbItemDescriptionEntry, * TreedbItemDescriptionEntryPtr;
typedef const TreedbItemDescriptionEntry * TreedbItemDescriptionEntryConstPtr;
	#define MetaTreedb_TreedbItemDescriptionEntry_DEFINED
  #endif

  #define TREEDB_ITEM_DESC_BEGIN(Name) \
	static TreedbItemDescriptionEntry Name[] = {
  #define TREEDB_ITEM_DESC_ENTRY(Size, Align) { Size, Align },
  #define TREEDB_ITEM_DESC_END {0, 0}}

  #define TreedbItemDescription_members \
	TreedbItemDescriptionEntry m_pointer, m_root, m_node; \
	uint16_t m_flags, m_aligned_node_size

  #define META_TREEDB_ITEM_DESCRIPTION_INIT_BARE \
	{0,0}, {0,0}, {0,0}, 0, 0

  #define META_TREEDB_ITEM_DESCRIPTION_INIT \
	{ META_TREEDB_ITEM_DESCRIPTION_INIT_BARE }

  #ifndef MetaTreedb_TreedbItemDescription_DEFINED
typedef struct TreedbItemDescription_tag {
	TreedbItemDescription_members;
} TreedbItemDescription;
typedef const TreedbItemDescription * TreedbItemDescriptionConstPtr;
	#define MetaTreedb_TreedbItemDescription_DEFINED
  #endif

/******************************************************************************/
/* enums                                                                      */
/******************************************************************************/
  #ifndef MetaTreedb_ETreedbTreeKind_DEFINED
	#define MetaTreedb_ETreedbTreeKind_DEFINED

typedef enum ETreedbContainerKind_tag
{ eTreedbContainerAllowsDuplicates = 1 /* n/a for lists */
, eTreedbContainerIsPointer = 2
, eTreedbContainerHasCount = 4
, eTreedbContainerHasDepth = 8 /* n/a for lists */
} ETreedbContainerKind;
  #endif

/******************************************************************************/
/* Later                                                                      */
/******************************************************************************/
typedef enum ETreedbComputeStructOffsetResult_tag
{ eTreedbComputeStructOffsetSuccess
, eTreedbComputeStructOffsetZeroAlign
, eTreedbComputeStructOffsetBadSize
} ETreedbComputeStructOffsetResult;

extern ETreedbComputeStructOffsetResult
treedb_compute_struct_offsets
( TreedbItemDescriptionEntryConstPtr p
, TreedbItemDescriptionEntryConstPtr schema
, uintptr_t * puStart
, uintptr_t * out
);

/******************************************************************************/
/* forward declarations                                                       */
/******************************************************************************/
  #ifndef MetaTreedb_typedefs_DEFINED
typedef struct treedb_tree_t treedb_tree_t
, * treedb_tree_ptr_t;
typedef const treedb_tree_t * treedb_const_tree_ptr_t;
	#ifdef __cplusplus
typedef treedb_tree_t & treedb_tree_ref_t;
	#else
	  #define treedb_tree_ref_t treedb_tree_ptr_t
	#endif

typedef struct treedb_allocator_t treedb_allocator_t
, * treedb_allocator_ptr_t;

	#ifdef __cplusplus
typedef treedb_allocator_t & treedb_allocator_ref_t;
	#else
	  #define treedb_allocator_ref_t treedb_allocator_ptr_t
	#endif

typedef struct treedb_allocator_on_storage_t
	treedb_allocator_on_storage_t
	, * treedb_allocator_on_storage_ptr_t;
	#ifdef __cplusplus
typedef treedb_allocator_on_storage_t & treedb_allocator_on_storage_ref_t;
	#else
	  #define treedb_allocator_on_storage_ref_t \
		treedb_allocator_on_storage_ptr_t
	#endif
  #endif

extern char * treedb_string_append(char * str, const char * more);

typedef size_t HTREEDB_HANDLE;

/******************************************************************************/
/* For arrays + handle_arrays                                                 */
/******************************************************************************/
typedef size_t HTREEDB_ARRAY;
typedef size_t HHTREEDB_ARRAY;
typedef PCVOID HTREEDB_ARRAYDATA;

  #ifdef __cplusplus
} /* extern "C" { */
  #endif

#define TREEDB_IFACE_IMPL_CCALL(Fn, This, ...) \
	((This)->m_pImpl->m_pVtbl->Fn(This, ## __VA_ARGS__))

  #include "meta-treedb-tree.h"
  #include "meta-treedb-list.h"
  #include "meta-treedb-storage.h"
  #include "meta-treedb-allocator.h"

#endif /* #ifndef v3c_meta_treedb_h_included */
