/* Copyright (C) 2010  Philip Ashmore (contact@philipashmore.com)             */
/* License: LGPLv3.  See LICENSE.txt for the full license.                    */

#ifndef v3c_meta_treedb_hpp_included
#define v3c_meta_treedb_hpp_included

#ifndef v3c_meta_treedb_h_included
#include "meta-treedb.h"
#endif

namespace v3c {
namespace treedb {
namespace meta {

template< typename T >
struct PackedPayload {
	typedef PackedPayload<T> * Ptr;
	explicit PackedPayload(uintptr_t value = 0) : m_value(value) {}
	size_t m_value;
};

template< typename T >
inline std::ostream & operator <<(std::ostream & s, const PackedPayload< T > & v)
{ s << v.m_value; return s; }

// Forward declarations.
template< typename Instance = treedb_allocator_t >
class AllocatorT;
template< typename Instance = treedb_two_tree_allocator_t >
class TwoTreeAllocatorT;
template< typename Instance = treedb_allocator_on_storage_t >
class AllocatorOnStorageT;

typedef AllocatorT<> Allocator, * AllocatorPtr;
typedef TwoTreeAllocatorT<> TwoTreeAllocator, * TwoTreeAllocatorPtr;
typedef AllocatorOnStorageT<> AllocatorOnStorage, * AllocatorOnStoragePtr;

} // namespace meta
} // namespace treedb
} // namespace v3c

#ifndef v3c_meta_treedb_tree_hpp_included
#include "meta-treedb-tree.hpp"
#endif
#ifndef v3c_meta_treedb_storage_hpp_included
#include "meta-treedb-storage.hpp"
#endif
#ifndef v3c_meta_treedb_allocator_hpp_included
#include "meta-treedb-allocator.hpp"
#endif
#ifndef v3c_meta_treedb_list_hpp_included
#include "meta-treedb-list.hpp"
#endif

namespace v3c {
namespace treedb {
namespace meta {

inline TreeImpl::TreeImpl
( Allocator & allocator
, treedb_tree_backend_ptr_t pBackend
, TreedbItemDescription * pd
, treedb_const_tree_impl_ptr_t ptemplate
) {
	treedb_tree_impl_t_Setup(& m_inst, & allocator.c_type(), pBackend, pd
		, ptemplate);
}
inline void TreeImpl::setup
( Allocator & allocator
, treedb_tree_backend_ptr_t pBackend
, TreedbItemDescription * pd
, treedb_const_tree_impl_ptr_t ptemplate
) {
	treedb_tree_impl_t_Setup(& m_inst, & allocator.c_type(), pBackend, pd
		, ptemplate);
}
inline void TreeImpl::setup
( AllocatorOnStorage & allocator
, treedb_tree_backend_ptr_t pBackend
, TreedbItemDescription * pd
, treedb_const_tree_impl_ptr_t ptemplate
) {
	treedb_tree_impl_t_Setup(& m_inst
		, (treedb_allocator_t *)(& allocator.c_type()), pBackend, pd
		, ptemplate);
}
template < typename Impl_ >
inline void TreeT< Impl_ >::dump()
{
	DataRenderer data_renderer(description().m_aligned_node_size);
#if META_TREEDB_DEBUG
	cout << __FILE__ << '[' << __LINE__ << "] Tree : aligned node size "
		<< description().m_aligned_node_size << '.' << endl;
#endif
	FdTreeRenderer renderer(STDOUT_FILENO);
	allocator().Dump(& renderer.c_type(), & data_renderer.c_type());
}

} // namespace meta
} // namespace treedb
} // namespace v3c

#endif /* #ifndef v3c_meta_treedb_hpp_included */
